#include "compiler.h"
#include "ctkparser.h"
#include "trampoline.h"
#include <malloc.h>

#ifndef MAX_PATH_LEN
#define MAX_PATH_LEN 1024
#endif 

#define TOP(t)      (t)->stack[(t)->sp-1]
#define POP(t)      ((t)->sp -= 1)
#define POP2(t)     ((t)->sp -= 2)
#define POP4(t)     ((t)->sp -= 4)
//#define POP(t)      assert(((t)->sp -= 1) >= 0)
//#define POP2(t)     assert(((t)->sp -= 2) >= 0)
//#define POP4(t)     assert(((t)->sp -= 4) >= 0)
#define PUSH(t, o)  ((t)->stack[(t)->sp] = (o), (t)->sp += 1)
#define PUSH_CS(t, o) { \
     CtkMemoryManager::instance.beginAccess(t); \
     (t)->stack[(t)->sp] = (o); \
     (t)->sp += 1; \
     CtkMemoryManager::instance.endAccess(t); }

#define PUSH_CHECK(t, o) \
     if ((t)->sp == (t)->stackSize) ctkThrowException("Stack overflow"); \
     else ((t)->stack[(t)->sp] = (o), (t)->sp += 1)

#define PUSH_CHECK_CS(t, o) { \
     CtkMemoryManager::instance.beginAccess(t); \
     if ((t)->sp == (t)->stackSize) { \
         ctkThrowException("Stack overflow"); \
     } else { \
         (t)->stack[(t)->sp] = (o); \
         (t)->sp += 1; \
     } \
     CtkMemoryManager::instance.endAccess(t); }

#define THROW_EXCEPTION(x) (frame->call = this, ctkThrowException(x))

CtkCompiler CtkCompiler::instance;

CtkObject CtkExpr::getLvalue(CtkFrame* frame) { 
    THROW_EXCEPTION("Lvalue expected");
    return ctkNull;
}

void CtkExpr::setLvalue(CtkFrame* frame, CtkObject) { 
    THROW_EXCEPTION("Lvalue expected");
}

void CtkExpr::assign(CtkFrame* frame, CtkObject) { 
    THROW_EXCEPTION("Lvalue expected");
}

bool CtkExpr::isLvalue() { 
    return false;
}

CtkFuncDef* CtkExpr::getFuncDef() { 
    return NULL;
}

void CtkExpr::compile(CtkCompiler*, CompileCtx) {}

CtkObject CtkExpr::getRvalue(CtkFrame* frame) {
    PUSH_CHECK_CS(frame->thread, ctkNull);
    return ctkNull;
}


void CtkStmt::compile(CtkCompiler*) {}

JumpType CtkStmt::execute(CtkFrame*) { 
    return JMP_NEXT;
}

JumpType CtkStmt::executeBlock(CtkFrame* frame) { 
    CtkStmt* stmt = this;
    while (stmt != NULL) { 
	JumpType jump = stmt->execute(frame);
	if (jump != JMP_NEXT) { 
	    return jump;
	}
	stmt = stmt->next;
    }
    return JMP_NEXT;
}


#ifdef _WIN32
#define FILE_SEP '\\'
#else
#define FILE_SEP '/'
#endif

CtkModule::~CtkModule()
{
    delete[] path;
    delete   ctx;
    delete   prog;
}

void CtkCompiler::reset()
{
    CtkModule *mod, *nextMod;
    for (mod = modules; mod != NULL; mod = nextMod) {         
        nextMod = mod->next;
        delete mod;
    }
    
    CtkImport *imp, *nextImp;
    for (imp = importList; imp != NULL; imp = nextImp) { 
        nextImp = imp->next;
        delete imp;
    }    
    program = NULL;
    modules = NULL;
    importList = NULL;
}

void CtkCompiler::compileBlock(CtkStmt* stmts) { 
    CtkSymbol* beforeSymbols = currCtx->symbols;
    while (stmts != NULL) { 
	stmts->compile(this);
	stmts = stmts->next;
    }
    CtkSymbol* afterSymbols = currCtx->symbols;
    while (beforeSymbols != afterSymbols) {
	CtkSymbol* symbol = afterSymbols;
	afterSymbols = afterSymbols->next;
	delete symbol;
    }
    currCtx->symbols = beforeSymbols;
}

extern int yyparse();

CtkModule* CtkCompiler::compile(char const* moduleName)
{
    critical_section cs(mutex);
    return compile(CtkSymbolTable::instance.add(moduleName, IDENT));
}
    
CtkModule* CtkCompiler::compile(CtkToken* moduleName)
{
    critical_section cs(mutex);
    for (CtkModule* module = modules; module != NULL; module = module->next) { 
	if (module->tkn->symId == moduleName->symId) {  
	    return module;
	}
    }	 
    for (CtkPath* p = path; p != NULL; p = p->next) { 
	char buf[MAX_PATH_LEN];
	sprintf(buf, "%s%c%s.ctk", p->dir, FILE_SEP, moduleName->name);
	FILE* f = fopen(buf, "r");
	if (f != NULL) { 
	    CtkScanner::instance.reset(f);
	    CtkModule* saveCurrModule = currModule;
	    CtkModule* curr = new CtkModule(moduleName, buf, modules);
	    currModule = modules = curr;
	    yyparse();
	    fclose(f);
	    program->compile(this);
	    int nVars = curr->prog->nVars;	
	    CtkThread* currThread = ctkGetCurrentThread();
	    CtkMemoryManager::instance.beginAccess(currThread);
	    CtkFrame* frame = 
		(CtkFrame*)ctkAllocateObject(sizeof(CtkFrame) 
					     + (nVars-1)*sizeof(CtkObject), NULL);
	    frame->nVars = nVars;
	    frame->call = NULL;
	    frame->up = NULL;
	    memset(frame->vars, 0, nVars*sizeof(CtkObject));
	    curr->frame = frame;
	    CtkFunction* func = (CtkFunction*)ctkAllocateObject(sizeof(CtkFunction), NULL);
	    func->nestedLevel = 0;
	    func->nParams = 0;
	    func->nLocalVars = nVars;
	    func->next = NULL;
	    func->exterior = NULL;
	    func->stmts = curr->prog->stmts; 
	    func->name = "<main>";
	    func->module = curr;
	    frame->func = func;
	    CtkFrame* topFrame = currThread->currFrame;
	    if (topFrame == NULL) { 
		currThread->currFrame = currThread->callStack = frame;
	    } else { 
		topFrame->up = frame;
	    }
	    frame->thread = currThread;
	    CtkMemoryManager::instance.endAccess(currThread);
	    curr->prog->execute(curr->frame);
	    currModule = saveCurrModule;
	    currThread->currFrame = topFrame;
	    if (topFrame != NULL) { 
		topFrame->up = NULL;
	    }
	    return curr;
	} 
    }
    return NULL;
}

void CtkCompiler::error(char const* errText, CtkNode* node)
{
    ctkTrace("%s:%d: %s\n", currModule->path, node->line, errText);
    exit(1);
}


void CtkProgram::compile(CtkCompiler* compiler) { 
    CtkImport* saveImport = compiler->importList;
    compiler->importList = NULL;
    for (CtkTokenList* imp = modules; imp != NULL; imp = imp->next) { 
	CtkModule* module = compiler->compile(imp->tkn);  
	if (module == NULL) { 
	    char buf[256];
	    sprintf(buf, "Module %s not found", imp->tkn->name);
	    compiler->error(buf, stmts);
	}
	compiler->importList = new CtkImport(module, compiler->importList);
    }
    CtkContext* ctx = new CtkContext();
    ctx->nSymbols = 0;
    ctx->symbols = NULL;
    ctx->exterior = NULL;
    compiler->currCtx = ctx;
    for (CtkStmt* stmt = stmts; stmt != NULL; stmt = stmt->next) { 
	 stmt->compile(compiler);	 
    }
    nVars = ctx->nSymbols;
    compiler->currModule->ctx = ctx;
    compiler->currModule->prog = this;
    compiler->importList = saveImport;
}

JumpType CtkProgram::execute(CtkFrame* frame)
{
    CtkThread* currThread = (CtkThread*)ctkGetCurrentThread();
#ifdef USE_CTALK_EXCEPTION
    try {
#else
    jmp_buf buf;
    jmp_buf* saveBuf = currThread->catchCtx;
    currThread->catchCtx = &buf;
    if (setjmp(buf) == 0) { 
#endif
	for (CtkStmt* stmt = stmts; stmt != NULL; stmt = stmt->next) { 
            JumpType jump = stmt->execute(frame);
	    switch (jump) { 
	      case JMP_EXCEPTION:
		ctkTrace("Uncatched exception %s\n", 
			IF_STRING(currThread->exceptionObject));
		ctkPrintStackTrace(frame);
		exit(1);
	      case JMP_RETURN:
#ifndef USE_CTALK_EXCEPTION
                currThread->catchCtx = saveBuf;
#endif
		return JMP_RETURN;
	      case JMP_NEXT:
		continue;
	      default:
		assert(false);
	    }
	} 
#ifndef USE_CTALK_EXCEPTION
        currThread->catchCtx = saveBuf;
#endif
#ifdef USE_CTALK_EXCEPTION
    } catch (CtkException const& x) {
#else
    } else { 
        currThread->catchCtx = saveBuf;
#endif
	ctkTrace("Uncatched exception %s\n", 
		IF_STRING(currThread->exceptionObject));
	ctkPrintStackTrace(frame);
	exit(1);
    }
    return JMP_NEXT;
}

void CtkIfStmt::compile(CtkCompiler* compiler)
{
    cond->compile(compiler);
    compiler->compileBlock(thenStmt);
    if (elseStmt != NULL) { 
	compiler->compileBlock(elseStmt);
    }
}

JumpType CtkIfStmt::execute(CtkFrame* frame)
{
    if (cond->testCondition(frame)) { 
	return thenStmt->executeBlock(frame);
    } else if (elseStmt != NULL) { 
	return elseStmt->executeBlock(frame);
    } else { 
	return JMP_NEXT;
    }
}

void CtkForStmt::compile(CtkCompiler* compiler)
{
    if (init != NULL) { 
	init->compile(compiler);
    }
    if (cond != NULL) { 
	cond->compile(compiler);
    }
    if (inc != NULL) { 
	inc->compile(compiler);
    }
    compiler->compileBlock(body);
}

JumpType CtkForStmt::execute(CtkFrame* frame)
{
    if (init != NULL) { 
	init->evaluate(frame);
    }
    while (cond == NULL || cond->testCondition(frame)) { 
	JumpType jump = body->executeBlock(frame);
	switch (jump) { 
	  case JMP_NEXT:
	  case JMP_CONTINUE:
	    break;
	  case JMP_BREAK:
	    return JMP_NEXT;	    
	  default:
	    return jump;
	}
	if (inc != NULL) { 
	    inc->evaluate(frame);
	}
    }
    return JMP_NEXT;	    
}


void CtkDoWhileStmt::compile(CtkCompiler* compiler)
{
    compiler->compileBlock(body);
    cond->compile(compiler);
}

JumpType CtkDoWhileStmt::execute(CtkFrame* frame)
{
    do { 
	JumpType jump = body->executeBlock(frame);
	switch (jump) { 
	  case JMP_NEXT:
	  case JMP_CONTINUE:
	    break;
	  case JMP_BREAK:
	    return JMP_NEXT;	    
	  default:
	    return jump;
	}	
    } while (cond->testCondition(frame));

    return JMP_NEXT;	    
}


void CtkWhileStmt::compile(CtkCompiler* compiler)
{
    cond->compile(compiler);
    compiler->compileBlock(body);
}

JumpType CtkWhileStmt::execute(CtkFrame* frame)
{
    while (cond->testCondition(frame)) { 
	JumpType jump = body->executeBlock(frame);
	switch (jump) { 
	  case JMP_NEXT:
	  case JMP_CONTINUE:
	    break;
	  case JMP_BREAK:
	    return JMP_NEXT;	    
	  default:
	    return jump;
	}	
    }
    return JMP_NEXT;	    
}



void CtkSwitchStmt::compile(CtkCompiler* compiler)
{
    expr->compile(compiler);
    for (CtkCaseStmt* cnode = cases; cnode != NULL; cnode = cnode->next) { 
	if (cnode->selector != NULL) { 
	    cnode->selector->compile(compiler);
	}
	compiler->compileBlock(cnode->stmt);
    }
}

JumpType CtkSwitchStmt::execute(CtkFrame* frame)
{
    CtkCaseStmt* selectedNode = NULL;
    CtkObject value = expr->getRvalue(frame);
    for (CtkCaseStmt* cnode = cases; cnode != NULL; cnode = cnode->next) { 
	if (cnode->selector != NULL) { 
	    CtkObject selector = cnode->selector->getRvalue(frame);
	    if (IS_EQUAL(selector, value)) { 
		POP(frame->thread);
		selectedNode = cnode;
		break;
	    }
	    POP(frame->thread);
	} else { 
	    selectedNode = cnode;
	}
    }
    POP(frame->thread);
    while (selectedNode != NULL) { 
	JumpType jump = selectedNode->stmt->executeBlock(frame);
	switch (jump) { 
	  case JMP_BREAK:
	    return JMP_NEXT;
	  case JMP_NEXT:
	    selectedNode = selectedNode->next;
	    continue;
	  default:
	    return jump;
	}
    }
    return JMP_NEXT;
}


JumpType CtkBreakStmt::execute(CtkFrame* frame)
{
    return JMP_BREAK;
}


JumpType CtkContinueStmt::execute(CtkFrame* frame)
{
    return JMP_CONTINUE;
}

void CtkReturnStmt::compile(CtkCompiler* compiler)
{
    if (expr != NULL) { 
	expr->compile(compiler);
    }
}

JumpType CtkReturnStmt::execute(CtkFrame* frame)
{    
    if (expr != NULL) { 
	expr->getRvalue(frame);
    } else { 
	PUSH_CHECK_CS(frame->thread, ctkNull);
    }
    return JMP_RETURN;
}



void CtkThrowStmt::compile(CtkCompiler* compiler)
{
    except->compile(compiler);
}

JumpType CtkThrowStmt::execute(CtkFrame* frame)
{
    CtkThread* currThread = frame->thread;
    frame->call = this;
    currThread->exceptionObject = except->getRvalue(frame);
    POP(currThread);
    return JMP_EXCEPTION;
}

void CtkTryStmt::compile(CtkCompiler* compiler)
{
    compiler->compileBlock(stmt);
    CtkSymbol* symbols = compiler->currCtx->symbols;
    symIdx = compiler->currCtx->nSymbols++;
    CtkSymbol sym(param, symIdx, symbols); 
    compiler->currCtx->symbols = &sym;
    compiler->compileBlock(handler);
    compiler->currCtx->symbols = symbols;
}

JumpType CtkTryStmt::execute(CtkFrame* frame)
{
    CtkThread* currThread = frame->thread;
    int sp = currThread->sp;
    CtkFrame* saveTrace = currThread->exceptionTrace;
#ifdef USE_CTALK_EXCEPTION
    try {
#else
    jmp_buf buf;
    jmp_buf* saveBuf = currThread->catchCtx;
    currThread->catchCtx = &buf;
    if (setjmp(buf) == 0) { 
#endif
	JumpType jump = stmt->executeBlock(frame);
#ifndef USE_CTALK_EXCEPTION
	currThread->catchCtx = saveBuf;
#endif
	currThread->sp = sp;
	switch (jump) { 
	  case JMP_EXCEPTION:
	    currThread->exceptionTrace = frame->up;
	    frame->up = NULL;
            frame->vars[symIdx] = currThread->exceptionObject;
	    jump = handler->executeBlock(frame);
	    currThread->exceptionTrace = saveTrace;
	    return jump;
	  case JMP_NEXT:
	    return JMP_NEXT;
	  default:
	    return jump;
	}
#ifdef USE_CTALK_EXCEPTION
    } catch (CtkException const& x) {
#else
    } else { 
	currThread->catchCtx = saveBuf;
#endif
        frame->vars[symIdx] = currThread->exceptionObject;
	currThread->exceptionTrace = frame->up;
	frame->up = NULL;
	currThread->sp = sp;
	JumpType jump = handler->executeBlock(frame);
	currThread->exceptionTrace = saveTrace;
	return jump;
    }
}
    
void CtkExprStmt::compile(CtkCompiler* compiler)
{
    expr->compile(compiler);
}

JumpType CtkExprStmt::execute(CtkFrame* frame)
{
    expr->getRvalue(frame);
    POP(frame->thread);
    return JMP_NEXT;
}

void CtkSynchronizedStmt::compile(CtkCompiler* compiler)
{
    mutex->compile(compiler);
    compiler->compileBlock(stmt);
}

JumpType CtkSynchronizedStmt::execute(CtkFrame* frame)
{
    CtkObject cs = mutex->getRvalue(frame);
    if (cs.type != CTK_MUTEX) { 
	THROW_EXCEPTION("synchronized parameter should be mutex");
    }
#ifdef USE_CTALK_EXCEPTION
    CtkCriticalSection lock((CtkMutex*)cs.u.ptr);
    return stmt->executeBlock(frame);
#else
    ctkLockMutex((CtkMutex*)cs.u.ptr);
    CtkThread* currThread = frame->thread;
    jmp_buf buf;
    jmp_buf* saveCtx = currThread->catchCtx;
    currThread->catchCtx = &buf;
    JumpType jump;
    if (setjmp(buf) == 0) { 
	jump = stmt->executeBlock(frame);
	ctkUnlockMutex((CtkMutex*)cs.u.ptr);
	POP(currThread);
	currThread->catchCtx = saveCtx;
    } else { 
	jump = stmt->executeBlock(frame);
	ctkUnlockMutex((CtkMutex*)cs.u.ptr);
	POP(currThread);
	currThread->catchCtx = saveCtx;	
	longjmp(*saveCtx, 1);
    }
    return jump;
#endif
}


void CtkBinaryExpr::compile(CtkCompiler* compiler, CompileCtx)
{
    left->compile(compiler);
    right->compile(compiler);
}


//////////////////////////////////////////////////////////
// Add by Jiye Zeng to apply operators to numeric array.
//

// The primitive modifies lvalue and return null in case of error
#define SET_OP2_EXPR(name) { \
	CtkPrimitive *prim = ctkGetPrimitive((name)); \
	if (prim != NULL) { \
		CtkObject params[2]; \
		params[0] = lvalue; \
		params[1] = rvalue; \
		if (!IS_TRUE(prim->func(2, params))) ctkThrowException(" : SET_OP2_EXPR error. "); \
		left->setLvalue(frame, lvalue); \
		TOP(frame->thread).u.ivalue = lvalue.u.ivalue; \
	} \
	else ctkThrowException(" : Operator primitive not found. "); \
}

// The primitive modifies result and return null in case of error
#define PRE_OP_EXPR(name) { \
	CtkPrimitive *prim = ctkGetPrimitive((name)); \
	if (prim != NULL) { \
		CtkObject param[1]; \
		param[0] = result; \
		if (IS_NULL(prim->func(1, param))) ctkThrowException(" : PRE_OP_EXPR error. "); \
		expr->setLvalue(frame, result); \
		PUSH_CHECK_CS(frame->thread, result); \
	} \
	else ctkThrowException(" : Operator primitive not found. "); \
}

CtkObject ctkCallOp2Func(CtkFrame* frame, CtkObject &lvalue, CtkObject &rvalue, const char* funcname)
{
	CtkPrimitive *prim = ctkGetPrimitive(funcname);
	if (prim != NULL) {
		CtkObject params[2];
		params[0] = lvalue;
		params[1] = rvalue;
		CtkObject result = prim->func(2, params);
		CtkMemoryManager::instance.beginAccess(frame->thread);
		POP2(frame->thread);
		PUSH(frame->thread, result);
		CtkMemoryManager::instance.endAccess(frame->thread);
		return result;
	}
	ctkThrowException(" : Operator primitive not found. ");
	return ctkNull;
}
/////////////////////////////////////////////////////////////////


CtkObject CtkAddExpr::getRvalue(CtkFrame* frame)
{
    CtkObject lvalue = left->getRvalue(frame);
    CtkObject rvalue = right->getRvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE || rvalue.type == CTK_USER_TYPE) {
	    return ctkCallOp2Func(frame, lvalue, rvalue, "AddExpr");
	}
    CtkObject result;
    if (lvalue.type == CTK_ARRAY && rvalue.type == CTK_ARRAY) { 
	CtkMemoryManager::instance.beginAccess(frame->thread);
	CtkArray* arr = ctkCloneArray((CtkArray*)lvalue.u.ptr);
	ctkAddArray(arr, (CtkArray*)rvalue.u.ptr);
	MAKE_ARRAY(result, arr);
	POP2(frame->thread);
	PUSH(frame->thread, result);
	CtkMemoryManager::instance.endAccess(frame->thread);
	return result;
    } else if (lvalue.type == CTK_STRING || rvalue.type == CTK_STRING) { 
	char buf[64];
	if (lvalue.type == CTK_INTEGER) { 
	    sprintf(buf, "%ld", (long)lvalue.u.ivalue);
	    lvalue.u.svalue = buf;
	} else if (lvalue.type == CTK_REAL) { 
	    sprintf(buf, "%f", lvalue.u.rvalue);
	    lvalue.u.svalue = buf;
	} else if (lvalue.type != CTK_STRING) { 
	    THROW_EXCEPTION("Invalid left operand for + operator");
	}
	if (rvalue.type == CTK_INTEGER) { 
	    sprintf(buf, "%ld", (long)rvalue.u.ivalue);
	    rvalue.u.svalue = buf;
	} else if (rvalue.type == CTK_REAL) { 
	    sprintf(buf, "%f", rvalue.u.rvalue);
	    rvalue.u.svalue = buf;
	} else if (rvalue.type != CTK_STRING) { 
	    THROW_EXCEPTION("Invalid right operand for + operator");
	}
	int headLen = strlen(lvalue.u.svalue);
	int tailLen = strlen(rvalue.u.svalue);
	CtkMemoryManager::instance.beginAccess(frame->thread);
	char* p = (char*)ctkAllocateObject(headLen + tailLen + 1, NULL);
	memcpy(p, lvalue.u.svalue, headLen);
	memcpy(p + headLen, rvalue.u.svalue, tailLen+1);
	MAKE_STRING(result, p);
	POP2(frame->thread);
	PUSH(frame->thread, result);
	CtkMemoryManager::instance.endAccess(frame->thread);
	return result;
    } else if (lvalue.type == CTK_REAL && rvalue.type == CTK_REAL) { 
	POP(frame->thread);
	TOP(frame->thread).u.rvalue = lvalue.u.rvalue += rvalue.u.rvalue;	
    } else if (lvalue.type == CTK_REAL && rvalue.type == CTK_INTEGER) { 
	POP(frame->thread);
	TOP(frame->thread).u.rvalue = lvalue.u.rvalue += rvalue.u.ivalue;	
    } else if (rvalue.type == CTK_REAL && lvalue.type == CTK_INTEGER) { 
	POP(frame->thread);
	lvalue.u.rvalue = lvalue.u.ivalue + rvalue.u.rvalue;	
	lvalue.type = CTK_REAL;
	TOP(frame->thread).u.rvalue = lvalue.u.rvalue;	
     } else if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_INTEGER) { 
	POP(frame->thread);
	TOP(frame->thread).u.ivalue = lvalue.u.ivalue += rvalue.u.ivalue;	
    } else { 
	THROW_EXCEPTION("Bad operands for + operator");
    } 
    return lvalue;
}


CtkObject CtkSubExpr::getRvalue(CtkFrame* frame)
{
    CtkObject lvalue = left->getRvalue(frame);
    CtkObject rvalue = right->getRvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE || rvalue.type == CTK_USER_TYPE) {
		return ctkCallOp2Func(frame, lvalue, rvalue, "SubExpr");
	}
    if (lvalue.type == CTK_ARRAY) { 
	CtkMemoryManager::instance.beginAccess(frame->thread);
	CtkArray* arr = ctkCloneArray((CtkArray*)lvalue.u.ptr);
	if (rvalue.type == CTK_ARRAY) { 
	    ctkSubArray(arr, (CtkArray*)rvalue.u.ptr);
	} else { 
	    ctkDelArray(arr, rvalue);
	}
	CtkObject result;
	MAKE_ARRAY(result, arr);
	POP2(frame->thread);
	PUSH(frame->thread, result);
	CtkMemoryManager::instance.endAccess(frame->thread);
	return result;
    }
    if (lvalue.type == CTK_REAL && rvalue.type == CTK_REAL) { 
	POP(frame->thread);
	TOP(frame->thread).u.rvalue = lvalue.u.rvalue -= rvalue.u.rvalue;	
    } else if (lvalue.type == CTK_REAL && rvalue.type == CTK_INTEGER) { 
	POP(frame->thread);
	TOP(frame->thread).u.rvalue = lvalue.u.rvalue -= rvalue.u.ivalue;	
    } else if (rvalue.type == CTK_REAL && lvalue.type == CTK_INTEGER) { 
	POP(frame->thread);
	lvalue.u.rvalue = lvalue.u.ivalue - rvalue.u.rvalue;	
	lvalue.type = CTK_REAL;
	TOP(frame->thread).u.rvalue = lvalue.u.rvalue;	
    } else if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_INTEGER) { 
	POP(frame->thread);
	TOP(frame->thread).u.ivalue = lvalue.u.ivalue -= rvalue.u.ivalue;
    } else { 
	THROW_EXCEPTION("Bad operands for - operator");
    } 
    return lvalue;
}

CtkObject CtkMulExpr::getRvalue(CtkFrame* frame)
{
    CtkObject lvalue = left->getRvalue(frame);
    CtkObject rvalue = right->getRvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE || rvalue.type == CTK_USER_TYPE) {
		return ctkCallOp2Func(frame, lvalue, rvalue, "MulExpr");
	}
    if (lvalue.type == CTK_REAL && rvalue.type == CTK_REAL) { 
	POP(frame->thread);
	TOP(frame->thread).u.rvalue = lvalue.u.rvalue *= rvalue.u.rvalue;	
    } else if (lvalue.type == CTK_REAL && rvalue.type == CTK_INTEGER) { 
	POP(frame->thread);
	TOP(frame->thread).u.rvalue = lvalue.u.rvalue *= rvalue.u.ivalue;	
    } else if (rvalue.type == CTK_REAL && lvalue.type == CTK_INTEGER) { 
	POP(frame->thread);
	lvalue.u.rvalue = lvalue.u.ivalue * rvalue.u.rvalue;	
	lvalue.type = CTK_REAL;
	TOP(frame->thread).u.rvalue = lvalue.u.rvalue;	
    } else if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_INTEGER) { 
	POP(frame->thread);
	TOP(frame->thread).u.ivalue = lvalue.u.ivalue *= rvalue.u.ivalue;
    } else { 
	THROW_EXCEPTION("Bad operands for * operator");
    } 
    return lvalue;
}

CtkObject CtkDivExpr::getRvalue(CtkFrame* frame)
{
    CtkObject lvalue = left->getRvalue(frame);
    CtkObject rvalue = right->getRvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE || rvalue.type == CTK_USER_TYPE) {
		return ctkCallOp2Func(frame, lvalue, rvalue, "DivExpr");
	}
    if (rvalue.type == CTK_INTEGER) { 
	if (rvalue.u.ivalue == 0) { 
	    THROW_EXCEPTION("Divided by zero");
	} 
    } else if (rvalue.type == CTK_REAL) { 
	if (rvalue.u.rvalue == 0.0) { 
	    THROW_EXCEPTION("Divided by zero");
	} 
    }	
    if (lvalue.type == CTK_REAL && rvalue.type == CTK_REAL) { 
	POP(frame->thread);
	TOP(frame->thread).u.rvalue = lvalue.u.rvalue /= rvalue.u.rvalue;	
    } else if (lvalue.type == CTK_REAL && rvalue.type == CTK_INTEGER) { 
	POP(frame->thread);
	TOP(frame->thread).u.rvalue = lvalue.u.rvalue /= rvalue.u.ivalue;	
    } else if (rvalue.type == CTK_REAL && lvalue.type == CTK_INTEGER) { 
	POP(frame->thread);
	lvalue.u.rvalue = lvalue.u.ivalue / rvalue.u.rvalue;	
	lvalue.type = CTK_REAL;
	TOP(frame->thread).u.rvalue = lvalue.u.rvalue;	
    } else if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_INTEGER) { 
	POP(frame->thread);
	TOP(frame->thread).u.ivalue = lvalue.u.ivalue /= rvalue.u.ivalue;
    } else { 
	THROW_EXCEPTION("Bad operands for / operator");
    } 
    return lvalue;
}

CtkObject CtkModExpr::getRvalue(CtkFrame* frame)
{
    CtkObject lvalue = left->getRvalue(frame);
    CtkObject rvalue = right->getRvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE || rvalue.type == CTK_USER_TYPE) {
		return ctkCallOp2Func(frame, lvalue, rvalue, "ModExpr");
	}
    if (rvalue.type == CTK_INTEGER) { 
	if (rvalue.u.ivalue == 0) { 
	    THROW_EXCEPTION("Divided by zero");
	} 
    } else if (rvalue.type == CTK_REAL) { 
	if (rvalue.u.rvalue == 0.0) { 
	    THROW_EXCEPTION("Divided by zero");
	} 
    }	
    if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_INTEGER) { 
	POP(frame->thread);
	TOP(frame->thread).u.ivalue = lvalue.u.ivalue %= rvalue.u.ivalue;
    } else { 
	THROW_EXCEPTION("Bad operands for % operator");
    } 
    return lvalue;
}

CtkObject CtkOrExpr::getRvalue(CtkFrame* frame)
{
    CtkObject lvalue = left->getRvalue(frame);
    CtkObject rvalue = right->getRvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE || rvalue.type == CTK_USER_TYPE) {
		return ctkCallOp2Func(frame, lvalue, rvalue, "OrExpr");
	}
    if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_INTEGER) { 
	POP(frame->thread);
	TOP(frame->thread).u.ivalue = lvalue.u.ivalue |= rvalue.u.ivalue;
    } else { 
	THROW_EXCEPTION("Bad operands for | operator");
    } 
    return lvalue;
}


CtkObject CtkAndExpr::getRvalue(CtkFrame* frame)
{
    CtkObject lvalue = left->getRvalue(frame);
    CtkObject rvalue = right->getRvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE || rvalue.type == CTK_USER_TYPE) {
		return ctkCallOp2Func(frame, lvalue, rvalue, "AndExpr");
	}
    if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_INTEGER) { 
	POP(frame->thread);
	TOP(frame->thread).u.ivalue = lvalue.u.ivalue &= rvalue.u.ivalue;
    } else { 
	THROW_EXCEPTION("Bad operands for & operator");
    } 
    return lvalue;
}


CtkObject CtkXorExpr::getRvalue(CtkFrame* frame)
{
    CtkObject lvalue = left->getRvalue(frame);
    CtkObject rvalue = right->getRvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE || rvalue.type == CTK_USER_TYPE) {
		return ctkCallOp2Func(frame, lvalue, rvalue, "XorExpr");
	}
    if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_INTEGER) { 
	POP(frame->thread);
	TOP(frame->thread).u.ivalue = lvalue.u.ivalue ^= rvalue.u.ivalue;
    } else { 
	THROW_EXCEPTION("Bad operands for ^ operator");
    } 
    return lvalue;
}


CtkObject CtkShlExpr::getRvalue(CtkFrame* frame)
{
    CtkObject lvalue = left->getRvalue(frame);
    CtkObject rvalue = right->getRvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE || rvalue.type == CTK_USER_TYPE) {
		return ctkCallOp2Func(frame, lvalue, rvalue, "ShlExpr");
	}
    if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_INTEGER) { 
	POP(frame->thread);
	TOP(frame->thread).u.ivalue = lvalue.u.ivalue <<= rvalue.u.ivalue & (sizeof(ctk_integer)*8-1);
    } else { 
	THROW_EXCEPTION("Bad operands for << operator");
    } 
    return lvalue;
}


CtkObject CtkShrExpr::getRvalue(CtkFrame* frame)
{
    CtkObject lvalue = left->getRvalue(frame);
    CtkObject rvalue = right->getRvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE || rvalue.type == CTK_USER_TYPE) {
		return ctkCallOp2Func(frame, lvalue, rvalue, "ShrExpr");
	}
    if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_INTEGER) { 
	POP(frame->thread);
	TOP(frame->thread).u.ivalue = lvalue.u.ivalue = (ctk_unsigned_integer)lvalue.u.ivalue >> (rvalue.u.ivalue & (sizeof(ctk_integer)*8-1));
    } else { 
	THROW_EXCEPTION("Bad operands for >> operator");
    } 
    return lvalue;
}

CtkObject CtkEqExpr::getRvalue(CtkFrame* frame)
{
    CtkObject lvalue = left->getRvalue(frame);
    CtkObject rvalue = right->getRvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE || rvalue.type == CTK_USER_TYPE) {
		return ctkCallOp2Func(frame, lvalue, rvalue, "EqExpr");
	}
    CtkObject result;
    MAKE_INTEGER(result, IS_EQUAL(lvalue, rvalue));
    POP2(frame->thread);
    PUSH_CS(frame->thread, result);
    return result;
}

CtkObject CtkNeExpr::getRvalue(CtkFrame* frame)
{
    CtkObject lvalue = left->getRvalue(frame);
    CtkObject rvalue = right->getRvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE || rvalue.type == CTK_USER_TYPE) {
		return ctkCallOp2Func(frame, lvalue, rvalue, "NeExpr");
	}
    CtkObject result;
    MAKE_INTEGER(result, !IS_EQUAL(lvalue, rvalue));
    POP2(frame->thread);
    PUSH_CS(frame->thread, result);
    return result;
}

CtkObject CtkGtExpr::getRvalue(CtkFrame* frame)
{
    CtkObject lvalue = left->getRvalue(frame);
    CtkObject rvalue = right->getRvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE || rvalue.type == CTK_USER_TYPE) {
		return ctkCallOp2Func(frame, lvalue, rvalue, "GtExpr");
	}
    int result = 0;
    if (lvalue.type == CTK_REAL && rvalue.type == CTK_REAL) { 
	result = lvalue.u.rvalue > rvalue.u.rvalue;	
    } else if (lvalue.type == CTK_REAL && rvalue.type == CTK_INTEGER) { 
	result = lvalue.u.rvalue > rvalue.u.ivalue;	
    } else if (rvalue.type == CTK_REAL && lvalue.type == CTK_INTEGER) { 
	result = lvalue.u.ivalue > rvalue.u.rvalue;	
    } else if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_INTEGER) { 
	result = lvalue.u.ivalue > rvalue.u.ivalue;	
    } else if (lvalue.type == CTK_STRING && rvalue.type == CTK_STRING) { 
	result = strcmp(lvalue.u.svalue, rvalue.u.svalue) > 0;
    } else { 
	THROW_EXCEPTION("Bad operands for > operator");
    } 	
    MAKE_INTEGER(lvalue, result);
    POP2(frame->thread);
    PUSH_CS(frame->thread, lvalue);
    return lvalue;
}

CtkObject CtkGeExpr::getRvalue(CtkFrame* frame)
{
    CtkObject lvalue = left->getRvalue(frame);
    CtkObject rvalue = right->getRvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE || rvalue.type == CTK_USER_TYPE) {
		return ctkCallOp2Func(frame, lvalue, rvalue, "GeExpr");
	}
    int result = 0;
    if (lvalue.type == CTK_REAL && rvalue.type == CTK_REAL) { 
	result = lvalue.u.rvalue >= rvalue.u.rvalue;	
    } else if (lvalue.type == CTK_REAL && rvalue.type == CTK_INTEGER) { 
	result = lvalue.u.rvalue >= rvalue.u.ivalue;	
    } else if (rvalue.type == CTK_REAL && lvalue.type == CTK_INTEGER) { 
	result = lvalue.u.ivalue >= rvalue.u.rvalue;	
    } else if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_INTEGER) { 
	result = lvalue.u.ivalue >= rvalue.u.ivalue;	
    } else if (lvalue.type == CTK_STRING && rvalue.type == CTK_STRING) { 
	result = strcmp(lvalue.u.svalue, rvalue.u.svalue) >= 0;
    } else { 
	THROW_EXCEPTION("Bad operands for >= operator");
    } 	
    MAKE_INTEGER(lvalue, result);
    POP2(frame->thread);
    PUSH_CS(frame->thread, lvalue);
    return lvalue;
}

CtkObject CtkLtExpr::getRvalue(CtkFrame* frame)
{
    CtkObject lvalue = left->getRvalue(frame);
    CtkObject rvalue = right->getRvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE || rvalue.type == CTK_USER_TYPE) {
		return ctkCallOp2Func(frame, lvalue, rvalue, "LtExpr");
	}
    int result = 0;
    if (lvalue.type == CTK_REAL && rvalue.type == CTK_REAL) { 
	result = lvalue.u.rvalue < rvalue.u.rvalue;	
    } else if (lvalue.type == CTK_REAL && rvalue.type == CTK_INTEGER) { 
	result = lvalue.u.rvalue < rvalue.u.ivalue;	
    } else if (rvalue.type == CTK_REAL && lvalue.type == CTK_INTEGER) { 
	result = lvalue.u.ivalue < rvalue.u.rvalue;	
    } else if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_INTEGER) { 
	result = lvalue.u.ivalue < rvalue.u.ivalue;	
    } else if (lvalue.type == CTK_STRING && rvalue.type == CTK_STRING) { 
	result = strcmp(lvalue.u.svalue, rvalue.u.svalue) < 0;
    } else { 
	THROW_EXCEPTION("Bad operands for < operator");
    } 	
    MAKE_INTEGER(lvalue, result);
    POP2(frame->thread);
    PUSH_CS(frame->thread, lvalue);
    return lvalue;
}

CtkObject CtkLeExpr::getRvalue(CtkFrame* frame)
{
    CtkObject lvalue = left->getRvalue(frame);
    CtkObject rvalue = right->getRvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE || rvalue.type == CTK_USER_TYPE) {
		return ctkCallOp2Func(frame, lvalue, rvalue, "LeExpr");
	}
    int result = 0;
    if (lvalue.type == CTK_REAL && rvalue.type == CTK_REAL) { 
	result = lvalue.u.rvalue <= rvalue.u.rvalue;	
    } else if (lvalue.type == CTK_REAL && rvalue.type == CTK_INTEGER) { 
	result = lvalue.u.rvalue <= rvalue.u.ivalue;	
    } else if (rvalue.type == CTK_REAL && lvalue.type == CTK_INTEGER) { 
	result = lvalue.u.ivalue <= rvalue.u.rvalue;	
    } else if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_INTEGER) { 
	result = lvalue.u.ivalue <= rvalue.u.ivalue;	
    } else if (lvalue.type == CTK_STRING && rvalue.type == CTK_STRING) { 
	result = strcmp(lvalue.u.svalue, rvalue.u.svalue) <= 0;
    } else { 
	THROW_EXCEPTION("Bad operands for <= operator");
    } 	
    MAKE_INTEGER(lvalue, result);
    POP2(frame->thread);
    PUSH_CS(frame->thread, lvalue);
    return lvalue;
}

CtkObject CtkLandExpr::getRvalue(CtkFrame* frame)
{
    CtkObject result;
    if (left->testCondition(frame)) { 
	MAKE_INTEGER(result, right->testCondition(frame));
    } else { 
	MAKE_INTEGER(result, 0);
    }
    PUSH_CS(frame->thread, result);
    return result;
}

CtkObject CtkLorExpr::getRvalue(CtkFrame* frame)
{
    CtkObject result;
    if (left->testCondition(frame)) { 
	MAKE_INTEGER(result, 1);
    } else { 
	MAKE_INTEGER(result, right->testCondition(frame));
    }
    PUSH_CS(frame->thread, result);
    return result;
}

void CtkAssignExpr::compile(CtkCompiler* compiler, CompileCtx) {
    left->compile(compiler, CC_LVALUE);
    right->compile(compiler, CC_RVALUE);
    if (!left->isLvalue()) { 
	compiler->error("Left operand should be L-Value", left);
    }
}

CtkObject CtkAssignExpr::getRvalue(CtkFrame* frame)
{
    CtkObject value = right->getRvalue(frame);    
    left->assign(frame, value);
    return value;
}

CtkObject CtkSetAddExpr::getRvalue(CtkFrame* frame)
{
    CtkObject rvalue = right->getRvalue(frame);
    CtkObject lvalue = left->getLvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE) {
		SET_OP2_EXPR("SetAddExpr");
		return lvalue;
	}
    if (lvalue.type == CTK_ARRAY && rvalue.type == CTK_ARRAY) { 
	CtkMemoryManager::instance.beginAccess(frame->thread); 
	ctkAddArray((CtkArray*)lvalue.u.ptr, (CtkArray*)rvalue.u.ptr);
	POP4(frame->thread);
	PUSH(frame->thread, lvalue);
	CtkMemoryManager::instance.endAccess(frame->thread); 
    } else if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_INTEGER) {
	lvalue.u.ivalue += rvalue.u.ivalue;
	left->setLvalue(frame, lvalue);
	TOP(frame->thread).u.ivalue = lvalue.u.ivalue;
    } else if (lvalue.type == CTK_REAL && rvalue.type == CTK_INTEGER) {
	lvalue.u.rvalue += rvalue.u.ivalue;
	left->setLvalue(frame, lvalue);
	TOP(frame->thread).u.rvalue = lvalue.u.rvalue;
	TOP(frame->thread).type = CTK_REAL;
    } else if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_REAL) {
	lvalue.u.ivalue = (ctk_integer)(lvalue.u.ivalue + rvalue.u.rvalue);
	left->setLvalue(frame, lvalue);
	TOP(frame->thread).u.ivalue = lvalue.u.ivalue;
	TOP(frame->thread).type = CTK_INTEGER;
    } else if (lvalue.type == CTK_REAL && rvalue.type == CTK_REAL) {
	lvalue.u.rvalue += rvalue.u.rvalue;
	left->setLvalue(frame, lvalue);
	TOP(frame->thread).u.rvalue = lvalue.u.rvalue;
    } else if (lvalue.type == CTK_STRING && rvalue.type == CTK_STRING) {
	CtkMemoryManager::instance.beginAccess(frame->thread); 
	int headLen = strlen(lvalue.u.svalue);
	int tailLen = strlen(rvalue.u.svalue);
	char* p = (char*)ctkAllocateObject(headLen + tailLen + 1, NULL);
	memcpy(p, lvalue.u.svalue, headLen);
	memcpy(p + headLen, rvalue.u.svalue, tailLen+1);
	lvalue.u.svalue = p;
	left->setLvalue(frame, lvalue);
	POP(frame->thread);
	PUSH(frame->thread, lvalue);
	CtkMemoryManager::instance.endAccess(frame->thread); 
    } else { 
	THROW_EXCEPTION("Bad operands for += operator");
    }
    return lvalue;
}

CtkObject CtkSetSubExpr::getRvalue(CtkFrame* frame)
{
    CtkObject rvalue = right->getRvalue(frame);
    CtkObject lvalue = left->getLvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE) {
		SET_OP2_EXPR("SetSubExpr")
		return lvalue;
	}
    if (lvalue.type == CTK_ARRAY) { 
	CtkMemoryManager::instance.beginAccess(frame->thread); 
	if (rvalue.type == CTK_ARRAY) { 
	    ctkSubArray((CtkArray*)lvalue.u.ptr, (CtkArray*)rvalue.u.ptr);
	} else { 
	    ctkDelArray((CtkArray*)lvalue.u.ptr, rvalue);
	}
	POP4(frame->thread);
	PUSH(frame->thread, lvalue);
	CtkMemoryManager::instance.endAccess(frame->thread); 
    } else if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_INTEGER) {
	lvalue.u.ivalue -= rvalue.u.ivalue;
	left->setLvalue(frame, lvalue);
	TOP(frame->thread).u.ivalue = lvalue.u.ivalue;
    } else if (lvalue.type == CTK_REAL && rvalue.type == CTK_INTEGER) {
	lvalue.u.rvalue -= rvalue.u.ivalue;
	left->setLvalue(frame, lvalue);
	TOP(frame->thread).u.rvalue = lvalue.u.rvalue;
	TOP(frame->thread).type = CTK_REAL;
    } else if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_REAL) {
	lvalue.u.ivalue = (ctk_integer)(lvalue.u.ivalue - rvalue.u.rvalue);
	left->setLvalue(frame, lvalue);
	TOP(frame->thread).u.ivalue = lvalue.u.ivalue;
	TOP(frame->thread).type = CTK_INTEGER;
    } else if (lvalue.type == CTK_REAL && rvalue.type == CTK_REAL) {
	lvalue.u.rvalue -= rvalue.u.rvalue;
	left->setLvalue(frame, lvalue);
	TOP(frame->thread).u.rvalue = lvalue.u.rvalue;
    } else { 
	THROW_EXCEPTION("Bad operands for -= operator");
    }
    return lvalue;
}

CtkObject CtkSetMulExpr::getRvalue(CtkFrame* frame)
{
    CtkObject rvalue = right->getRvalue(frame);
    CtkObject lvalue = left->getLvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE) {
		SET_OP2_EXPR("SetMulExpr");
		return lvalue;
	}
    if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_INTEGER) {
	lvalue.u.ivalue *= rvalue.u.ivalue;
	left->setLvalue(frame, lvalue);
	TOP(frame->thread).u.ivalue = lvalue.u.ivalue;
    } else if (lvalue.type == CTK_REAL && rvalue.type == CTK_INTEGER) {
	lvalue.u.rvalue *= rvalue.u.ivalue;
	left->setLvalue(frame, lvalue);
	TOP(frame->thread).u.rvalue = lvalue.u.rvalue;
	TOP(frame->thread).type = CTK_REAL;
    } else if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_REAL) {
	lvalue.u.ivalue = (ctk_integer)(lvalue.u.ivalue * rvalue.u.rvalue);
	left->setLvalue(frame, lvalue);
	TOP(frame->thread).u.ivalue = lvalue.u.ivalue;
	TOP(frame->thread).type = CTK_INTEGER;
    } else if (lvalue.type == CTK_REAL && rvalue.type == CTK_REAL) {
	lvalue.u.rvalue *= rvalue.u.rvalue;
	left->setLvalue(frame, lvalue);
	TOP(frame->thread).u.rvalue = lvalue.u.rvalue;
    } else { 
	THROW_EXCEPTION("Bad operands for *= operator");
    }
    return lvalue;
}

CtkObject CtkSetDivExpr::getRvalue(CtkFrame* frame)
{
    CtkObject rvalue = right->getRvalue(frame);
    CtkObject lvalue = left->getLvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE) {
		SET_OP2_EXPR("SetDivExpr");
		return lvalue;
	}
    if (rvalue.type == CTK_INTEGER) {
	if (rvalue.u.ivalue == 0) { 
	    THROW_EXCEPTION("Divided by zero");
	}
	if (lvalue.type == CTK_INTEGER) { 
	    lvalue.u.ivalue /= rvalue.u.ivalue;
	    left->setLvalue(frame, lvalue);
	    TOP(frame->thread).u.ivalue = lvalue.u.ivalue;
	} else if (lvalue.type == CTK_REAL) { 
	    lvalue.u.rvalue /= rvalue.u.ivalue;
	    left->setLvalue(frame, lvalue);
	    TOP(frame->thread).u.rvalue = lvalue.u.rvalue;
	    TOP(frame->thread).type = CTK_REAL;
	} else {
	    THROW_EXCEPTION("Bad operands for /= operator");
	}
    } else if (rvalue.type == CTK_REAL) {
	if (rvalue.u.rvalue == 0.0) { 
	    THROW_EXCEPTION("Divided by zero");
	}
	if (lvalue.type == CTK_INTEGER) { 
	    lvalue.u.ivalue = (ctk_integer)(lvalue.u.ivalue / rvalue.u.rvalue);
	    left->setLvalue(frame, lvalue);
	    TOP(frame->thread).u.ivalue = lvalue.u.ivalue;
	    TOP(frame->thread).type = CTK_INTEGER;
	} else if (lvalue.type == CTK_REAL) { 
	    lvalue.u.rvalue /= rvalue.u.rvalue;
	    left->setLvalue(frame, lvalue);
	    TOP(frame->thread).u.rvalue = lvalue.u.rvalue;
	} else {
	    THROW_EXCEPTION("Bad operands for /= operator");
	}
    } else {
	THROW_EXCEPTION("Bad operands for /= operator");
    }
    return lvalue;
}

CtkObject CtkSetModExpr::getRvalue(CtkFrame* frame)
{
    CtkObject rvalue = right->getRvalue(frame);
    CtkObject lvalue = left->getLvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE) {
		SET_OP2_EXPR("SetModExpr");
		return lvalue;
	}
    if (rvalue.type == CTK_INTEGER && lvalue.type == CTK_INTEGER) {
	if (rvalue.u.ivalue == 0) { 
	    THROW_EXCEPTION("Divided by zero");
	}
	lvalue.u.ivalue %= rvalue.u.ivalue;
	left->setLvalue(frame, lvalue);
	TOP(frame->thread).u.ivalue = lvalue.u.ivalue;
    } else {
	THROW_EXCEPTION("Bad operands for %= operator");
    }
    return lvalue;
}

CtkObject CtkSetOrExpr::getRvalue(CtkFrame* frame)
{
    CtkObject rvalue = right->getRvalue(frame);
    CtkObject lvalue = left->getLvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE) {
		SET_OP2_EXPR("SetOrExpr");
		return lvalue;
	}
    if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_INTEGER) {
	lvalue.u.ivalue |= rvalue.u.ivalue;
    } else { 
	THROW_EXCEPTION("Bad operands for |= operator");
    }
    left->setLvalue(frame, lvalue);
    TOP(frame->thread).u.ivalue = lvalue.u.ivalue;
    return lvalue;
}

CtkObject CtkSetXorExpr::getRvalue(CtkFrame* frame)
{
    CtkObject rvalue = right->getRvalue(frame);
    CtkObject lvalue = left->getLvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE) {
		SET_OP2_EXPR("SetXorExpr");
		return lvalue;
	}
    if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_INTEGER) {
	lvalue.u.ivalue ^= rvalue.u.ivalue;
    } else { 
	THROW_EXCEPTION("Bad operands for ^= operator");
    }
    left->setLvalue(frame, lvalue);
    TOP(frame->thread).u.ivalue = lvalue.u.ivalue;
    return lvalue;
}

CtkObject CtkSetAndExpr::getRvalue(CtkFrame* frame)
{
    CtkObject rvalue = right->getRvalue(frame);
    CtkObject lvalue = left->getLvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE) {
		SET_OP2_EXPR("SetAndExpr");
		return lvalue;
	}
    if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_INTEGER) {
	lvalue.u.ivalue &= rvalue.u.ivalue;
    } else { 
	THROW_EXCEPTION("Bad operands for &= operator");
    } 
    left->setLvalue(frame, lvalue);
    TOP(frame->thread).u.ivalue = lvalue.u.ivalue;
    return lvalue;
}

CtkObject CtkSetShlExpr::getRvalue(CtkFrame* frame)
{
    CtkObject rvalue = right->getRvalue(frame);
    CtkObject lvalue = left->getLvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE) {
		SET_OP2_EXPR("SetShlExpr");
		return lvalue;
	}
    if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_INTEGER) {
	lvalue.u.ivalue <<= rvalue.u.ivalue & (sizeof(ctk_integer)*8-1);
    } else { 
	THROW_EXCEPTION("Bad operands for <<= operator");
    }
    left->setLvalue(frame, lvalue);
    TOP(frame->thread).u.ivalue = lvalue.u.ivalue;
    return lvalue;
}

CtkObject CtkSetShrExpr::getRvalue(CtkFrame* frame)
{
    CtkObject rvalue = right->getRvalue(frame);
    CtkObject lvalue = left->getLvalue(frame);
	// Added by Jiye Zeng
	if (lvalue.type == CTK_USER_TYPE) {
		SET_OP2_EXPR("SetShrExpr");
		return lvalue;
	}
    if (lvalue.type == CTK_INTEGER && rvalue.type == CTK_INTEGER) {
	lvalue.u.ivalue = (ctk_unsigned_integer)lvalue.u.ivalue >> (rvalue.u.ivalue & (sizeof(ctk_integer)*8-1));
    } else { 
	THROW_EXCEPTION("Bad operands for >>= operator");
    }
    left->setLvalue(frame, lvalue);
    TOP(frame->thread).u.ivalue = lvalue.u.ivalue;
    return lvalue;
}



void CtkCondExpr::compile(CtkCompiler* compiler, CompileCtx)
{
    cond->compile(compiler);
    trueExpr->compile(compiler);
    falseExpr->compile(compiler);
}

CtkObject CtkCondExpr::getRvalue(CtkFrame* frame)
{
    CtkObject result = cond->getRvalue(frame);
    if (IS_TRUE(result)) { 
	POP(frame->thread);
	result = trueExpr->getRvalue(frame);
    } else { 
	POP(frame->thread);
	result = falseExpr->getRvalue(frame);
    }
    return result;
}


void CtkUnaryExpr::compile(CtkCompiler* compiler, CompileCtx)
{
    expr->compile(compiler);
}

CtkObject CtkNegExpr::getRvalue(CtkFrame* frame)
{
    CtkObject result = expr->getRvalue(frame);
	// Added by Jiye Zeng
	if (result.type == CTK_USER_TYPE) {
		CtkPrimitive *prim = ctkGetPrimitive("NegExpr");
		if (prim != NULL) {
			CtkObject param[1];
			param[0] = result;
			result = prim->func(1, param);
			TOP(frame->thread).u.ptr = result.u.ptr;
		}
		else {
			ctkThrowException(" : Operator primitive not found. ");
		}
		return result;
	}
    if (result.type == CTK_INTEGER) { 
	TOP(frame->thread).u.ivalue = result.u.ivalue = -result.u.ivalue;
    } else if (result.type == CTK_REAL) { 
	TOP(frame->thread).u.rvalue = result.u.rvalue = -result.u.rvalue;
    } else { 
	THROW_EXCEPTION("Bad operand for unary minus operator");
    } 
    return result;
}

CtkObject CtkNotExpr::getRvalue(CtkFrame* frame)
{
    CtkObject result = expr->getRvalue(frame);
    if (result.type != CTK_INTEGER) { 
	THROW_EXCEPTION("Bad operand for ! operator");
    } 
    TOP(frame->thread).u.ivalue = result.u.ivalue = !result.u.ivalue;
    return result;
}

CtkObject CtkComExpr::getRvalue(CtkFrame* frame)
{
    CtkObject result = expr->getRvalue(frame);
    if (result.type != CTK_INTEGER) { 
	THROW_EXCEPTION("Bad operand for ~ operator");
    } 
    TOP(frame->thread).u.ivalue = result.u.ivalue = ~result.u.ivalue;
    return result;
}

void CtkPreIncExpr::compile(CtkCompiler* compiler, CompileCtx ctx)
{
    CtkUnaryExpr::compile(compiler, ctx);
    if (!expr->isLvalue()) { 
	compiler->error("Lvalue expected for ++ operator", expr);
    }
}

CtkObject CtkPreIncExpr::getRvalue(CtkFrame* frame)
{
    CtkObject result = expr->getLvalue(frame);
	// Added by Jiye Zeng
	if (result.type == CTK_USER_TYPE) {
		PRE_OP_EXPR("PreIncExpr");
		return result;
	}
    switch (result.type) { 
      case CTK_INTEGER: 
	result.u.ivalue += 1;
	expr->setLvalue(frame, result);
	break;
      case CTK_REAL:
	result.u.rvalue += 1.0;
	expr->setLvalue(frame, result);
	break;
      default:
	THROW_EXCEPTION("Bad operand for ++ operator");
    } 
    PUSH_CHECK_CS(frame->thread, result);
    return result;
}
    

void CtkPreDecExpr::compile(CtkCompiler* compiler, CompileCtx ctx)
{
    CtkUnaryExpr::compile(compiler, ctx);
    if (!expr->isLvalue()) { 
	compiler->error("Lvalue expected for -- operator", expr);
    }
}

CtkObject CtkPreDecExpr::getRvalue(CtkFrame* frame)
{
    CtkObject result = expr->getLvalue(frame);
	// Added by Jiye Zeng
	if (result.type == CTK_USER_TYPE) {
		PRE_OP_EXPR("PreDecExpr");
		return result;
	}
    switch (result.type) { 
      case CTK_INTEGER: 
	result.u.ivalue -= 1;
	expr->setLvalue(frame, result);
	break;
      case CTK_REAL:
	result.u.rvalue -= 1.0;
	expr->setLvalue(frame, result);
	break;
      default:
	THROW_EXCEPTION("Bad operand for ++ operator");
    } 
    PUSH_CHECK_CS(frame->thread, result);
    return result;
}
    

void CtkPostIncExpr::compile(CtkCompiler* compiler, CompileCtx ctx)
{
    CtkUnaryExpr::compile(compiler, ctx);
    if (!expr->isLvalue()) { 
	compiler->error("Lvalue expected for ++ operator", expr);
    }
}

CtkObject CtkPostIncExpr::getRvalue(CtkFrame* frame)
{
    CtkObject result = expr->getLvalue(frame);
	// Added by Jiye Zeng
	if (result.type == CTK_USER_TYPE) {
		PRE_OP_EXPR("PreIncExpr");
		return result;
	}
    CtkObject newResult = result;
    switch (result.type) { 
      case CTK_INTEGER:
	newResult.u.ivalue += 1;
	expr->setLvalue(frame, newResult);
	break;
      case CTK_REAL:
	newResult.u.rvalue += 1.0;
	expr->setLvalue(frame, newResult);
	break;
      default:
	THROW_EXCEPTION("Bad operand for ++ operator");
    }
    PUSH_CHECK_CS(frame->thread, result);
    return result;
}
    



void CtkPostDecExpr::compile(CtkCompiler* compiler, CompileCtx ctx)
{
    CtkUnaryExpr::compile(compiler, ctx);
    if (!expr->isLvalue()) { 
	compiler->error("Lvalue expected for -- operator", expr);
    }
}


CtkObject CtkPostDecExpr::getRvalue(CtkFrame* frame)
{
    CtkObject result = expr->getLvalue(frame);
	// Added by Jiye Zeng
	if (result.type == CTK_USER_TYPE) {
		PRE_OP_EXPR("PreDecExpr");
		return result;
	}
    CtkObject newResult = result;
    switch (result.type) { 
      case CTK_INTEGER:
	newResult.u.ivalue -= 1;
	expr->setLvalue(frame, newResult);
	break;
      case CTK_REAL:
	newResult.u.rvalue -= 1.0;
	expr->setLvalue(frame, newResult);
	break;
      default:
	THROW_EXCEPTION("Bad operand for ++ operator");
    }
    PUSH_CHECK_CS(frame->thread, result);
    return result;
}


CtkObject CtkEnvExpr::getRvalue(CtkFrame* frame)
{
    CtkObject name = expr->getRvalue(frame);
    char* value = getenv(TO_STRING(name));
    CtkObject result;
    if (value == NULL) { 
	result = ctkNull;
	POP(frame->thread);
	PUSH_CS(frame->thread, result);
    } else {
	CtkMemoryManager::instance.beginAccess(frame->thread);
	result.u.svalue = (char*)ctkAllocateObject(strlen(value)+1, NULL);
	TOP(frame->thread).u.svalue = result.u.svalue;
	strcpy(result.u.svalue, value);
	result.type = CTK_STRING;
	CtkMemoryManager::instance.endAccess(frame->thread);
    }
    return result;
}

CtkObject CtkEnvExpr::getLvalue(CtkFrame* frame) 
{
    CtkObject base = expr->getRvalue(frame);
    char* value = getenv(TO_STRING(base));
    CtkMemoryManager::instance.beginAccess(frame->thread);
    PUSH_CHECK(frame->thread, ctkNull);
    if (value == NULL) { 
	PUSH_CHECK(frame->thread, ctkNull);
	CtkMemoryManager::instance.endAccess(frame->thread);
	return ctkNull;
    }
    CtkObject result;
    result.u.svalue = (char*)ctkAllocateObject(strlen(value)+1, NULL);
    strcpy(result.u.svalue, value);
    result.type = CTK_STRING;
    PUSH_CHECK(frame->thread, result);
    CtkMemoryManager::instance.endAccess(frame->thread);
    return result;
}

void CtkEnvExpr::assign(CtkFrame* frame, CtkObject value)
{
    CtkObject name = expr->getRvalue(frame);
    char* nam = TO_STRING(name);
    char* val = TO_STRING(value);
    char* buf = (char*)alloca(strlen(nam) + strlen(val) + 2);
    sprintf(buf, "%s=%s", nam, val);
    putenv(buf);
    POP(frame->thread);
}

void CtkEnvExpr::setLvalue(CtkFrame* frame, CtkObject value)
{
    POP2(frame->thread);
    char* nam = TO_STRING(TOP(frame->thread));
    char* val = TO_STRING(value);
    char* buf = (char*)alloca(strlen(nam) + strlen(val) + 2);
    sprintf(buf, "%s=%s", nam, val);
    putenv(buf);
    POP(frame->thread);
}

bool CtkEnvExpr::isLvalue() { 
    return true;
}

void CtkArrayAccessExpr::compile(CtkCompiler* compiler, CompileCtx) 
{ 
    array->compile(compiler);
    index->compile(compiler);
}

CtkObject CtkArrayAccessExpr::getRvalue(CtkFrame* frame)
{
    CtkObject base = array->getRvalue(frame);
    CtkObject key = index->getRvalue(frame);
	// Jiye Zeng
	if (base.type == CTK_USER_TYPE && key.type == CTK_INTEGER) {
		CtkPrimitive *prim = ctkGetPrimitive("GetAtExpr");
		if (!prim) ctkThrowException(" : Operator primitive not found. ");
		CtkObject params[2];
		params[0] = base;
		params[1] = key;
		CtkObject result = prim->func(2, params);
		POP2(frame->thread);
		PUSH_CS(frame->thread, result);
		return result;
	}
    if (base.type != CTK_ARRAY) { 
	THROW_EXCEPTION("Applying index to non-array type");
    }
    CtkObject result = ctkGetArray((CtkArray*)base.u.ptr, key);
    POP2(frame->thread);
    PUSH_CS(frame->thread, result);
    return result;
}

CtkObject CtkArrayAccessExpr::getLvalue(CtkFrame* frame) 
{
    CtkObject base = array->getRvalue(frame);
    CtkObject key = index->getRvalue(frame);
	/*/ Jiye Zeng
	if (base.type == CTK_USER_TYPE && key.type == CTK_INTEGER) {
		printf("key: %d\n", key.u.ivalue);
		CtkObject result;
		CtkArray *arr = ctkCreateArray();
		MAKE_INTEGER(result, 0);
		ctkPutArray(arr, result, base);
		MAKE_INTEGER(result, 1);
		ctkPutArray(arr, result, base);
	    CtkMemoryManager::instance.beginAccess(frame->thread);
		PUSH_CHECK(frame->thread, result);
		CtkMemoryManager::instance.endAccess(frame->thread);
		return result;
	}*/
    if (base.type != CTK_ARRAY) { 
	THROW_EXCEPTION("Applying index to non-array type");
    }
    CtkMemoryManager::instance.beginAccess(frame->thread);
    CtkObject result = ctkGetArray((CtkArray*)base.u.ptr, key);
    PUSH_CHECK(frame->thread, result);
    CtkMemoryManager::instance.endAccess(frame->thread);
    return result;
}


void CtkArrayAccessExpr::assign(CtkFrame* frame, CtkObject value)
{
    CtkObject base = array->getRvalue(frame);
    CtkObject key = index->getRvalue(frame);
	// Jiye Zeng
	// getRvalue and assign for a[i] = ...
	if (base.type == CTK_USER_TYPE && key.type == CTK_INTEGER) {
		CtkPrimitive *prim = ctkGetPrimitive("SetAtExpr");
		if (!prim) ctkThrowException(" : Operator primitive not found. ");
		CtkObject params[3];
		params[0] = base;
		params[1] = key;
		params[2] = value;
	    CtkMemoryManager::instance.beginAccess(frame->thread);
		prim->func(3, params);
	    POP2(frame->thread);
	    CtkMemoryManager::instance.endAccess(frame->thread);
		return;
	}
    if (base.type != CTK_ARRAY) { 
	THROW_EXCEPTION("Applying index to non-array type");
    }
    CtkMemoryManager::instance.beginAccess(frame->thread); 
    ctkPutArray((CtkArray*)base.u.ptr, key, value);
    POP2(frame->thread);
    CtkMemoryManager::instance.endAccess(frame->thread); 
}

void CtkArrayAccessExpr::setLvalue(CtkFrame* frame, CtkObject value)
{
	/*/ Jiye Zeng
	// getLvalue and setLvalue for a[i] += ...
	if (value.type == CTK_USER_TYPE) {
		CtkPrimitive *prim = ctkGetPrimitive("SetLeftExpr");
		if (!prim) ctkThrowException(" : Operator primitive not found. ");
		CtkObject params[1];
		params[0] = value;
	    CtkMemoryManager::instance.beginAccess(frame->thread);
		prim->func(1, params);
	    POP1(frame->thread);
	    CtkMemoryManager::instance.endAccess(frame->thread);
		return;
	}*/
    CtkMemoryManager::instance.beginAccess(frame->thread); 
    ctkPutArray((CtkArray*)frame->thread->stack[frame->thread->sp-3].u.ptr, 
		frame->thread->stack[frame->thread->sp-2], value);
    frame->thread->sp -= 3;
    CtkMemoryManager::instance.endAccess(frame->thread); 
}

bool CtkArrayAccessExpr::isLvalue() { 
    return true;
}

void CtkFuncCallExpr::compile(CtkCompiler* compiler, CompileCtx)
{
    func->compile(compiler);
    int n = 0;
    for (CtkExpr* arg = args; arg != NULL; arg = arg->next, n++) { 
	arg->compile(compiler);
    }
    CtkFuncDef* fdef = func->getFuncDef();
    if (fdef != NULL && fdef->nParams != n) {  
	char buf[256];
	sprintf(buf, "Too %s parameters for function %s", 
		fdef->nParams < n ? "many" : "few", fdef->tok->name);	
	compiler->error(buf, func);
    }
    nArgs = n;
}

static CtkObject ctkDllFunctionTrampoline(CtkPrimitive* prim, int nArgs, CtkObject* args)
{
    if (nArgs >= MAX_DLL_FUNC_ARGS) { 
	char buf[256];
	sprintf(buf, "Attempt to pass %d arguments to DLL function %s (limit for maximal number of arguments is %d)", 
		nArgs, prim->name, MAX_DLL_FUNC_ARGS);	
	ctkThrowException(buf);
    }
    int profile = 1 << nArgs;
    for (int i = 0; i < nArgs; i++) { 
        switch (args[i].type) { 
          case CTK_INTEGER:
            profile |= 1 << i;
            break;
          case CTK_REAL:
            ctkThrowException("Real type of parameter is not supported for transparent DLL invocations");
        }
    }
    void* rc = ctkTrampoline[profile]((void*)prim->func, args);
    CtkObject result;
    MAKE_RAW_POINTER(result, rc);
    return result;                    
}

CtkObject CtkFuncCallExpr::getRvalue(CtkFrame* frame)
{
    CtkObject f = func->getRvalue(frame);
    if (f.type == CTK_STRING) { 
	CtkPrimitive* prim = ctkGetPrimitive(f.u.svalue);
	if (prim != NULL) { 
	    MAKE_PRIMITIVE(f, prim);
	} else { 
	    CtkFunction* func = ctkGetFunction(f.u.svalue);
	    if (func != NULL) { 
		MAKE_FUNCTION(f, func);
	    } else {
                void* dllFunc = ctkFindFunction(f.u.svalue);
                if (dllFunc != NULL) { 
                    prim = new CtkPrimitive(f.u.svalue, (CtkPrimitiveFunc)dllFunc, false, &ctkDllFunctionTrampoline);
                    MAKE_PRIMITIVE(f, prim);
                } else { 
                    char buf[256];
                    sprintf(buf, "Function %s not found", f.u.svalue);
                    THROW_EXCEPTION(buf);
                }
	    }
	} 
    }
    frame->call = func;
    if (f.type == CTK_PRIMITIVE) { 
	CtkPrimitive* p = (CtkPrimitive*)f.u.ptr;
	CtkObject* params = (CtkObject*)alloca(nArgs*sizeof(CtkObject));
	int n = 0;
	for (CtkExpr* arg = args; arg != NULL; arg = arg->next) { 
	    params[n++] = arg->getRvalue(frame);	    
	}
	CtkMemoryManager::instance.beginAccess(frame->thread);
	frame->thread->sp -= n + 1;
	frame->thread->currFrame = frame;
	CtkObject result = p->trampoline(p, n, params);
	PUSH_CHECK(frame->thread, result);
	CtkMemoryManager::instance.endAccess(frame->thread);
	return result;
    }
    if (f.type != CTK_FUNCTION) { 
	THROW_EXCEPTION("Attempt to call non-function object");
    }
    CtkFunction* fp = (CtkFunction*)f.u.ptr;
    if (fp->nParams != nArgs) { 
	THROW_EXCEPTION("Number of actual and formal parameters doesn't match");
    }
    CtkMemoryManager::instance.beginAccess(frame->thread); 
    CtkFrame* callFrame = ctkCreateFrame(fp);
    CtkFrame* up = frame->up;
    frame->up = callFrame;
    callFrame->thread = frame->thread;
    int n = 0;
    for (CtkExpr* arg = args; arg != NULL; arg = arg->next) { 
	CtkObject value = arg->getRvalue(frame);
	callFrame->vars[n++] = value;
	POP(frame->thread);
    }
    CtkStmt* stmts = fp->stmts;
    POP(frame->thread);
    CtkMemoryManager::instance.endAccess(frame->thread); 
    switch (stmts->executeBlock(callFrame)) { 
      case JMP_RETURN:
	frame->up = up;
	return TOP(frame->thread);
      case JMP_BREAK:
	THROW_EXCEPTION("Invalid use of BREAK statement");
      case JMP_CONTINUE:
	THROW_EXCEPTION("Invalid use of CONTINUE statement");
      case JMP_NEXT:
	frame->up = up;
	PUSH_CHECK_CS(frame->thread, ctkNull);
	return ctkNull;
      case JMP_EXCEPTION:
	ctkPropagateException();
    }
    return ctkNull;
}

CtkObject CtkRunThreadExpr::getRvalue(CtkFrame* frame)
{
    CtkObject f = func->getRvalue(frame);
    if (f.type != CTK_FUNCTION) { 
	THROW_EXCEPTION("Attemp to call non-function object");
    }
    CtkFunction* fp = (CtkFunction*)f.u.ptr;
    if (fp->nParams != nArgs) { 
	THROW_EXCEPTION("Number of actual and formal parameters doesn't match");
    }
    CtkFrame* callFrame = ctkCreateFrame(fp);
    frame->call = func;
    int n = 0;
    CtkMemoryManager::instance.beginAccess(frame->thread);
    for (CtkExpr* arg = args; arg != NULL; arg = arg->next) { 
	callFrame->vars[n++] = arg->getRvalue(frame);
    }
    CtkThread* thr = ctkCreateThread(ctkRunThread, callFrame, CTK_MIN_STACK_SIZE);
    CtkObject result;
    MAKE_THREAD(result, thr);
    frame->thread->sp -= n + 1;
    PUSH_CHECK(frame->thread, result);
    CtkMemoryManager::instance.endAccess(frame->thread);
    return result;
}

CtkObject CtkIntConst::getRvalue(CtkFrame* frame)
{
    PUSH_CHECK_CS(frame->thread, value);
    return value;
}

CtkObject CtkRealConst::getRvalue(CtkFrame* frame)
{
    PUSH_CHECK_CS(frame->thread, value);
    return value;
}

CtkObject CtkNullConst::getRvalue(CtkFrame* frame)
{
    PUSH_CHECK_CS(frame->thread, ctkNull);
    return ctkNull;
}

CtkObject CtkStringConst::getRvalue(CtkFrame* frame)
{
    PUSH_CHECK_CS(frame->thread, value);
    return value;
}

void CtkArrayElemExpr::compile(CtkCompiler* compiler, CompileCtx)
{
    key->compile(compiler);
    value->compile(compiler);
}
	
void CtkArrayConst::compile(CtkCompiler* compiler, CompileCtx)
{
    int i = 0;
    for (CtkArrayElemExpr* elem = elements; elem != NULL; i++, elem = (CtkArrayElemExpr*)elem->next) {        
        if (elem->key == NULL) { 
            elem->key = new CtkIntConst(i);
        }
	elem->compile(compiler, CC_RVALUE);
    }
}
 
CtkObject CtkArrayConst::getRvalue(CtkFrame* frame)
{
    CtkMemoryManager::instance.beginAccess(frame->thread);
    CtkArray* array = ctkCreateArray();
    for (CtkArrayElemExpr* elem = elements; elem != NULL; elem = (CtkArrayElemExpr*)elem->next) {
	CtkObject key = elem->key->getRvalue(frame);
	CtkObject value = elem->value->getRvalue(frame);
	ctkPutArray(array, key, value);
	POP2(frame->thread);
    }
    CtkObject result;
    MAKE_ARRAY(result, array);
    PUSH_CHECK(frame->thread, result);
    CtkMemoryManager::instance.endAccess(frame->thread);
    return result;
}

void CtkVariable::compile(CtkCompiler* compiler, CompileCtx cc)
{
    int ctxLevel = 0;
    CtkContext* ctx;
    prim = NULL;
    func = NULL;
    for (ctx = compiler->currCtx; ctx != NULL; ctxLevel++, ctx = ctx->exterior) {
	for (CtkSymbol* symb = ctx->symbols; symb != NULL; symb = symb->next) { 
	    if (symb->tok == tok) { 
		level = ctxLevel;
		index = symb->index;
		func = symb->func;
		module = NULL;
		if (cc != CC_RVALUE && func != NULL) { 
		    compiler->error("Function is not l-value", this);
		}	    
		return;
	    }
	}
    }
    for (CtkImport* imp = compiler->importList; imp != NULL; imp = imp->next) {
	ctx = imp->module->ctx;
	for (CtkSymbol* symb = ctx->symbols; symb != NULL; symb = symb->next) { 
	    if (symb->tok == tok) { 
		index = symb->index;
		func = symb->func;
		module = imp->module->frame;
		if (cc != CC_RVALUE && func != NULL) { 
		    compiler->error("Function is not l-value", this);
		}	    
		return;
	    }
	}
    }
    // no symbol was found
    if (cc == CC_RVALUE) { 
	if ((prim = ctkGetPrimitive(tok->name)) == NULL) { 
            void* dllFunc = ctkFindFunction(tok->name);
            if (dllFunc != NULL) { 
                prim = new CtkPrimitive(tok->name, (CtkPrimitiveFunc)dllFunc, false, &ctkDllFunctionTrampoline);
            } else { 
                char buf[256];
                sprintf(buf, "Access to uninitialized variable %s", tok->name);
                compiler->error(buf, this);
            }
        }
    } else {
	level = 0;
	index = compiler->currCtx->nSymbols++;
	module = NULL;
	compiler->currCtx->symbols = new CtkSymbol(tok, index, 
						   compiler->currCtx->symbols); 
    }
}

CtkFuncDef* CtkVariable::getFuncDef() { 
    return func;
}


CtkObject CtkVariable::getRvalue(CtkFrame* frame)
{
    CtkObject result;
    if (prim != NULL) { 
	MAKE_PRIMITIVE(result, prim);
    } else if (module != NULL) { 
	result = module->vars[index];
    } else {
	CtkFrame* scope = frame;
	for (int n = level; --n >= 0; scope = scope->func->exterior); 
	result = scope->vars[index];
    }
    PUSH_CHECK_CS(frame->thread, result);
    return result;
}

CtkObject CtkVariable::getLvalue(CtkFrame* frame)
{
    PUSH_CHECK_CS(frame->thread, ctkNull);
    PUSH_CHECK_CS(frame->thread, ctkNull);
    return getRvalue(frame); 
}

void CtkVariable::assign(CtkFrame* frame, CtkObject value)
{
    CtkMemoryManager::instance.beginAccess(frame->thread); 
    if (module != NULL) { 
        module->vars[index] = value;
    } else { 
	CtkFrame* scope = frame;
	for (int n = level; --n >= 0; scope = scope->func->exterior); 
	scope->vars[index] = value;
    }    
    CtkMemoryManager::instance.endAccess(frame->thread); 	
}

bool CtkVariable::isLvalue() 
{ 
    return true;
}

void CtkVariable::setLvalue(CtkFrame* frame, CtkObject value)
{
    assign(frame, value);
    frame->thread->sp -= 3;
}


void CtkFuncDef::compile(CtkCompiler* compiler)
{
    CtkContext* exterior = compiler->currCtx;
    for (CtkContext* cpp = compiler->currCtx; cpp != NULL; cpp = cpp->exterior) {
	for (CtkSymbol* symb = cpp->symbols; symb != NULL; symb = symb->next) { 
	    if (symb->tok == tok) { 
		compiler->error("Symbol is already defined", this);
		return;
	    }
	}
    }
    index = compiler->currCtx->nSymbols++;
    CtkSymbol* symb = new CtkSymbol(tok, index, compiler->currCtx->symbols); 
    symb->func = this;
    compiler->currCtx->symbols = symb;
    CtkContext ctx;
    CtkSymbol* next = NULL;
    int n = 0;
    for (CtkTokenList* p = params; p != NULL; p = p->next) { 
	next = new CtkSymbol(p->tkn, n++, next);
    }
    ctx.nSymbols = nParams = n;
    ctx.symbols = next;
    ctx.exterior = exterior;
    compiler->currCtx = &ctx;
    compiler->currLevel += 1;
    compiler->compileBlock(stmts);
    compiler->currCtx = exterior;
    compiler->currLevel -= 1;
    nLocals = ctx.nSymbols;
}

JumpType CtkFuncDef::execute(CtkFrame* frame)
{
    CtkMemoryManager::instance.beginAccess(frame->thread); 
    CtkFunction* func = (CtkFunction*)ctkAllocateObject(sizeof(CtkFunction), NULL);
    func->stmts = stmts;
    func->nestedLevel = frame->func->nestedLevel + 1;
    func->nParams = nParams;
    func->nLocalVars = nLocals;
    func->module = frame->func->module;
    func->name = tok->name;
    func->exterior = frame;
    if (func->nestedLevel == 1) { 
	ctkAddFunction(func);
    }
    CtkObject f;
    MAKE_FUNCTION(f, func);
    frame->vars[index] = f;
    CtkMemoryManager::instance.endAccess(frame->thread); 
    return JMP_NEXT;
}











