#ifndef __COMPILER_H__
#define __COMPILER_H__

#include "ctalk.h"
#include "sync.h"
#include "memmgr.h"
#include "scanner.h"

enum CompileCtx { 
    CC_RVALUE,
    CC_LVALUE
}; 

enum JumpType { 
    JMP_NEXT, 
    JMP_RETURN, 
    JMP_BREAK, 
    JMP_CONTINUE, 
    JMP_EXCEPTION
};

struct CtkToken { 
    unsigned  hash;
    int       symId;
    int       tag;
    char*     name;
    CtkToken* next;
};

struct CtkTokenList { 
    CtkTokenList* next;
    CtkToken*     tkn;

    CtkTokenList(CtkToken* t, CtkTokenList* tail) { 
	tkn = t;
	next = tail;
    }
};
 
struct CtkExpr;
struct CtkStmt;
struct CtkFuncDef;

struct CtkSymbol { 
    CtkToken*   tok;
    CtkSymbol*  next;
    int         index;   // index of variable in the frame
    CtkFuncDef* func;

    CtkSymbol(CtkToken* t, int idx, CtkSymbol* chain) { 
	tok = t;
	index = idx;
	next = chain;
	func = NULL;
    }
};

struct CtkContext { 
    CtkContext* exterior;
    int         nSymbols;
    CtkSymbol*  symbols;

    CtkContext() { 
	exterior = NULL;
	nSymbols = 0;
	symbols = NULL;
    }
    ~CtkContext() { 
        CtkSymbol *sym, *next;
        for (sym = symbols; sym != NULL; sym = next) { 
            next = sym->next;
            delete sym;
        }
    }
};

struct CtkFrame;
struct CtkProgram;

struct CtkModule { 
    CtkToken*   tkn;
    CtkModule*  next;
    CtkFrame*   frame;
    CtkContext* ctx;
    CtkProgram* prog;
    char*       path;

    CtkModule(CtkToken* t, char const* modPath, CtkModule* chain) { 
	tkn = t;
	path = new char[strlen(modPath) + 1];
	strcpy(path, modPath);
	next = chain;
	frame = NULL;
	ctx = NULL;
    }
    ~CtkModule();
};

struct CtkImport { 
    CtkModule*  module;
    CtkImport*  next; 

    CtkImport(CtkModule* mod, CtkImport* chain) {
	module = mod;
	next = chain;
    }
};

struct CtkPath { 
    CtkPath*    next;
    char*       dir;

    CtkPath(char* path, CtkPath* chain) { 
	next = chain;
	dir = path;
    }
};
    


struct DLL_ENTRY CtkCompiler {      // Modified by Jiye Zeng
    mutex_t      mutex;
    CtkContext*  currCtx;
    int          currLevel;
    CtkProgram*  program;
    CtkImport*   importList;
    CtkModule*   modules;
    CtkPath*     path;
    CtkModule*   currModule;

    CtkModule*   compile(CtkToken* moduleName);
    CtkModule*   compile(char const* moduleName);
    
    void         compileBlock(CtkStmt* stmts);

    void         error(char const* errtext, CtkNode* node);

    void         reset();

    static CtkCompiler instance;
};

struct CtkStmt : CtkNode { 
    CtkStmt* next;

    virtual JumpType execute(CtkFrame* frame);
    virtual void compile(CtkCompiler* compiler);

    JumpType  executeBlock(CtkFrame* frame);

    CtkStmt() { 
        line = CtkScanner::instance.yyline;
        next = NULL; 
    }
    virtual ~CtkStmt() {}
};


struct CtkProgram : CtkStmt { 
    CtkTokenList* modules;
    CtkStmt*      stmts;
    int           nVars;

    CtkProgram(CtkTokenList* modules, CtkStmt* stmts) { 
	this->modules = modules;
	this->stmts = stmts;
    }
    
    ~CtkProgram() { 
        CtkStmt *stmt, *nextStmt;
        for (stmt = stmts; stmt != NULL; stmt = nextStmt) { 
            nextStmt = stmt->next;
            delete stmt;
        }
        CtkTokenList *tkn, *nextTkn;
        for (tkn = modules; tkn != NULL; tkn = nextTkn) { 
            nextTkn = tkn->next;
            delete tkn;
        }            
    }
        
    virtual JumpType execute(CtkFrame* frame);
    
    virtual void compile(CtkCompiler* compiler);
};
    
struct CtkFuncDef : CtkStmt { 
    CtkToken*     tok; 
    CtkTokenList* params; 
    CtkStmt*      stmts;
    int           index;
    int           nParams;
    int           nLocals;

    virtual void compile(CtkCompiler* compiler);
    virtual JumpType execute(CtkFrame* frame);

    CtkFuncDef(CtkToken* tok, CtkTokenList* params, CtkStmt* stmts) { 
	this->tok = tok;
	this->params = params;
	this->stmts = stmts;
    }
    ~CtkFuncDef() { 
        CtkStmt *stmt, *nextStmt;
        for (stmt = stmts; stmt != NULL; stmt = nextStmt) { 
            nextStmt = stmt->next;
            delete stmt;
        }
        CtkTokenList *tkn, *nextTkn;
        for (tkn = params; tkn != NULL; tkn = nextTkn) { 
            nextTkn = tkn->next;
            delete tkn;
        }            
    }
};

struct CtkExpr : CtkNode { 
    CtkExpr* next;

    virtual void compile(CtkCompiler* compiler, CompileCtx ctx = CC_RVALUE);

    inline  bool testCondition(CtkFrame* frame) { 
	CtkObject result = getRvalue(frame);
	frame->thread->sp -= 1;
	return IS_TRUE(result);
    }

    inline void evaluate(CtkFrame* frame) { 
	getRvalue(frame);
	frame->thread->sp -= 1;
    }

    virtual CtkObject   getRvalue(CtkFrame* frame);
    virtual CtkObject   getLvalue(CtkFrame* frame);
    virtual void        setLvalue(CtkFrame* frame, CtkObject value);
    virtual void        assign(CtkFrame* frame, CtkObject value);
    virtual bool        isLvalue();
    virtual CtkFuncDef* getFuncDef();

    CtkExpr() { 
        next = NULL; 
        line = CtkScanner::instance.yyline;
    }
    virtual ~CtkExpr() {}
}; 

struct CtkIfStmt : CtkStmt { 
    CtkExpr* cond;
    CtkStmt* thenStmt;
    CtkStmt* elseStmt;

    virtual void compile(CtkCompiler* compiler);
    virtual JumpType execute(CtkFrame* frame);

    CtkIfStmt(CtkExpr* cond, CtkStmt* thenStmt, CtkStmt* elseStmt) { 
	this->cond = cond;
	this->thenStmt = thenStmt;
	this->elseStmt = elseStmt;
    }
    ~CtkIfStmt() { 
        delete cond;
        delete thenStmt;
        delete elseStmt;
    }
};

struct CtkForStmt : CtkStmt { 
    CtkExpr* init;
    CtkExpr* cond;
    CtkExpr* inc;
    CtkStmt* body;

    virtual void compile(CtkCompiler* compiler);
    virtual JumpType execute(CtkFrame* frame);

    CtkForStmt(CtkExpr* init, CtkExpr* cond, CtkExpr* inc, CtkStmt* body) { 
	this->init = init;
	this->cond = cond;
	this->inc  = inc;
	this->body = body;
    }
    ~CtkForStmt() { 
        delete init;
        delete cond;
        delete inc;
        delete body;
    }
};

struct CtkDoWhileStmt : CtkStmt { 
    CtkExpr* cond;
    CtkStmt* body;

    virtual void compile(CtkCompiler* compiler);
    virtual JumpType execute(CtkFrame* frame);

    CtkDoWhileStmt(CtkStmt* body, CtkExpr* cond) { 
	this->cond = cond;
	this->body = body;
    }
    ~CtkDoWhileStmt() { 
        delete cond;
        delete body;
    }       
};

struct CtkWhileStmt : CtkStmt { 
    CtkExpr* cond;
    CtkStmt* body;

    virtual void compile(CtkCompiler* compiler);
    virtual JumpType execute(CtkFrame* frame);

    CtkWhileStmt(CtkExpr* cond, CtkStmt* body) { 
	this->cond = cond;
	this->body = body;
    }
    ~CtkWhileStmt() { 
        delete cond;
        delete body;
    }       
};

struct CtkCaseStmt : CtkStmt { 
    CtkCaseStmt* next;
    CtkExpr*     selector;
    CtkStmt*     stmt;

    CtkCaseStmt(CtkExpr* selector, CtkStmt* stmt) { 
	next = NULL;
	this->selector = selector;
	this->stmt = stmt;
    }
    ~CtkCaseStmt() { 
        delete selector;
        delete stmt;
    }       
};

struct CtkSwitchStmt : CtkStmt { 
    CtkExpr* expr;
    CtkCaseStmt* cases;

    virtual void compile(CtkCompiler* compiler);
    virtual JumpType execute(CtkFrame* frame);

    CtkSwitchStmt(CtkExpr* expr, CtkCaseStmt* cases) { 
	this->expr = expr;
	this->cases = cases;
    }
    ~CtkSwitchStmt() { 
        delete expr;
        CtkCaseStmt *cs, *next;
        for (cs = cases; cs != NULL; cs = next) { 
            next = cs->next;
            delete cs;
        }
    }
};

struct CtkContinueStmt : CtkStmt { 
    virtual JumpType execute(CtkFrame* frame);
};


struct CtkBreakStmt : CtkStmt { 
    virtual JumpType execute(CtkFrame* frame);
};

struct CtkReturnStmt : CtkStmt { 
    CtkExpr* expr;

    virtual void compile(CtkCompiler* compiler);
    virtual JumpType execute(CtkFrame* frame);

    CtkReturnStmt(CtkExpr* expr = NULL) { 
	this->expr = expr;
    }
    ~CtkReturnStmt() { 
        delete expr;
    }
};
    
struct CtkThrowStmt : CtkStmt { 
    CtkExpr* except;

    virtual void compile(CtkCompiler* compiler);
    virtual JumpType execute(CtkFrame* frame);

    CtkThrowStmt(CtkExpr* except = NULL) { 
	this->except = except;
    }
    ~CtkThrowStmt() { 
        delete except;
    }
};

struct CtkTryStmt : CtkStmt { 
    CtkStmt*    stmt;
    CtkToken*   param;
    CtkStmt*    handler;
    int         symIdx;

    virtual void compile(CtkCompiler* compiler);
    virtual JumpType execute(CtkFrame* frame);

    CtkTryStmt(CtkStmt* stmt, CtkToken* param, CtkStmt* handler) { 
	this->stmt = stmt;
	this->param = param;
	this->handler = handler;
    }
    ~CtkTryStmt() { 
        delete stmt;
        delete handler;
    }
};

struct CtkExprStmt : CtkStmt { 
    CtkExpr* expr;
    
    virtual void compile(CtkCompiler* compiler);
    virtual JumpType execute(CtkFrame* frame);

    CtkExprStmt(CtkExpr* expr) { 
	this->expr = expr;
    }
    ~CtkExprStmt() { 
        delete expr;
    }
};

struct CtkEmptyStmt : CtkStmt {};

struct CtkSynchronizedStmt : CtkStmt { 
    CtkExpr* mutex;
    CtkStmt* stmt;
    
    virtual void compile(CtkCompiler* compiler);
    virtual JumpType execute(CtkFrame* frame);

    CtkSynchronizedStmt(CtkExpr* mutex, CtkStmt* stmt) { 
	this->mutex = mutex;
	this->stmt = stmt;
    }
    ~CtkSynchronizedStmt() { 
        delete mutex;
        delete stmt;
    }
};

struct CtkBinaryExpr : CtkExpr {
    CtkExpr* left;
    CtkExpr* right;

    virtual void compile(CtkCompiler* compiler, CompileCtx ctx);
 
    CtkBinaryExpr(CtkExpr* left, CtkExpr* right) { 
	this->left = left;
	this->right = right;
    }
    ~CtkBinaryExpr() { 
        delete left;
        delete right;
    }
};

struct CtkAddExpr : CtkBinaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkAddExpr(CtkExpr* left, CtkExpr* right) : CtkBinaryExpr(left, right) {}
};

struct CtkSubExpr : CtkBinaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkSubExpr(CtkExpr* left, CtkExpr* right) : CtkBinaryExpr(left, right) {}
};

struct CtkMulExpr : CtkBinaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkMulExpr(CtkExpr* left, CtkExpr* right) : CtkBinaryExpr(left, right) {}
};
    
struct CtkDivExpr : CtkBinaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkDivExpr(CtkExpr* left, CtkExpr* right) : CtkBinaryExpr(left, right) {}
};

struct CtkModExpr : CtkBinaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkModExpr(CtkExpr* left, CtkExpr* right) : CtkBinaryExpr(left, right) {}
};

struct CtkOrExpr : CtkBinaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkOrExpr(CtkExpr* left, CtkExpr* right) : CtkBinaryExpr(left, right) {}
};
    
struct CtkAndExpr : CtkBinaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkAndExpr(CtkExpr* left, CtkExpr* right) : CtkBinaryExpr(left, right) {}
};

struct CtkXorExpr : CtkBinaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkXorExpr(CtkExpr* left, CtkExpr* right) : CtkBinaryExpr(left, right) {}
};


struct CtkShlExpr : CtkBinaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkShlExpr(CtkExpr* left, CtkExpr* right) : CtkBinaryExpr(left, right) {}
};
    
struct CtkShrExpr : CtkBinaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkShrExpr(CtkExpr* left, CtkExpr* right) : CtkBinaryExpr(left, right) {}
};

struct CtkEqExpr : CtkBinaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkEqExpr(CtkExpr* left, CtkExpr* right) : CtkBinaryExpr(left, right) {}
};

struct CtkNeExpr : CtkBinaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkNeExpr(CtkExpr* left, CtkExpr* right) : CtkBinaryExpr(left, right) {}
};

struct CtkGtExpr : CtkBinaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkGtExpr(CtkExpr* left, CtkExpr* right) : CtkBinaryExpr(left, right) {}
};

struct CtkGeExpr : CtkBinaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkGeExpr(CtkExpr* left, CtkExpr* right) : CtkBinaryExpr(left, right) {}
};

struct CtkLtExpr : CtkBinaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkLtExpr(CtkExpr* left, CtkExpr* right) : CtkBinaryExpr(left, right) {}
};
    
struct CtkLeExpr : CtkBinaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkLeExpr(CtkExpr* left, CtkExpr* right) : CtkBinaryExpr(left, right) {}
};


struct CtkLandExpr : CtkBinaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkLandExpr(CtkExpr* left, CtkExpr* right) : CtkBinaryExpr(left, right) {}
};


struct CtkLorExpr : CtkBinaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkLorExpr(CtkExpr* left, CtkExpr* right) : CtkBinaryExpr(left, right) {}
};

struct CtkAssignExpr : CtkBinaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);
    virtual void compile(CtkCompiler* compiler, CompileCtx);

    CtkAssignExpr(CtkExpr* left, CtkExpr* right) : CtkBinaryExpr(left, right) {}
};

struct CtkSetAddExpr : CtkAssignExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkSetAddExpr(CtkExpr* left, CtkExpr* right) : CtkAssignExpr(left, right) {}
};

struct CtkSetSubExpr : CtkAssignExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkSetSubExpr(CtkExpr* left, CtkExpr* right) : CtkAssignExpr(left, right) {}
};

struct CtkSetMulExpr : CtkAssignExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkSetMulExpr(CtkExpr* left, CtkExpr* right) : CtkAssignExpr(left, right) {}
};

struct CtkSetDivExpr : CtkAssignExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkSetDivExpr(CtkExpr* left, CtkExpr* right) : CtkAssignExpr(left, right) {}
};

struct CtkSetModExpr : CtkAssignExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkSetModExpr(CtkExpr* left, CtkExpr* right) : CtkAssignExpr(left, right) {}
};

struct CtkSetOrExpr : CtkAssignExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkSetOrExpr(CtkExpr* left, CtkExpr* right) : CtkAssignExpr(left, right) {}
};

struct CtkSetXorExpr : CtkAssignExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkSetXorExpr(CtkExpr* left, CtkExpr* right) : CtkAssignExpr(left, right) {}
};

struct CtkSetAndExpr : CtkAssignExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkSetAndExpr(CtkExpr* left, CtkExpr* right) : CtkAssignExpr(left, right) {}
};

struct CtkSetShlExpr : CtkAssignExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkSetShlExpr(CtkExpr* left, CtkExpr* right) : CtkAssignExpr(left, right) {}
};

struct CtkSetShrExpr : CtkAssignExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkSetShrExpr(CtkExpr* left, CtkExpr* right) : CtkAssignExpr(left, right) {}
};

struct CtkCondExpr : CtkExpr {
    CtkExpr* cond;
    CtkExpr* trueExpr;
    CtkExpr* falseExpr;

    virtual void compile(CtkCompiler* compiler, CompileCtx ctx);
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkCondExpr(CtkExpr* cond, CtkExpr* trueExpr, CtkExpr* falseExpr) { 
	this->cond = cond;
	this->trueExpr = trueExpr;
	this->falseExpr = falseExpr;
    }
    ~CtkCondExpr() { 
        delete cond;
        delete trueExpr;
        delete falseExpr;
    }
};

struct CtkUnaryExpr : CtkExpr { 
    CtkExpr* expr;

    virtual void compile(CtkCompiler* compiler, CompileCtx ctx);

    CtkUnaryExpr(CtkExpr* expr) { 
	this->expr = expr;
    }
    ~CtkUnaryExpr() { 
        delete expr;
    }
};

struct CtkNegExpr : CtkUnaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkNegExpr(CtkExpr* expr) : CtkUnaryExpr(expr) {}
};

struct CtkPreIncExpr : CtkUnaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);
    virtual void compile(CtkCompiler* compiler, CompileCtx ctx);

    CtkPreIncExpr(CtkExpr* expr) : CtkUnaryExpr(expr) {}
};

struct CtkPreDecExpr : CtkUnaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);
    virtual void compile(CtkCompiler* compiler, CompileCtx ctx);

    CtkPreDecExpr(CtkExpr* expr) : CtkUnaryExpr(expr) {}
};

struct CtkPostIncExpr : CtkUnaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);
    virtual void compile(CtkCompiler* compiler, CompileCtx ctx);

    CtkPostIncExpr(CtkExpr* expr) : CtkUnaryExpr(expr) {}
};

struct CtkPostDecExpr : CtkUnaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);
    virtual void compile(CtkCompiler* compiler, CompileCtx ctx);

    CtkPostDecExpr(CtkExpr* expr) : CtkUnaryExpr(expr) {}
};

struct CtkNotExpr : CtkUnaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkNotExpr(CtkExpr* expr) : CtkUnaryExpr(expr) {}
};

struct CtkComExpr : CtkUnaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkComExpr(CtkExpr* expr) : CtkUnaryExpr(expr) {}
};

struct CtkEnvExpr : CtkUnaryExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);
    virtual CtkObject getLvalue(CtkFrame* frame);
    virtual void      setLvalue(CtkFrame* frame, CtkObject value);
    virtual void      assign(CtkFrame* frame, CtkObject value);
    virtual bool      isLvalue();

    CtkEnvExpr(CtkExpr* expr) : CtkUnaryExpr(expr) {}
};

struct CtkArrayAccessExpr : CtkExpr { 
    CtkExpr* array;
    CtkExpr* index;

    virtual void      compile(CtkCompiler* compiler, CompileCtx ctx);
    virtual CtkObject getRvalue(CtkFrame* frame);
    virtual CtkObject getLvalue(CtkFrame* frame);
    virtual void      setLvalue(CtkFrame* frame, CtkObject value);
    virtual void      assign(CtkFrame* frame, CtkObject value);
    virtual bool      isLvalue();

    CtkArrayAccessExpr(CtkExpr* array, CtkExpr* index) { 
	this->array = array;
	this->index = index;
    }
    ~CtkArrayAccessExpr() { 
        delete array;
        delete index;
    }
};
    
struct CtkFuncCallExpr : CtkExpr { 
    CtkExpr* func;
    CtkExpr* args;
    int      nArgs;

    virtual void compile(CtkCompiler* compiler, CompileCtx ctx);
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkFuncCallExpr(CtkExpr* func, CtkExpr* args) { 
	this->func = func;
	this->args = args;
    }
    ~CtkFuncCallExpr() {
        delete func;
        CtkExpr* arg, *next;
        for (arg = args; arg != NULL; arg = next) { 
            next = arg->next;
            delete arg;
        }
    }
};
    
struct CtkRunThreadExpr : CtkFuncCallExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);    

    CtkRunThreadExpr(CtkExpr* func, CtkExpr* args) : CtkFuncCallExpr(func, args) {}
};
    
struct CtkArrayElemExpr : CtkExpr { 
    CtkExpr* key;
    CtkExpr* value;

    virtual void compile(CtkCompiler* compiler, CompileCtx ctx);

    CtkArrayElemExpr(CtkExpr* key, CtkExpr* value) { 
	this->key = key;
	this->value = value;
    }
    ~CtkArrayElemExpr() { 
        delete key;
        delete value;
    }
};


struct CtkArrayConst: CtkExpr { 
    CtkArrayElemExpr* elements;

    virtual void compile(CtkCompiler* compiler, CompileCtx ctx);
    virtual CtkObject getRvalue(CtkFrame* frame);

    void enumerate();

    CtkArrayConst(CtkArrayElemExpr* elements) { 
	this->elements = elements;
    }
    ~CtkArrayConst() { 
        CtkExpr* elem, *next;
        for (elem = elements; elem != NULL; elem = next) { 
            next = elem->next;
            delete elem;
        }
    }
};

struct CtkStringConst : CtkExpr { 
    CtkObject value;
    
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkStringConst(char* s) { 
        MAKE_STRING(value, s);	
    }
    ~CtkStringConst() { 
        ctkDeallocateStringLiteral(value.u.svalue);
    }
};

struct CtkIntConst : CtkExpr { 
    CtkObject value;
    
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkIntConst(ctk_integer i) { 
	MAKE_INTEGER(value, i);	
    }
};

struct CtkRealConst : CtkExpr { 
    CtkObject value;
    
    virtual CtkObject getRvalue(CtkFrame* frame);

    CtkRealConst(ctk_real r) { 
	MAKE_REAL(value, r);	
    }
};

struct CtkNullConst : CtkExpr { 
    virtual CtkObject getRvalue(CtkFrame* frame);
};

struct CtkVariable : CtkExpr { 
    CtkToken*     tok; 
    CtkFrame*     module;
    CtkPrimitive* prim;
    CtkFuncDef*   func;
    int           level;
    int           index;

    virtual void compile(CtkCompiler* compiler, CompileCtx ctx);

    virtual CtkObject   getRvalue(CtkFrame* frame);
    virtual CtkObject   getLvalue(CtkFrame* frame);
    virtual void        setLvalue(CtkFrame* frame, CtkObject value);
    virtual void        assign(CtkFrame* frame, CtkObject value);
    virtual bool        isLvalue();
    virtual CtkFuncDef* getFuncDef();

    CtkVariable(CtkToken* tok) { 
	this->tok = tok;
	module = NULL;
	prim = NULL;
    }
};

extern int yylex();

#endif




