#define INSIDE_CTALK 1

#include "compiler.h"
#include "ctkparser.h"

DLL_ENTRY CtkObject ctkNull  = {CTK_NULL, {0}};
DLL_ENTRY CtkObject ctkFalse = {CTK_INTEGER, {0}};
DLL_ENTRY CtkObject ctkTrue  = {CTK_INTEGER, {1}};

int ctkMaxStackSize = 1024;

#ifdef _WIN32
DLL_ENTRY 
void ctkLockMutex(CtkMutex* mutex) { 
    if (mutex->initialized) { 
	EnterCriticalSection(&mutex->cs);
    } else { 
	ctkThrowException("Mutex was deleted");
    }
}

DLL_ENTRY 
void ctkUnlockMutex(CtkMutex* mutex) { 
    if (mutex->initialized) { 
	LeaveCriticalSection(&mutex->cs); 
    } else { 
	ctkThrowException("Mutex was deleted");
    }
}

DLL_ENTRY 
CtkMutex* ctkCreateMutex() { 
    CtkMutex* mutex = 
	(CtkMutex*)ctkAllocateObject(sizeof(CtkMutex), CtkFinalizer(ctkDeleteMutex));
    InitializeCriticalSection(&mutex->cs);
    mutex->initialized = true;
    return mutex;
}
    
DLL_ENTRY 
void ctkDeleteMutex(CtkMutex* mutex) { 
    if (mutex->initialized) { 
	mutex->initialized = false;
	DeleteCriticalSection(&mutex->cs);
    }
} 


#else

#include <sys/time.h>

void ctkLockMutex(CtkMutex* mutex) { 
    if (mutex->initialized) { 
	pthread_t self = pthread_self();
	if (mutex->owner != self) { 
	    pthread_mutex_lock(&mutex->mutex); 
	    mutex->owner = self;
	}
	mutex->count += 1;
    } else { 
	ctkThrowException("Mutex was deleted");
    }
}

void ctkUnlockMutex(CtkMutex* mutex) { 
    if (mutex->initialized) { 
	if (pthread_self() != mutex->owner) { 
	    ctkThrowException("Not owner");
	}
	if (--mutex->count == 0) {
	    mutex->owner = 0;
	    pthread_mutex_unlock(&mutex->mutex);
	}
    } else {
	ctkThrowException("Mutex was deleted");
    }
}

CtkMutex* ctkCreateMutex() { 
    CtkMutex* mutex = (CtkMutex*)ctkAllocateObject(sizeof(CtkMutex), CtkFinalizer(ctkDeleteMutex));
    mutex->count = 0;
    mutex->owner = 0;
    mutex->initialized = true;
    pthread_mutex_init(&mutex->mutex, NULL);
    return mutex;
}
    
void ctkDeleteMutex(CtkMutex* mutex) { 
    pthread_mutex_destroy(&mutex->mutex);
} 

#endif


#define INIT_HASH_SIZE 17		// Jiye Zeng

DLL_ENTRY 
CtkArray* ctkCreateArray() { 
    CtkArray* arr = (CtkArray*)ctkAllocateObject(sizeof(CtkArray), CtkFinalizer(ctkDeleteArray));
    arr->entries = new CtkHashEntry*[INIT_HASH_SIZE];
    CHECK_ALLOC(arr->entries); 
    memset(arr->entries, 0, INIT_HASH_SIZE*sizeof(CtkHashEntry*));
    arr->nUsed = 0;
    arr->nAllocated = INIT_HASH_SIZE;
    return arr;
}

inline int hashCode(CtkObject obj) { 
    if (obj.type == CTK_STRING) { 
	int h = 0;
	byte* p = (byte*)obj.u.svalue;
	int ch;
	while ((ch = *p++) != 0) { 
	    h = h*31 + ch;
	}
	return h;
    } else { 
	return (int)obj.u.ivalue;
    }
}

DLL_ENTRY 
void ctkPutArray(CtkArray* arr, CtkObject key, CtkObject value) { 
    if (arr->entries == NULL) {
	ctkThrowException("Array was deleted");
    }
    CtkHashEntry* entry;
    unsigned hash = hashCode(key);
    int i = hash % arr->nAllocated;
    for (entry = arr->entries[i]; entry != NULL; entry = entry->next) { 
	if (entry->hashCode == hash && IS_EQUAL(key, entry->key)) { 
	    entry->value = value;
	    return;
	}
    } 
    if (++arr->nUsed == arr->nAllocated) { 
	// rehash
	int newCapacity = arr->nAllocated * 2;
	CtkHashEntry** newEntries = new CtkHashEntry*[newCapacity];
	CHECK_ALLOC(newEntries);
	memset(newEntries, 0, newCapacity*sizeof(CtkHashEntry*));
	for (int j = arr->nAllocated; --j >= 0;) { 
	    CtkHashEntry* next;
	    for (entry = arr->entries[j]; entry != NULL; entry = next) { 
		next = entry->next;
		int k = entry->hashCode % newCapacity;
		entry->next = newEntries[k];
		newEntries[k]= entry;
	    }
	}
	arr->nAllocated = newCapacity;
	delete[] arr->entries;
	arr->entries = newEntries;
	i = hash % arr->nAllocated;
    }
    entry = new CtkHashEntry();
    CHECK_ALLOC(entry); 
    entry->next = arr->entries[i];
    arr->entries[i] = entry;
    entry->value = value;
    entry->key = key;
    entry->hashCode = hash;
}

DLL_ENTRY 
int ctkDelArray(CtkArray* arr, CtkObject key)
{
    if (arr->entries == NULL) {
	ctkThrowException("Array was deleted");
    }
    CtkHashEntry **epp, *ep;
    unsigned hash = hashCode(key);
    int i = hash % arr->nAllocated;
    for (epp = &arr->entries[i]; (ep = *epp) != NULL; epp = &ep->next) { 	
	if (ep->hashCode == hash && IS_EQUAL(key, ep->key)) { 
	    *epp = ep->next;
	    delete ep;
	    arr->nUsed -= 1;
	    return 1;
	}
    }
    return 0;
}

DLL_ENTRY 
CtkObject ctkGetArray(CtkArray* arr, CtkObject key)
{
    if (arr->entries == NULL) {
	ctkThrowException("Array was deleted");
    }
    CtkHashEntry* entry;
    unsigned hash = hashCode(key);
    int i = hash % arr->nAllocated;
    for (entry = arr->entries[i]; entry != NULL; entry = entry->next) { 
	if (entry->hashCode == hash && IS_EQUAL(key, entry->key)) { 
	    return entry->value;
	}
    } 
    return ctkNull;
}

DLL_ENTRY 
void ctkClearArray(CtkArray* arr)
{
    CtkHashEntry *entry, *next;

    if (arr->entries != NULL) { 
	for (int i = arr->nAllocated; --i >= 0;) { 
	    for (entry = arr->entries[i]; entry != NULL; entry = next) { 
		next = entry->next;
		delete entry;
	    }
	    arr->entries[i] = NULL;
	}
	arr->nUsed = 0;
    }
}

DLL_ENTRY 
void ctkDeleteArray(CtkArray* arr)
{
    ctkClearArray(arr);
    delete[] arr->entries;
    arr->entries = NULL;
}

DLL_ENTRY 
CtkArray* ctkCloneArray(CtkArray* arr)
{
    CtkArray* clone = (CtkArray*)ctkAllocateObject(sizeof(CtkArray), CtkFinalizer(ctkDeleteArray));
    clone->entries = new CtkHashEntry*[arr->nAllocated];
    CHECK_ALLOC(clone->entries); 
    memset(clone->entries, 0, arr->nAllocated*sizeof(CtkHashEntry*));
    clone->nUsed = arr->nUsed;
    clone->nAllocated = arr->nAllocated;
    for (int i = arr->nAllocated; --i >= 0;) { 
	CtkHashEntry **hpp = &clone->entries[i];
	for (CtkHashEntry* entry = arr->entries[i]; entry != NULL; entry = entry->next) { 
	    CtkHashEntry *hp = new CtkHashEntry();
	    CHECK_ALLOC(hp); 
	    hp->key = entry->key;
	    hp->value = entry->value;
	    hp->hashCode = entry->hashCode;
	    *hpp = hp;
	    hpp = &hp->next;
	}
	*hpp = NULL;
    }
    return clone;
}


DLL_ENTRY 
void ctkAddArray(CtkArray* aDst, CtkArray* aSrc)
{
    for (int i = aSrc->nAllocated; --i >= 0;) { 
	for (CtkHashEntry* entry = aSrc->entries[i]; entry != NULL; entry = entry->next) { 
	    ctkPutArray(aDst, entry->key, entry->value);
	}
    }
}

DLL_ENTRY 
void ctkSubArray(CtkArray* aDst, CtkArray* aSrc)
{
    for (int i = aSrc->nAllocated; --i >= 0;) { 
	for (CtkHashEntry* entry = aSrc->entries[i]; entry != NULL; entry = entry->next) { 
	    ctkDelArray(aDst, entry->key);
	}
    }
}

DLL_ENTRY 
void ctkInitializeThread(CtkThread* thr) 
{
    thr->queue.firstItem = thr->queue.lastItem = NULL;
    thr->queue.nItems = 0;
    thr->queue.nBlocked = 0;
    thr->markFunc = NULL;
#ifdef _WIN32
    InitializeCriticalSection(&thr->queue.cs);
    thr->queue.event = CreateEvent(NULL, FALSE, FALSE, NULL);
#else
    pthread_mutex_init(&thr->queue.mutex, NULL);
    pthread_cond_init(&thr->queue.cond, NULL);
#endif   
    thr->exceptionTrace = NULL;
    thr->stackSize = ctkMaxStackSize;
    thr->sp = 0;
    thr->stack = new CtkObject[thr->stackSize];
    CHECK_ALLOC(thr->stack);
    thr->memoryAccess = 0;
    thr->mallocList = NULL;
    thr->lastHeader = NULL;
    thr->exceptionObject = ctkNull;
    thr->usedObject = ctkNull;
    thr->callStack = NULL;
    thr->catchCtx = NULL;
    thr->currFrame = NULL; 
    thr->arg = NULL;
    thr->initialized = 1;

    // Added by KDB
    thr->ctkThreadId = 0;
    CtkThreadPool::instance.addThread(thr);
}

#ifdef _WIN32
DWORD WINAPI threadProg(void* arg) 
#else
void* threadProg(void* arg) 
#endif
{
    CtkThread* t = (CtkThread*)arg; 
    CtkThreadPool::instance.setupThread(t);
    (*t->func)(t->callStack);  
    CtkThreadPool::instance.removeThread(t);
    return 0;
}

DLL_ENTRY 
CtkThread* ctkCreateThread(CtkThreadFunc f, CtkFrame* frame, size_t stackSize)
{
    CtkThread* thr = (CtkThread*)ctkAllocateObject(sizeof(CtkThread), 
						   CtkFinalizer(ctkDeleteThread));
    ctkInitializeThread(thr);
    thr->func = f;
    thr->callStack = frame;

	static int threads_count = 0;
	thr->ctkThreadId = ++threads_count;

    frame->thread = thr;
#ifdef _WIN32    
    HANDLE h = CreateThread(NULL, stackSize, threadProg, thr,
			    0, &thr->threadId);
    CloseHandle(h);
#else
    pthread_attr_t attr;
    pthread_attr_init(&attr);
    pthread_attr_setstacksize(&attr, stackSize);
    pthread_create(&thr->thread, &attr, threadProg, thr);
    pthread_detach(thr->thread);
    pthread_attr_destroy(&attr);
#endif    
    return thr;
}

void ctkRunThread(CtkFrame* frame) { 
#ifdef USE_CTALK_EXCEPTION
    try {
#else
    jmp_buf buf;
    frame->thread->catchCtx = &buf;
    if (setjmp(buf) == 0) { 
#endif
	JumpType jump = frame->func->stmts->executeBlock(frame);
	frame->thread->callStack = NULL;
	switch (jump) { 
	  case JMP_EXCEPTION:
	    ctkTrace("Uncatched exception %s\n", 
		     IF_STRING(frame->thread->exceptionObject));
	    break;
	  case JMP_BREAK:
	  case JMP_CONTINUE:
	    ctkTrace("Unexpected BREAK or CONTINUE statement at line\n"); 
	    break;
	  default:
	    return;
	}
#ifdef USE_CTALK_EXCEPTION
    } catch (CtkException const& x) {
#else
    } else { 
#endif
	ctkTrace("Uncatched exception %s\n", 
		 IF_STRING(frame->thread->exceptionObject));
    } 
    ctkPrintStackTrace(frame);
}
    
DLL_ENTRY 
int ctkGetThreadQueueLength(CtkThread* thr)
{
    return thr->queue.nItems;
}

DLL_ENTRY 
CtkObject ctkGetMessage(CtkThread* thr)
{
    if (!thr->initialized) {
	ctkThrowException("Thread was deleted");
    }
    CtkQueueItem* item;

    CtkThread* curthr = ctkGetCurrentThread();
    int i, nAccesses = curthr->memoryAccess;
    for (i = nAccesses; --i >= 0; CtkMemoryManager::instance.endAccess(curthr));
    
#ifdef _WIN32    
    EnterCriticalSection(&thr->queue.cs);
#else
    pthread_mutex_lock(&thr->queue.mutex);
#endif
    while (thr->queue.nItems == 0) { 
	thr->queue.nBlocked += 1;
#ifdef _WIN32    
	LeaveCriticalSection(&thr->queue.cs);
	WaitForSingleObject(thr->queue.event, INFINITE);
	EnterCriticalSection(&thr->queue.cs);
#else
	pthread_cond_wait(&thr->queue.cond, &thr->queue.mutex);
#endif
    }
    item = thr->queue.firstItem;
    CtkObject value = item->value;
    thr->usedObject = value;
    thr->queue.firstItem = item->next;
    if (item == thr->queue.lastItem) { 
	thr->queue.lastItem = NULL;
    }
    thr->queue.nItems -= 1;

#ifdef _WIN32
    LeaveCriticalSection(&thr->queue.cs);
#else
    pthread_mutex_unlock(&thr->queue.mutex);
#endif
    delete item;
    for (i = nAccesses; --i >= 0; CtkMemoryManager::instance.beginAccess(curthr));
    return value;
}

DLL_ENTRY 
void ctkPutMessage(CtkThread* thr, CtkObject msg)
{
    if (!thr->initialized) {
	ctkThrowException("Thread was deleted");
    }
    CtkQueueItem* item = new CtkQueueItem();
    CHECK_ALLOC(item);
    item->value = msg;
#ifdef _WIN32    
    EnterCriticalSection(&thr->queue.cs);
#else
    pthread_mutex_lock(&thr->queue.mutex);
#endif
    item->next = NULL;
    if (thr->queue.lastItem != NULL) { 
	thr->queue.lastItem->next = item;
    } else { 
	thr->queue.firstItem = item;
    }
    thr->queue.lastItem = item;
    if (thr->queue.nBlocked > 0) { 
	thr->queue.nBlocked -= 1;
#ifdef _WIN32
	SetEvent(thr->queue.event);
#else 
	pthread_cond_signal(&thr->queue.cond);
#endif
    }
    thr->queue.nItems += 1;
#ifdef _WIN32
    LeaveCriticalSection(&thr->queue.cs);
#else
    pthread_mutex_unlock(&thr->queue.mutex);
#endif    
}

DLL_ENTRY 
void ctkDeleteThread(CtkThread* thr) 
{
    if (thr->initialized) { 
	thr->initialized = false;
#ifdef _WIN32
	CloseHandle(thr->queue.event);
	DeleteCriticalSection(&thr->queue.cs);
#else
	pthread_mutex_destroy(&thr->queue.mutex);
	pthread_cond_destroy(&thr->queue.cond);
#endif
	CtkQueueItem *item, *next;
	for (item = thr->queue.firstItem; item != NULL; item = next) { 
	    next = item->next;
	    delete item;
	}
	thr->queue.nItems = 0;
    }
}

DLL_ENTRY 
CtkThread* ctkGetCurrentThread()
{
#ifdef _WIN32
    return (CtkThread*)TlsGetValue(CtkThreadPool::instance.threadKey);
#else
    return (CtkThread*)pthread_getspecific(CtkThreadPool::instance.threadKey);
#endif
}

DLL_ENTRY 
void* ctkThrowException(char const* msg)
{
    CtkThread* thr = ctkGetCurrentThread();
    CtkMemoryManager::instance.beginAccess(thr);
    ALLOCATE_STRING(thr->exceptionObject, (char*)msg);
    CtkMemoryManager::instance.endAccess(thr);
#ifdef USE_CTALK_EXCEPTION
    throw CtkException(thr->exceptionObject.u.svalue);
#else
    longjmp(*thr->catchCtx, 1);
#endif
    return NULL;
}

DLL_ENTRY 
void ctkPropagateException()
{
    CtkThread* thr = ctkGetCurrentThread();
#ifdef USE_CTALK_EXCEPTION
    throw CtkException(thr->exceptionObject.u.svalue);
#else
    longjmp(*thr->catchCtx, 1);
#endif
}

static void defaultTraceFunction(char const* msg, va_list args)
{
    vfprintf(stderr, msg, args);
}

static ctkTraceFunction traceHook = &defaultTraceFunction;

void ctkTrace(char const* msg, ...)
{
    va_list args;
    va_start(args, msg);
    (*traceHook)(msg, args);
    va_end(args);
}

DLL_ENTRY 
ctkTraceFunction ctkSetTraceFunction(ctkTraceFunction hook)
{
    ctkTraceFunction saveHook = traceHook;
    traceHook = hook;
    return saveHook;
}


#define PRIMITIVE_HASH_TABLE_SIZE 1031

CtkObject ctkDefaultPrimitiveTrampolile(CtkPrimitive* prim, int nArgs, CtkObject* args)
{
    return prim->func(nArgs, args);
}

static CtkPrimitive* primHashTable[PRIMITIVE_HASH_TABLE_SIZE];

inline int stringHashCode(char const* s)
{ 
    int h = 0;
    unsigned char* p = (unsigned char*)s;
    while (*p != 0) { 
	h = h*31 + *p++;
    }
    return h;
}

DLL_ENTRY 
void ctkAddPrimitive(CtkPrimitive* prim)
{
    static mutex_t primMutex;
    critical_section cs(primMutex);
    unsigned h = stringHashCode(prim->name);
    int i = h % PRIMITIVE_HASH_TABLE_SIZE;
    for (CtkPrimitive* pp = primHashTable[i]; pp != NULL; pp = pp->next) { 
	if (strcmp(pp->name, prim->name) == 0) { 
	    if (pp->overload) { 
		return;
	    } else if (!prim->overload) { 
		ctkTrace("Redefinition of primitive %s\n", prim->name);
	    }
	    break;
	}
    }
    prim->next = primHashTable[i];
    primHashTable[i] = prim;
}

DLL_ENTRY 
CtkPrimitive* ctkGetPrimitive(char const* name)
{
    unsigned h = stringHashCode(name);
    int i = h % PRIMITIVE_HASH_TABLE_SIZE;
    for (CtkPrimitive* pp = primHashTable[i]; pp != NULL; pp = pp->next) { 
	if (strcmp(pp->name, name) == 0) { 
	    return pp;
	}
    }
    return NULL;
}

DLL_ENTRY 
void* ctkAllocateObject(size_t size, CtkFinalizer finalizer) { 
    return CtkMemoryManager::instance.allocate(size, finalizer);
}

DLL_ENTRY 
char* ctkAllocateStringLiteral(char* str)
{
    CtkAllocHeader* hdr = (CtkAllocHeader*)malloc(strlen(str) + 1 + sizeof(CtkAllocHeader));
    hdr->finalizer = NULL;
    hdr->next = NULL;
    char* dst = (char*)(hdr + 1);
    strcpy(dst, str);
    return dst;
}

DLL_ENTRY 
void ctkDeallocateStringLiteral(char* str)
{
    free((CtkAllocHeader*)str-1);
}

DLL_ENTRY 
CtkFile* ctkCreateFile(char const* name, char const* fmt)
{
    FILE* f = fopen(name, fmt);
    if (f == NULL) {
	return NULL;
    }
    CtkFile* file = (CtkFile*)ctkAllocateObject(sizeof(CtkFile), 
						CtkFinalizer(ctkDeleteFile));
    file->name = new char[strlen(name) + 1];
    CHECK_ALLOC(file->name);
    strcpy(file->name, name);
    file->f = f;
    return file;
}

DLL_ENTRY 
void ctkDeleteFile(CtkFile* file) { 
	// Modified by Jiye Zeng
	if (file->f == stdout || file->f == stderr || file->f == stdin) return;
    if (file->f != NULL) { 
	fclose(file->f);
	delete[] file->name;
	file->f = NULL;
    }
}

DLL_ENTRY 
void ctkPrint(CtkObject obj) { 
    if (obj.type == CTK_STRING) { 
	fputs(obj.u.svalue, stdout);
    } else { 
	char buf[256];
	ctkToString(obj, buf);
	fputs(buf, stdout);
    }
}

DLL_ENTRY 
void ctkPrintFile(CtkFile* file, CtkObject obj) { 
    if (file->f == NULL) { 
	ctkThrowException("File was deleted");
    }
    if (obj.type == CTK_STRING) { 
	fputs(obj.u.svalue, file->f);
    } else { 
	char buf[256];
	ctkToString(obj, buf);
	fputs(buf, file->f);
    }
}

DLL_ENTRY 
void ctkFlushFile(CtkFile* file) { 
    if (file->f == NULL) { 
	ctkThrowException("File was deleted");
    }
    fflush(file->f);
}

#include <math.h>

DLL_ENTRY 
char* ctkToString(CtkObject obj, char* buf)
{
	// Modified by Jiye Zeng
    switch (obj.type) { 
      case CTK_NULL:
	strcpy(buf, " NULL");
	break;
      case CTK_INTEGER:
	sprintf(buf, " %ld", (long)obj.u.ivalue);
	break;
      case CTK_REAL:
	if (fabs(obj.u.rvalue) < 1.e-5 || fabs(obj.u.rvalue) > 1.e5)
		sprintf(buf, " %1.6e", obj.u.rvalue);
	else
		sprintf(buf, " %f", obj.u.rvalue);
	break;
      case CTK_PRIMITIVE:
	sprintf(buf, " <%p:PRIMITIVE %s>", obj.u.ptr, ((CtkPrimitive*)obj.u.ptr)->name);
	break;	
      case CTK_FILE:
	sprintf(buf, " <%p:FILE %s>", obj.u.ptr, ((CtkFile*)obj.u.ptr)->name);
	break;		
      case CTK_RAW_POINTER:
	sprintf(buf, " <%p:RAW POINTER>", obj.u.ptr);
	break;		
      case CTK_STRING:
	sprintf(buf, "<%p:STRING \"%.64s\">", obj.u.svalue, obj.u.svalue);
	break;	
      case CTK_ARRAY:
	sprintf(buf, " <%p:ARRAY %d items>", obj.u.ptr, ((CtkArray*)obj.u.ptr)->nUsed);
	break;	
      case CTK_MUTEX:
	sprintf(buf, " <%p:MUTEX>", obj.u.ptr);	
	break;	
      case CTK_THREAD:
	sprintf(buf, " <%p:THREAD>", obj.u.ptr);	
	break;	
      case CTK_FUNCTION:
	sprintf(buf, " <%p:FUNCTION %s>", obj.u.ptr, ((CtkFunction*)obj.u.ptr)->name);	
	break;	
      case CTK_FRAME:
	sprintf(buf, " <%p:FRAME>", obj.u.ptr);	       
	break;	
      case CTK_USER_TYPE:
	sprintf(buf, " <%p:UDT>", obj.u.ptr);	       
	break;	
      default:
	sprintf(buf, " %p:UNKNOWN type %d>", obj.u.ptr, obj.type);	  
    }
    return buf;
}


static CtkArray* options;

DLL_ENTRY 
CtkArray* ctkGetCommandLineOptions() { 
    return options;
}


DLL_ENTRY 
void ctkPrintStackTrace(CtkFrame* frame) { 
    if (frame->up != NULL) { 
	ctkPrintStackTrace(frame->up);
    }
    ctkTrace("%s from module %s", 
	     frame->func->name, frame->func->module->tkn->name);
    if (frame->call != NULL) { 
	ctkTrace(" line %d\n", frame->call->line);
    } else { 
	ctkTrace("\n");
    }
}
    
#define FUNC_HASH_TABLE_SIZE 1031;

static CtkFunction* funcHashTable[PRIMITIVE_HASH_TABLE_SIZE];
static mutex_t funcMutex;

DLL_ENTRY 
void ctkAddFunction(CtkFunction* f)
{
    critical_section cs(funcMutex);
    unsigned h = stringHashCode(f->name);
    int i = h % PRIMITIVE_HASH_TABLE_SIZE;
    f->next = funcHashTable[i];
    funcHashTable[i] = f;
}

DLL_ENTRY 
CtkFunction* ctkGetFunction(char const* name)
{
    unsigned h = stringHashCode(name);
    int i = h % PRIMITIVE_HASH_TABLE_SIZE;
    for (CtkFunction* f = funcHashTable[i]; f != NULL; f = f->next) { 
	if (strcmp(f->name, name) == 0) { 
	    return f;
	}
    }
    return NULL;
}

DLL_ENTRY 
CtkFrame* ctkCreateFrame(CtkFunction* func) 
{ 
    CtkFrame* frame = (CtkFrame*)ctkAllocateObject(sizeof(CtkFrame) 
		    + (func->nLocalVars-1)*sizeof(CtkObject), NULL);
    memset(frame->vars, 0, func->nLocalVars*sizeof(CtkObject));
    frame->nVars = func->nLocalVars;
    frame->func = func;
    frame->up = NULL;
    frame->call = NULL;
    frame->thread = NULL;
    return frame;
}

DLL_ENTRY 
CtkObject ctkCallFunction(CtkFunction* func, CtkObject* args)
{
    CtkFrame* frame = ctkCreateFrame(func);
    CtkThread* thread = (CtkThread*)ctkGetCurrentThread();
    frame->thread = thread;
    CtkFrame* saveTopFrame = thread->currFrame;
	if (saveTopFrame != NULL) { 
        saveTopFrame->up = frame;
	}
    thread->currFrame = frame;
    memcpy(frame->vars, args, func->nParams*sizeof(CtkObject));
    int sp = thread->sp;
    CtkObject result;
    switch (func->stmts->executeBlock(frame)) { 
      case JMP_RETURN:
	result = thread->stack[thread->sp-1];
	// no break
      case JMP_NEXT:
	thread->currFrame = saveTopFrame;
	thread->sp = sp;
	if (saveTopFrame != NULL) {
		saveTopFrame->up = NULL;
	}
	break;
      case JMP_BREAK:
	ctkThrowException("Invalid use of BREAK statement");
      case JMP_CONTINUE:
	ctkThrowException("Invalid use of CONTINUE statement");
      case JMP_EXCEPTION:
	ctkPropagateException();	    
    }
    return result;
}

DLL_ENTRY 
void ctkSleep(unsigned seconds) 
{
#ifdef _WIN32
    Sleep(seconds*1000);
#else
    sleep(seconds);
#endif
}

DLL_ENTRY 
void ctkNotEnoughMemory()
{
    CtkMemoryManager::instance.notEnoughMemory();
}


#ifndef _WIN32
#include <dlfcn.h>
#endif

struct CtkDllDescriptor { 
    CtkDllDescriptor*        next;
    static CtkDllDescriptor* chain;
    void*                    dll;

    CtkDllDescriptor(void* h) { 
        dll = h;
        next = chain;
        chain = this;
    }
};
CtkDllDescriptor* CtkDllDescriptor::chain;

DLL_ENTRY 
int ctkLoadLibrary(char const* name) 
{
    char buf[1024];
    char* p = strrchr(name, '.');
    bool hasExtension = p != NULL && p[1] != '.' && p[1] != '/' && p[1] != '\\';
    void* dll;
#ifdef _WIN32
    if (!hasExtension) { 
	sprintf(buf, "%s.dll", name);
	name = buf;
    }
#ifdef _UNICODE
    wchar_t wcbuf[1024];
    mbstowcs(wcbuf, name, sizeof wcbuf);
    dll = LoadLibrary(wcbuf);
#else
    dll = LoadLibrary(name);
#endif
#else
    if (!hasExtension) { 
	sprintf(buf, "%s.so", name);
	name = buf;
    }
    dll = dlopen(name, RTLD_NOW);
#endif
    if (dll != NULL) { 
        new CtkDllDescriptor(dll);
        return true;
    }
    return false;
}

DLL_ENTRY 
void* ctkFindFunction(char const* name) 
{
    for (CtkDllDescriptor* desc = CtkDllDescriptor::chain; desc != NULL; desc = desc->next) { 
        void* func;
#ifdef _WIN32
        func = GetProcAddress((HMODULE)desc->dll, name);
#else
        func = dlsym(desc->dll, name);
#endif
        if (func != NULL) { 
            return func;
        }
    }
    return NULL;
}
    
DLL_ENTRY 
long ctkGetTimeMillis()
{
#ifdef _WIN32
    return GetTickCount();
#else
    struct timeval cur_tv;
    gettimeofday(&cur_tv, NULL);
    return (cur_tv.tv_sec & 0x1FFFFF)*1000 + cur_tv.tv_usec/1000;    
#endif
}


DLL_ENTRY 
char* ctkGetModuleName(CtkFrame* frame)
{
    return (char*)frame->func->module->tkn->name;
}

DLL_ENTRY 
char* ctkGetModuleFilePath(CtkFrame* frame)
{
    return (char*)frame->func->module->path;
}

DLL_ENTRY 
int ctkEqual(CtkObject o1, CtkObject o2)
{ 
    if (o1.type == CTK_STRING && o2.type == CTK_STRING) { 
	return strcmp(o1.u.svalue, o2.u.svalue) == 0;
    } else if (o1.type == CTK_REAL) { 
	if (o2.type == CTK_REAL) { 
	    return o1.u.rvalue == o2.u.rvalue;
	} else if (o2.type == CTK_INTEGER) { 
	    return o1.u.rvalue == o2.u.ivalue;
	}
    } else if (o1.type == CTK_INTEGER) { 
	if (o2.type == CTK_REAL) { 
	    return o1.u.ivalue == o2.u.rvalue;
	} else if (o2.type == CTK_INTEGER) { 
	    return o1.u.ivalue == o2.u.ivalue;
	}
    } else if (o1.type == o2.type) {
	return o1.u.ptr == o2.u.ptr;
    }
    return false; 
}
	
DLL_ENTRY 
void ctkParseArguments(int nArgs, CtkObject* args, char const* format, ...)
{
    va_list params;
    va_start(params, format);
    bool optionalParameters = false;
    for (int i = 0; i < nArgs; i++) { 
        CtkObject& o = args[i];
        int type = o.type;
        switch (*format++) { 
          case ',':
            optionalParameters = true;
            i -= 1;
            continue;
          case 'i':
            if (type != CTK_INTEGER) { 
                ctkThrowException("Integer value expected");
            }
            *va_arg(params, ctk_integer*) = o.u.ivalue;
            continue;
          case 'I':
            if (type != CTK_INTEGER) { 
                ctkThrowException("Integer value expected");
            }
            *va_arg(params, CtkObject*) = o;
            continue;
          case 'r':
            if (type != CTK_REAL) { 
                ctkThrowException("Real value expected");
            }
            *va_arg(params, ctk_real*) = o.u.rvalue;
            continue;
          case 'R':
            if (type != CTK_REAL) { 
                ctkThrowException("Real value expected");
            }
            *va_arg(params, CtkObject*) = o;
            continue;
          case 'p':
            if (type != CTK_RAW_POINTER && type != CTK_NULL) { 
                ctkThrowException("Raw pointer expected");
            }
            *va_arg(params, void**) = o.u.ptr;
            continue;
          case 'P':
            if (type != CTK_RAW_POINTER) { 
                ctkThrowException("Raw pointer expected");
            }
            *va_arg(params, CtkObject*) = o;
            continue;
          case 'o':
            *va_arg(params, CtkObject*) = o;
            continue;
          case 'S':
            if (type != CTK_STRING) { 
                ctkThrowException("String expected");
            }
            *va_arg(params, CtkObject*) = o;
            continue;
          case 's':
            if (type != CTK_STRING && type != CTK_NULL) { 
                ctkThrowException("String expected");
            }
            *va_arg(params, char**) = o.u.svalue;
            continue;
          case 'A':
            if (type != CTK_ARRAY) { 
                ctkThrowException("Array expected");
            }
            *va_arg(params, CtkObject*) = o;
            continue;
          case 'a':
            if (type != CTK_ARRAY && type != CTK_NULL) { 
                ctkThrowException("Array expected");
            }
            *va_arg(params, CtkArray**) = (CtkArray*)o.u.ptr;
            continue;
          case 'F':
            if (type != CTK_FILE) { 
                ctkThrowException("File expected");
            }
            *va_arg(params, CtkObject*) = o;
            continue;
          case 'f':
            if (type != CTK_FILE && type != CTK_NULL) { 
                ctkThrowException("File expected");
            }
            *va_arg(params, CtkFile**) = (CtkFile*)o.u.ptr;
            continue;
          case 'M':
            if (type != CTK_MUTEX) { 
                ctkThrowException("Mutex expected");
            }
            *va_arg(params, CtkObject*) = o;
            continue;
           case 'm':
            if (type != CTK_MUTEX && type != CTK_NULL) { 
                ctkThrowException("Mutex expected");
            }
            *va_arg(params, CtkMutex**) = (CtkMutex*)o.u.ptr;
            continue;
          case 'T':
            if (type != CTK_THREAD) { 
                ctkThrowException("Thread expected");
            }
            *va_arg(params, CtkObject*) = o;
            continue;
          case 't':
            if (type != CTK_THREAD && type != CTK_NULL) { 
                ctkThrowException("Thread expected");
            }
            *va_arg(params, CtkThread**) = (CtkThread*)o.u.ptr;
            continue;
          case 'L':
            if (type != CTK_FUNCTION) { 
                ctkThrowException("Function expected");
            }
            *va_arg(params, CtkObject*) = o;
            continue;
          case 'l':
            if (type != CTK_FUNCTION && type != CTK_NULL) { 
                ctkThrowException("Function expected");
            }
            *va_arg(params, CtkFunction**) = (CtkFunction*)o.u.ptr;
            continue;
          case 'U':
            if (type != CTK_USER_TYPE) { 
                ctkThrowException("Value of user defined type expected");
            }
            *va_arg(params, CtkObject*) = o;
            continue;
          case 'u':
            if (type != CTK_USER_TYPE && type != CTK_NULL) { 
                ctkThrowException("Value of user defined type expected");
            }
            *va_arg(params, void**) = o.u.ptr;
            continue;
          case '\0':
            ctkThrowException("Too many arguments for function");
          default:
            ctkThrowException("Invalid format");
        }
    }
    if (*format != '\0' && *format != ',' && !optionalParameters) {
        ctkThrowException("Too few arguments for function");
    }
    va_end(params);
}

/*
#ifndef EMBEDDED_CTALK
int main(int argc, char* argv[])
{
    if (argc == 1) { 
	printf("C-Talk intergreter version 1.0\n");
	printf("Usage: ctalk {options} {module-name}\n"
	       "Options:\n"
	       "\t-D<var>=<value>\n"
	       "\t-p <module-path>{;<module-path>}\n\n");
	return 1;
    }
    CtkCompiler* compiler = &CtkCompiler::instance;
    compiler->path = new CtkPath(".", NULL);
    CHECK_ALLOC(compiler->path); 
    options = ctkCreateArray();
    CtkObject opt;
    MAKE_ARRAY(opt, options);
    CtkThreadPool::instance.pushVariable(opt);

    for (int i = 1; i < argc; i++) { 
	if (*argv[i] == '-') { 
	    switch (argv[i][1]) { 
	      case 'p':
	      case 'P':
		if (i+1 < argc) { 
		    char *p, *cp = argv[++i];
		    while ((p = strchr(cp, ';')) != NULL) { 
			*p = '\0';
			compiler->path = new CtkPath(cp, compiler->path);
			CHECK_ALLOC(compiler->path);
			cp = p + 1;
		    }
		    if (*cp != '\0') { 
			compiler->path = new CtkPath(cp, compiler->path);
			CHECK_ALLOC(compiler->path);
		    }
		    continue;
		}
		break;
	      case 'd':
	      case 'D':
		{
		    char* p = strchr(argv[i] + 2, '='); 
		    if (p == NULL) { 
			break;
		    } else { 
			*p = '\0';
			CtkObject key;
			CtkObject value;
			ALLOCATE_STRING(key, argv[i] + 2);
			ALLOCATE_STRING(value, p + 1);
			ctkPutArray(options, key, value);
			continue;
		    }		    
		}
	    }
	    fprintf(stderr, "Invalid option %s\n", argv[i]);
	    return 1;
	} else { 
	    if (compiler->compile(CtkSymbolTable::instance.add(argv[i], IDENT)) == NULL) { 
		fprintf(stderr, "Failed to locate module \"%s.ctk\"\n", argv[i]);
		return 1;
	    }
	}
    }
    return 0;
}
#endif
*/