#ifndef __ARRAY_TMP__
#define __ARRAY_TMP__

#include <string.h>
#include <stdio.h>

#pragma warning(once: 4018)


#define MAX(a, b) ((a) > (b) ? (a) : (b))
#define MIN(a, b) ((a) < (b) ? (a) : (b))

typedef unsigned int size_t;

enum {
	ARRAY_NULL = 'ARR',
	ARRAY_CHAR, ARRAY_UCHAR, ARRAY_SHORT, ARRAY_USHORT, ARRAY_INT, ARRAY_UINT,
	ARRAY_FLOAT, ARRAY_DOUBLE, ARRAY_COMPLEX, ARRAY_TRIPLEX, ARRAY_QUADRUPLE,
	ARRAY_TOP
};

typedef struct sComplex{double r, i;} sComplex;
typedef struct sTriplex{double x, y, z;} sTriplex;
typedef struct sQuadruple{double x, y, z, w;} sQuadruple;

class zeUint;
class zeChar;

///
/// zeArray template
///

template<class T>
class zeArrayTmp
{
protected:
	int				_type;

	size_t			_ncol;

	size_t			_nrow;

	T				*_data;

public:
	/// Constructor.
	zeArrayTmp() : _type(ARRAY_NULL), _nrow(0), _ncol(0), _data(0) { }

	/// Copy constructor.
	zeArrayTmp(const zeArrayTmp & arr) : _type(ARRAY_NULL), _nrow(0), _ncol(0), _data(0)
	{
		(*this) = arr;
	}

	/// Destructor.
	virtual ~zeArrayTmp()
	{
		if (_data) delete[] _data;
		_data = 0;
	}

	/// The infamous 'operator='.
	zeArrayTmp & operator =(const zeArrayTmp & arr)
	{
		if (&arr != this) {
			_type = arr.type();
			resize(arr.nrow(), arr.ncol());
			memcpy(_data, arr.data(), ndat()*sizeof(T));
		}
		return *this;
	}

	/// Array access
	size_t ndat() const {return _ncol * _nrow;}
	size_t ncol() const {return _ncol;}
	size_t nrow() const {return _nrow;}
	int    type() const {return _type;}

	const T * data() const {return _data;}
		  T * data()       {return _data;}

  	const T & operator()(size_t i)           const {return _data[i];}
		  T & operator()(size_t i)                 {return _data[i];}
#ifdef ROW_MAJOR
	const T & operator()(size_t i, size_t j) const {return _data[i * _ncol + j];}
		  T & operator()(size_t i, size_t j)       {return _data[i * _ncol + j];}
#else
	const T & operator()(size_t i, size_t j) const {return _data[i + _nrow * j];}
		  T & operator()(size_t i, size_t j)       {return _data[i + _nrow * j];}
#endif

	/// Reshape
	int reshape(size_t nrow, size_t ncol)
	{
		if (ndat() != ncol * nrow) return 0;
		_ncol = ncol;
		_nrow = nrow;
		return 1;
	}

	/// Reset array size.
	int resize(size_t nrow, size_t ncol, bool keep = false)
	{
   		if (nrow * ncol == 0) return 0;
		if (nrow * ncol == ndat())
		{
			_nrow = nrow;
			_ncol = ncol;
			return 1;
		}
		T *data = new T[nrow * ncol];
		if (!data) return 0;
		if (_data)
		{
			if (keep) memcpy(data, _data, nrow*ncol*sizeof(T));
			delete[] _data;
		}
		_data = data;
		_nrow = nrow;
		_ncol = ncol;
		return 1;
	}

	/// Range.
	int fill(T t0, T dt)
	{
		size_t i, n = ndat();
		for (i = 0; i < n; i++) _data[i] = t0 + dt * i;
		return 1;
	}

	/// Negate.
	int negate()
	{
		size_t i, n = ndat();
		for (i = 0; i < n; i++) _data[i] = -_data[i];
		return 1;
	}

	/// Add
	int add(T v)
	{
		size_t i, n = ndat();
		for (i = 0; i < n; i++) _data[i] += v;
		return 1;
	}

	int add(const zeArrayTmp<T> & arr)
	{
		size_t i, n = ndat();
		if (arr.ndat() != n) return 0;
		for (i = 0; i < n; i++) _data[i] += arr(i);
		return 1;
	}

	/// Subtract
	int subtract(T v)
	{
		size_t i, n = ndat();
		for (i = 0; i < n; i++) _data[i] -= v;
		return 1;
	}

	int subtract(const zeArrayTmp<T> & arr)
	{
		size_t i, n = ndat();
		if (arr.ndat() != n) return 0;
		for (i = 0; i < n; i++) _data[i] -= arr(i);
		return 1;
	}

	/// Multiply
	int multiply(T v)
	{
		size_t i, n = ndat();
		for (i = 0; i < n; i++) _data[i] *= v;
		return 1;
	}
	
	int multiply(const zeArrayTmp<T> & arr)
	{
		size_t i, n = ndat();
		if (arr.ndat() != n) return 0;
		for (i = 0; i < n; i++) _data[i] *= arr(i);
		return 1;
	}

	/// Divide
	int divide(T v)
	{
		if (v == 0) return 0;
		size_t i, n = ndat();
		for (i = 0; i < n; i++) _data[i] /= v;
		return 1;
	}

	int divide(const zeArrayTmp<T> & arr)
	{
		size_t i, n = ndat();
		if (arr.ndat() != n) return 0;
		for (i = 0; i < n; i++) {
			if (arr(i) == 0) return 0;
			_data[i] /= arr(i);
		}
		return 1;
	}
	
	/// count elements that equal v
	int counteq(T v) const
	{
		size_t i, m = 0, n = ndat();
		for (i = 0; i < n; i++) {
			if (_data[i] == v) m++;
		}
		return m;
	}

	/// count elements that do not equal v
	int countne(T v) const
	{
		size_t i, m = 0, n = ndat();
		for (i = 0; i < n; i++) {
			if (_data[i] != v) m++;
		}
		return m;
	}

	/// find index of elements that equal v
	int findeq(zeUint & idx, T v) const
	{
		size_t i, m = counteq(v), n = ndat();
		if (m > 0) {
			idx.resize(1, m);
			m = 0;
			for (i = 0; i < n; i++) {
				if (_data[i] == v) {
					idx(m) = i;
					m++;
				}
			}
		}
		return 1;
	}

	/// find index of elements that do not equal v
	int findne(zeUint & idx, T v) const
	{
		size_t i, m = countne(v), n = ndat();
		if (m > 0) {
			idx.resize(1, m);
			m = 0;
			for (i = 0; i < n; i++) {
				if (_data[i] != v) {
					idx(m) = i;
					m++;
				}
			}
		}
		return 1;
	}

	/// Get the index of  non-zero elements in both arrays
	int and(zeChar & idx, const zeArrayTmp<T> & arr) const
	{
		size_t  i, count = 0, n = ndat();
		if (arr.ndat() != n) return 0;
		idx.resize(_nrow, _ncol);
		for (i = 0; i < n; i++) {
			if (_data[i] != 0 && arr(i) != 0) {
				idx(i) = 1;
				count++;
			}
			else {
				idx(i) = 0;
			}
		}
		return 1;
	}

	/// Get the index of  non-zero elements in either arrays 
	int or(zeChar & idx, const zeArrayTmp<T> & arr) const
	{
		size_t  i, count = 0, n = ndat();
		if (arr.ndat() != n) return 0;
		idx.resize(_nrow, _ncol);
		for (i = 0; i < n; i++) {
			if (_data[i] != 0 || arr(i) != 0) {
				idx(i) = 1;
				count++;
			}
			else {
				idx(i) = 0;
			}
		}
		return 1;
	}

	/// Get the index of exclusive elements between the arrays
	int xor(zeChar & idx, const zeArrayTmp<T> & arr) const
	{
		size_t  i, count = 0, n = ndat();
		if (arr.ndat() != n) return 0;
		idx.resize(_nrow, _ncol);
		for (i = 0; i < n; i++) {
			if (_data[i] == 0) {
				if (arr(i) != 0) {
					idx(i) = 1;
					count++;
				}
				else {
					idx(i) = 0;
				}
			}
			else {
				if (arr(i) == 0) {
					idx(i) = 1;
					count++;
				}
				else {
					idx(i) = 0;
				}
			}
		}
		return 1;
	}

	/// Get the index of elements that equals the input array
	int equal(zeChar& idx, const zeArrayTmp<T> & arr) const
	{
		size_t i, count = 0, n = ndat();
		idx.resize(_nrow, _ncol);
		for (i = 0; i < n; i++) {
			if (_data[i] == arr(i)) {
				idx(i) = 1;
				count++;
			}
			else {
				idx(i) = 0;
			}
		}
		return 1;
	}

	int equal(zeChar& idx, T v) const
	{
		size_t i, count = 0, n = ndat();
		idx.resize(_nrow, _ncol);
		for (i = 0; i < n; i++) {
			if (_data[i] == v) {
				idx(i) = 1;
				count++;
			}
			else {
				idx(i) = 0;
			}
		}
		return 1;
	}

	/// Get the index of elements that do not equal the input array
	int nequal(zeChar& idx, const zeArrayTmp<T> & arr) const
	{
		size_t i, count = 0, n = ndat();
		idx.resize(_nrow, _ncol);
		for (i = 0; i < n; i++) {
			if (_data[i] != arr(i)) {
				idx(i) = 1;
				count++;
			}
			else {
				idx(i) = 0;
			}
		}
		return 1;
	}

	int nequal(zeChar& idx, T v) const
	{
		size_t i, count = 0, n = ndat();
		idx.resize(_nrow, _ncol);
		for (i = 0; i < n; i++) {
			if (_data[i] != v) {
				idx(i) = 1;
				count++;
			}
			else {
				idx(i) = 0;
			}
		}
		return 1;
	}

	/// Get the index of elements that are less than the input array
	int less(zeChar& idx, const zeArrayTmp<T> & arr) const
	{
		size_t i, count = 0, n = ndat();
		idx.resize(_nrow, _ncol);
		for (i = 0; i < n; i++) {
			if (_data[i] < arr(i)) {
				idx(i) = 1;
				count++;
			}
			else {
				idx(i) = 0;
			}
		}
		return 1;
	}

	int less(zeChar& idx, T v) const
	{
		size_t i, count = 0, n = ndat();
		idx.resize(_nrow, _ncol);
		for (i = 0; i < n; i++) {
			if (_data[i] < v) {
				idx(i) = 1;
				count++;
			}
			else {
				idx(i) = 0;
			}
		}
		return 1;
	}

	/// Get the index of elements that are less than or equal to the input array
	int lequal(zeChar& idx, const zeArrayTmp<T> & arr) const
	{
		size_t i, count = 0, n = ndat();
		idx.resize(_nrow, _ncol);
		for (i = 0; i < n; i++) {
			if (_data[i] <= arr(i)) {
				idx(i) = 1;
				count++;
			}
			else {
				idx(i) = 0;
			}
		}
		return 1;
	}

	int lequal(zeChar& idx, T v) const
	{
		size_t i, count = 0, n = ndat();
		idx.resize(_nrow, _ncol);
		for (i = 0; i < n; i++) {
			if (_data[i] <= v) {
				idx(i) = 1;
				count++;
			}
			else {
				idx(i) = 0;
			}
		}
		return 1;
	}

	/// Get the index of elements that are greater than the input array
	int greater(zeChar& idx, const zeArrayTmp<T> & arr) const
	{
		size_t i, count = 0, n = ndat();
		idx.resize(_nrow, _ncol);
		for (i = 0; i < n; i++) {
			if (_data[i] > arr(i)) {
				idx(i) = 1;
				count++;
			}
			else {
				idx(i) = 0;
			}
		}
		return 1;
	}

	int greater(zeChar& idx, T v) const
	{
		size_t i, count = 0, n = ndat();
		idx.resize(_nrow, _ncol);
		for (i = 0; i < n; i++) {
			if (_data[i] > v) {
				idx(i) = 1;
				count++;
			}
			else {
				idx(i) = 0;
			}
		}
		return 1;
	}

	/// Get the index of elements that are greater than or equal to the input array
	int gequal(zeChar& idx, const zeArrayTmp<T> & arr) const
	{
		size_t i, count = 0, n = ndat();
		idx.resize(_nrow, _ncol);
		for (i = 0; i < n; i++) {
			if (_data[i] >= arr(i)) {
				idx(i) = 1;
				count++;
			}
			else {
				idx(i) = 0;
			}
		}
		return 1;
	}

	int gequal(zeChar& idx, T v) const
	{
		size_t i, count = 0, n = ndat();
		idx.resize(_nrow, _ncol);
		for (i = 0; i < n; i++) {
			if (_data[i] >= v) {
				idx(i) = 1;
				count++;
			}
			else {
				idx(i) = 0;
			}
		}
		return 1;
	}

	/// Replace zeArrayTmp elements by index
	int setidx(const zeChar & idx, const zeArrayTmp<T> & arr)
	{
		size_t  i, n = ndat();
		if (arr.ndat() != n || idx.ndat() != n) return 0;
		for (i = 0; i < n; i++) {
			if (idx(i)) _data[i] = arr(i);
    	}
		return 1;
	}

	int setidx(const zeChar & idx, T v)
	{
		size_t  i, n = ndat();
		if (idx.ndat() != n) return 0;
		for (i = 0; i < n; i++) {
			if (idx(i)) _data[i] = v;
		}
		return 1;
	}

	/// Get zeArrayTmp elements by index
	int getidx(zeArrayTmp<T> & arr, const zeChar & idx) const
	{
		size_t n = idx.ndat();
		if (ndat() != n) return 0;
		size_t count = idx.countne(0);
		if (count > 0) {
			arr.resize(count, 1);
			count = 0;
			for (size_t i = 0; i < n; i++) {
				if (idx(i)) {
					arr(count) = _data[i];
					count++;
				}
			}
		}
		return 1;
	}

	/// Set array block
	int setblc(const zeArrayTmp<T> & arr, size_t i, size_t j)
	{
		if (i + arr.nrow() > nrow() || j + arr.ncol() > ncol()) return 0;
		for (size_t l = 0; l < arr.ncol(); l++) {
			for (size_t k = 0; k < arr.nrow(); k++) (*this)(i+k, j+l) = arr(k, l);
		}
		return 1;
	}

	/// Get array block
	int getblc(zeArrayTmp<T> & arr, size_t i, size_t m, size_t j, size_t n) const
	{
		if (i + m > nrow() || j + n > ncol()) return 0;
		arr.resize(m, n);
		for (size_t l = 0; l < n; l++) {
			for (size_t k = 0; k < m; k++) arr(k, l) = (*this)(i+k, j+l);
		}
		return 1;
	}

	/// Replace a column
	int setcol(const zeArrayTmp<T> & arr, const size_t j)
	{
		size_t  i;
		if (j >= ncol() || arr.ndat() != nrow()) return 0;
		for (i = 0; i < nrow(); i++) (*this)(i, j) = arr(i);
		return 1;
	}

	int setcol(T v, const size_t j)
	{
		size_t  i;
		if (j >= ncol()) return 0;
		for (i = 0; i < nrow(); i++) (*this)(i, j) = v;
		return 1;
	}

	/// Replace a row
	int setrow(const zeArrayTmp<T> & arr, const size_t i)
	{
		size_t  j;
		if (i >= nrow() || arr.ndat() != ncol()) return 0;
		for (j = 0; j < ncol(); j++) (*this)(i, j) = arr(j);
		return 1;
	}

	int setrow(T v, const size_t i)
	{
		size_t  j;
		if (i >= nrow()) return 0;
		for (j = 0; j < ncol(); j++) (*this)(i, j) = v;
		return 1;
	}

	/// Get a column
	int getcol(zeArrayTmp<T> & arr, const size_t j) const
	{
		if (j >= ncol()) return 0;
		arr.resize(nrow(), 1);
#ifdef ROW_MAJOR
		for (size_t i = 0; i < nrow(); i++) arr(i) = (*this)(i, j);
#else
		memcpy(arr.data(), &_data[j*_nrow], sizeof(T)*_nrow);
#endif
		return 1;
	}

	/// Get a row
	int getrow(zeArrayTmp<T> & arr, const size_t i) const
	{
		if (i >= nrow()) return 0;
		arr.resize(1, ncol());
#ifdef ROW_MAJOR
		memcpy(arr.data(), &_data[i*_ncol], sizeof(T)*_ncol);
#else
		for (size_t j = 0; j < ncol(); j++) arr(j) = (*this)(i, j);
#endif
		return 1;
	}

	/// transpose.
	int transpose()
	{
		size_t i, j;
		if (nrow() == ncol()) {
			for (j = 0; j < ncol(); j++) {
				for (i = j+1; i < nrow(); i++) {
					T v = (*this)(i, j);
					(*this)(i, j) = (*this)(j, i);
					(*this)(j, i) = v;
				}
			}
		}
		else {
			zeArrayTmp<T> tmp = (*this);
			resize(ncol(), nrow());
			for (j = 0; j < tmp.ncol(); j++) {
				for (i = 0; i < tmp.nrow(); i++) (*this)(j, i) = tmp(i, j);
			}
		}
		return 1;
	}

	/// swap columns
	int swapcol(const size_t i, const size_t j)
	{
		if (i >= ncol() || j >= ncol()) return 0;
		for (size_t k = 0; k < nrow(); k++) {
			T value = (*this)(k, i);
			(*this)(k, i) = (*this)(k, j);
			(*this)(k, j) = value;
		}
		return 1;
	}

	/// swap rows
	int swaprow(const size_t i, const size_t j)
	{
		if (i >= nrow() || j >= nrow()) return 0;
		for (size_t k = 0; k < ncol(); k++) {
			T value = (*this)(i, k);
			(*this)(i, k) = (*this)(j, k);
			(*this)(j, k) = value;
		}
		return 1;
	}

	/// flip columns
	int flipcol()
	{
		size_t k, n = ncol() / 2;
		for (k = 0; k < n; k++) swapcol(k, ncol() - k - 1);
		return 1;
	}

	/// flip rows
	int fliprow()
	{
		size_t k, n = nrow() / 2;
		for (k = 0; k < n; k++) swaprow(k, nrow() - k - 1);
		return 1;
	}

	/// shift columns
	int shiftcol(int n)
	{
		if (n < 0) n = ncol() + n;		// shift right
		if (n == 0 || n == ncol()) return 1;
		if (n > ncol()) return 0;

		zeArrayTmp<T> tmp = (*this), p;
		size_t k;

		for (k = n; k < ncol(); k++) {
			tmp.getcol(p, k);
			setcol(p, k - n);
		}
		for (k = 0; k < n; k++) {
			tmp.getcol(p, k);
			setcol(p, ncol() - n + k);
		}

		return 1;
	}

	/// shift rows
	int shiftrow(int n)
	{
		if (n < 0) n = nrow() + n;		// shift down
		if (n == 0 || n == nrow()) return 1;
		if (n > nrow()) return 0;

		zeArrayTmp<T> tmp = (*this), p;
		size_t k;

		for (k = n; k < nrow(); k++) {
			tmp.getrow(p, k);
			setrow(p, k - n);
		}
		for (k = 0; k < n; k++) {
			tmp.getrow(p, k);
			setrow(p, nrow() - n + k);
		}

		return 1;
	}

	/// insert column
	int insertcol(const zeArrayTmp<T> & arr, const size_t j)
	{
		if (arr.nrow() != nrow() || j > ncol()) return 0;
		zeArrayTmp<T> tmp = (*this), p;
		resize(nrow(), ncol() + arr.ncol());
		size_t k;
		for (k = 0; k < j; k++) {
			tmp.getcol(p, k);
			setcol(p, k);
		}
		for (k = 0; k < arr.ncol(); k++) {
			arr.getcol(p, k);
			setcol(p, k + j);
		}
		for (k = j; k < tmp.ncol(); k++) {
			tmp.getcol(p, k);
			setcol(p, k + arr.ncol());
		}
		return 1;
	}

	int insertcol(T v, const size_t j)
	{
		if (j > ncol()) return 0;
		zeArrayTmp<T> tmp = (*this), p;
		resize(nrow(), ncol() + 1);
		size_t k;
		for (k = 0; k < j; k++) {
			tmp.getcol(p, k);
			setcol(p, k);
		}
		setcol(v, j);
		for (k = j; k < tmp.ncol(); k++) {
			tmp.getcol(p, k);
			setcol(p, k + 1);
		}
		return 1;
	}
	
	/// insert row
	int insertrow(const zeArrayTmp<T> & arr, const size_t i)
	{
		if (arr.ncol() != ncol() || i > nrow()) return 0;
		zeArrayTmp<T> tmp = (*this), p;
		resize(nrow() + arr.nrow(), ncol());
		size_t k;
		for (k = 0; k < i; k++) {
			tmp.getrow(p, k);
			setrow(p, k);
		}
		for (k = 0; k < arr.nrow(); k++) {
			arr.getrow(p, k);
			setrow(p, k + i);
		}
		for (k = i; k < tmp.nrow(); k++) {
			tmp.getrow(p, k);
			setrow(p, k + arr.nrow());
		}
		return 1;
	}

	int insertrow(const T v, const size_t i)
	{
		if (i > nrow()) return 0;
		zeArrayTmp<T> tmp = (*this), p;
		resize(nrow() + 1, ncol());
		size_t k;
		for (k = 0; k < i; k++) {
			tmp.getrow(p, k);
			setrow(p, k);
		}
		setrow(v, i);
		for (k = i; k < tmp.nrow(); k++) {
			tmp.getrow(p, k);
			setrow(p, k + 1);
		}
		return 1;
	}
	
	/// delete column
	int delcol(const size_t j)
	{
		if (ncol() < 2 || j > ncol()) return 0;
		zeArrayTmp<T> tmp = (*this), p;
		resize(nrow(), ncol()-1);
		size_t k;
		for (k = 0; k < j; k++) {
			tmp.getcol(p, k);
			setcol(p, k);
		}
		for (k = j + 1; k < tmp.ncol(); k++) {
			tmp.getcol(p, k);
			setcol(p, k - 1);
		}
		return 1;
	}

	/// delete row
	int delrow(const size_t i)
	{
		if (nrow() < 2 || i > nrow()) return 0;
		zeArrayTmp<T> tmp = (*this), p;
		resize(nrow()-1 , ncol());
		size_t k;
		for (k = 0; k < i; k++) {
			tmp.getrow(p, k);
			setrow(p, k);
		}
		for (k = i + 1; k < tmp.nrow(); k++) {
			tmp.getrow(p, k);
			setrow(p, k-1);
		}
		return 1;
	}

	/// Concat product
	int concat(const zeArrayTmp<T> & arr)
	{
		if (arr.nrow() != ncol()) return 0;
		
		zeArrayTmp<T> tmp = (*this);
		resize(nrow(), arr.ncol());
		fill(static_cast<T>(0), static_cast<T>(0));

		if (arr.data() == data())
		{
			for (size_t i = 0; i < tmp.nrow(); i++) {
				for (size_t j = 0; j < arr.ncol(); j++) {
					for (size_t k = 0; k < tmp.ncol(); k++) (*this)(i, j) += tmp(i, k) * tmp(k, j);
				}
			}
		}
		else
		{
			for (size_t i = 0; i < tmp.nrow(); i++) {
				for (size_t j = 0; j < arr.ncol(); j++) {
					for (size_t k = 0; k < tmp.ncol(); k++) (*this)(i, j) += tmp(i, k) * arr(k, j);
				}
			}
		}

		return 1;
	}
	
	/// Sort array in assending or dessending order.
	int sort(const size_t k, int assending)
	{
		if (k >= ncol()) return 0;
		
		zeArrayTmp<T> col;
		getcol(col, k);

		zeUint idx;
		idx.resize(nrow(), 1);

		HeapSort(col.data(), idx.data(), nrow(), assending);
	
		setcol(col, k);

		for (size_t j = 0; j < ncol(); j++) {
			if (j == k) continue;
			for (size_t i = 0; i < nrow(); i++) {
				col(i) = (*this)(idx(i), j);
			}
			setcol(col, j);
		}

		return 1;
	}

	/// Remove duplicated data in the array.
	int unique(int assending)
	{
		size_t i, count = 1, n = ndat();
		if (n == 1) return 1;
		HeapSort(_data, 0, n, assending);
		for (i = 1;  i < n; i++) {
			if (_data[i-1] != _data[i]) {
				if (i != 1) _data[count] = _data[i];
				count++;
			}
		}
		resize(count, 1, true);
		return 1;
	}

	/// Confine
	int confine(T vmin, T vmax)
	{
		size_t k, n = ndat();
		for (k = 0; k < n; k++)
		{
			(*this)(k) = MIN(vmax, (*this)(k));
			(*this)(k) = MAX(vmin, (*this)(k));
		}
		return 1;
	}

	T vmin(T v) const
	{
		size_t k, n = ndat();
		for (k = 0; k < n; k++) v = MIN(v, (*this)(k));
		return v;
	}

	T vmax(T v) const
	{
		size_t k, n = ndat();
		for (k = 0; k < n; k++) v = MAX(v, (*this)(k));
		return v;
	}

	double sum() const
	{
		double v = 0;
		size_t k, n = ndat();
		for (k = 0; k < n; k++) v += (*this)(k);
		return v;
	}

	int print(const char *format, FILE *fp) const
	{
		size_t i, j;
		for (i = 0; i < nrow(); i++) {
			for (j = 0; j < ncol(); j++) fprintf(fp, format, (*this)(i, j));
			fprintf(fp, "\n");
		}
		return 1;
	}

	/// parse string to data
	int parse(const size_t i, const char *source)
	{
		if (i >= nrow()) return 0;

		std::istringstream istr(source);
		size_t j = 0;

		while (!istr.eof())
		{
			if (j >= ncol()) return 0;
			T tmp;
			istr >> tmp;
			(*this)(i, j) = tmp;
			j++;
		}

		if (j != nrow()) return 0;

		return 1;
	}
};

class zeChar : public zeArrayTmp<char>
{
public:
	 zeChar() {_type = ARRAY_CHAR;}
	~zeChar() { }
};

class zeUchar : public zeArrayTmp<unsigned char>
{
public:
	 zeUchar() {_type = ARRAY_UCHAR;}
	~zeUchar() { }
};

class zeShort : public zeArrayTmp<short>
{
public:
	 zeShort() {_type = ARRAY_SHORT;}
	~zeShort() { }
};

class zeUshort : public zeArrayTmp<unsigned short>
{
public:
	 zeUshort() {_type = ARRAY_USHORT;} 
	~zeUshort() { }
};

class zeInt : public zeArrayTmp<int>
{
public:
	 zeInt() {_type = ARRAY_INT;}
	~zeInt() { }
};

class zeUint : public zeArrayTmp<unsigned int>
{
public:
	 zeUint() {_type = ARRAY_UINT;}
	~zeUint() { }
};

class zeFloat : public zeArrayTmp<float>
{
public:
	 zeFloat() {_type = ARRAY_FLOAT;}
	~zeFloat() { }
};

class zeDouble : public zeArrayTmp<double>
{
public:
	zeDouble()  {_type = ARRAY_DOUBLE;}
	~zeDouble() { }
};

class zeComplex : public zeArrayTmp<sComplex>
{
public:
	 zeComplex() {_type = ARRAY_COMPLEX; }
	~zeComplex() { }
};

class zeTriplex : public zeArrayTmp<sTriplex>
{
public:
	 zeTriplex() {_type = ARRAY_TRIPLEX; }
	~zeTriplex() { }
};

class zeQuadruple : public zeArrayTmp<sQuadruple>
{
public:
	 zeQuadruple() {_type = ARRAY_QUADRUPLE; }
	~zeQuadruple() { }
};


//////////////////////////////////////////////////////////////////////
//
template <typename T>
inline void Swap(T& a, T& b)
{
	T c = a; a = b; b = c;
}

//////////////////////////////////////////////////////////////////////
// For heap sort of assending order
//
template <typename T>
void SiftDown(T v[], size_t idx[], int parent, int n)
{
	// 0 based child-parent relationship
	int child = 2 * parent + 1;
	// find where element being sifted belongs
	while (child < n)
	{
		// use right child if it is bigger
		if (child != n - 1 && v[child] < v[child + 1])
			child++;
		// check if it is done
		if (v[parent] >= v[child])
			break;
		// if not, swap elements and move parent down
		Swap(v[parent], v[child]);
		if (idx) Swap (idx[parent], idx[child]);
		parent = child;
		child = 2 * parent + 1;
	}
}

//////////////////////////////////////////////////////////////////////
// For heap sort of decending order
//
template <typename T>
void SiftUp(T v[], size_t idx[], int parent, int n)
{
	// 0 based child-parent relationship
	int child = 2 * parent + 1;
	// find where element being sifted belongs
	while (child < n)
	{
		// use right child if it is smaller
		if (child != n - 1 && v[child] > v[child + 1])
			child++;
		if (v[parent] <= v[child])
			break;
		// if not, swap elements and move parent down
		Swap(v[parent], v[child]);
		if (idx) Swap (idx[parent], idx[child]);
		parent = child;
		child = 2 * parent + 1;
	}
}

//////////////////////////////////////////////////////////////////////
// Sort v[] in assending or dessending order according to the flag.
// If an index vector is provided, corresponding sorted indices can be
// used to rearrange accompany vectors.
//
template <typename T>
void HeapSort(T v[], size_t idx[], int n, int flag = 1)
{
	int i;
	
	if (idx) for (i = 0; i < n; i++) idx[i] = i;

	// construct heap
	for (i = (n - 1) / 2; i >= 0; i--)
	{
		if (flag > 0)
			SiftDown(v, idx, i, n);
		else
			SiftUp(v, idx, i, n);
	}

	for (i = n - 1; i >= 1; i--)
	{
		// put top element in its correct position
		Swap(v[0], v[i]);
		if (idx) Swap (idx[0], idx[i]);
		// rebuild heap
		if (flag > 0)
			SiftDown(v, idx, 0, i);
		else
			SiftUp(v, idx, 0, i);
	}
}

#endif	// __ZE_ARRAY__