#include "matrix.h"
#include "array.h"
#include "amath.h"
#include <math.h>
#include <vector>;

using namespace std;

#pragma warning(once: 4244)

#define RAD_TO_DEG		57.295779513082300 

CtkObject ctkNull  = {CTK_NULL, {0}};

inline void invalid_input()
{
	ctkThrowException(": Invalid arguments. ");
}

inline void *CtkObject2Ptr(CtkObject &o)
{
	CtkObject *p = (CtkObject*)o.u.ptr;
	if (p->type != 'MAT') ctkThrowException(": Invalid matrix object. ");;
	return p->u.ptr;
}

static void mat_delete(void *ptr)
{
	if (array_free(((CtkObject*)ptr)->u.ptr) == 0) ctkThrowException(array_get_error_message());
}

static CtkObject mat_create_array(const char* name, int nrow, int ncol)
{
	void *arr = array_create(name, nrow, ncol);
	if (!arr) ctkThrowException(array_get_error_message());
    CtkObject* o = (CtkObject*)ctkAllocateObject(sizeof(CtkObject), CtkFinalizer(mat_delete));
	o->type = 'MAT';
	o->u.ptr = arr;
	CtkObject ret;
	MAKE_USER_TYPE(ret, o);
	return ret;
}

static CtkObject mat_make_array(CtkObject &obj, bool clone)
{
	CtkObject ret = mat_create_array(array_get_typename(CtkObject2Ptr(obj)), 0, 0);
	if (clone) {
		if (array_clone(CtkObject2Ptr(ret),
						CtkObject2Ptr(obj)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}

////////////////// basic array operations //////////////////////////////

static CtkObject matCreate(int nArgs, CtkObject* args)
{
	if (nArgs < 1 || !IS_STRING(args[0])) invalid_input();
	if (nArgs == 1) {
		return mat_create_array(args[0].u.svalue, 0, 0);
	}
	else {
		if (nArgs < 3 ||
			!IS_NUMBER(args[1]) ||
			!IS_NUMBER(args[2])) invalid_input();
		return mat_create_array(args[0].u.svalue, GET_NUMBER(args[1]), GET_NUMBER(args[2]));
	}
	return ctkNull;
}
static CtkPrimitive mat1("matCreate", matCreate);


static CtkObject matPrint(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	if (nArgs == 1) {
		if (array_print(CtkObject2Ptr(args[0]),
						0, stdout) == 0) ctkThrowException(array_get_error_message());
	}
	else if (nArgs > 2) {
		if (!IS_STRING(args[1]) || !IS_STRING(args[2])) invalid_input();
		char *mode = "a";
		if (nArgs > 3 && IS_STRING(args[3])) mode = args[3].u.svalue;
		FILE *f = fopen(args[2].u.svalue, mode);
		if (!f) ctkThrowException(": Connot open the file.");
		if (array_print(CtkObject2Ptr(args[0]),
						args[1].u.svalue,
						f) == 0) ctkThrowException(array_get_error_message());
		fclose(f);
	}
	else {
		if (!IS_STRING(args[1])) invalid_input();
		if (array_print(CtkObject2Ptr(args[0]),
						args[1].u.svalue,
						stdout) == 0) ctkThrowException(array_get_error_message());
	}
	return ctkNull;
}
static CtkPrimitive mat2("matPrint", matPrint);


static CtkObject matGetPtr(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	size_t esize, n;
	void *ptr = array_get_ptr(CtkObject2Ptr(args[0]), &esize, &n);
	if (!ptr) ctkThrowException(array_get_error_message());
	CtkArray* arr = ctkCreateArray();
	CtkObject key, value;
	MAKE_INTEGER(key, 0);
	MAKE_RAW_POINTER(value, ptr);
	ctkPutArray(arr, key, value);
	MAKE_INTEGER(key, 1);
	MAKE_INTEGER(value, n);
	ctkPutArray(arr, key, value);
	MAKE_INTEGER(key, 2);
	MAKE_INTEGER(value, esize);
	ctkPutArray(arr, key, value);
	MAKE_ARRAY(value, arr);
	return value;
}
static CtkPrimitive mat3("matGetPtr", matGetPtr);


static CtkObject matGetSize(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	size_t nrow, ncol;
	if (array_get_size(CtkObject2Ptr(args[0]), &nrow, &ncol) == 0) ctkThrowException(array_get_error_message());
	CtkArray* arr = ctkCreateArray();
	CtkObject key, value;
	MAKE_INTEGER(key, 0);
	MAKE_INTEGER(value, nrow);
	ctkPutArray(arr, key, value);
	MAKE_INTEGER(key, 1);
	MAKE_INTEGER(value, ncol);
	ctkPutArray(arr, key, value);
	MAKE_INTEGER(key, 2);
	MAKE_INTEGER(value, nrow * ncol);
	ctkPutArray(arr, key, value);
	MAKE_ARRAY(value, arr);
	return value;
}
static CtkPrimitive mat4("matGetSize", matGetSize);


static CtkObject matResize(int nArgs, CtkObject* args)
{
	if (nArgs < 3 || !IS_NUMBER(args[1]) || !IS_NUMBER(args[2]) ) invalid_input();
	if (array_resize(CtkObject2Ptr(args[0]),
					 GET_NUMBER(args[1]),
					 GET_NUMBER(args[2]),
					 true) == 0) ctkThrowException(array_get_error_message());
	return ctkNull;
}
static CtkPrimitive mat5("matResize", matResize);


static CtkObject matReshape(int nArgs, CtkObject* args)
{
	if (nArgs < 3 || !IS_NUMBER(args[1]) || !IS_NUMBER(args[2]) ) invalid_input();
	if (array_reshape(CtkObject2Ptr(args[0]),
					  GET_NUMBER(args[1]),
					  GET_NUMBER(args[2])) == 0) ctkThrowException(array_get_error_message());
	return ctkNull;
}
static CtkPrimitive mat6("matReshape", matReshape);


static CtkObject matFill(int nArgs, CtkObject* args)
{
	if (nArgs < 3 || !IS_NUMBER(args[1]) || !IS_NUMBER(args[2])) invalid_input();
	if (array_fill(CtkObject2Ptr(args[0]),
				   GET_NUMBER(args[1]),
				   GET_NUMBER(args[2])) == 0) ctkThrowException(array_get_error_message());
	return ctkNull;
}
static CtkPrimitive mat7("matFill", matFill);


static CtkObject matRand(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	if (nArgs > 1) {
		if (!IS_NUMBER(args[1])) invalid_input();
		if (amath_rand(CtkObject2Ptr(args[0]),
					   GET_NUMBER(args[1])) == 0) ctkThrowException(array_get_error_message());
	}
	else {
		if (amath_rand(CtkObject2Ptr(args[0]),
					   10) == 0) ctkThrowException(array_get_error_message());
	}
	return ctkNull;
}
static CtkPrimitive mat8("matRand", matRand);


static CtkObject matGetRow(int nArgs, CtkObject* args)
{
	if (nArgs < 2 || !IS_NUMBER(args[1])) invalid_input();
	CtkObject ret = mat_make_array(args[0], false);
	if (array_get_row(CtkObject2Ptr(ret),
					  CtkObject2Ptr(args[0]),
					  GET_NUMBER(args[1])) == 0) ctkThrowException(array_get_error_message());
	return ret;
}
static CtkPrimitive mat17("matGetRow", matGetRow);


static CtkObject matSetRow(int nArgs, CtkObject* args)
{
	if (nArgs < 3) invalid_input();
	void *ptr = CtkObject2Ptr(args[0]);
	for (int i = 2; i < nArgs; i += 2) {
		if (!IS_NUMBER(args[i-1])) invalid_input();
		if (IS_NUMBER(args[i])) {
			if (array_set_row(ptr,
							  GET_NUMBER(args[i]),
							  GET_NUMBER(args[i-1])) == 0) ctkThrowException(array_get_error_message());
		}
		else {
			if (array_set_row2(ptr,
							   CtkObject2Ptr(args[i]),
							   GET_NUMBER(args[i-1])) == 0) ctkThrowException(array_get_error_message());
		}
	}
	return ctkNull;
}
static CtkPrimitive mat19("matSetRow", matSetRow);


static CtkObject matGetColumn(int nArgs, CtkObject* args)
{
	if (nArgs < 2 || !IS_NUMBER(args[1])) invalid_input();
	CtkObject ret = mat_make_array(args[0], false);
	if (array_get_column(CtkObject2Ptr(ret),
						 CtkObject2Ptr(args[0]),
						 GET_NUMBER(args[1])) == 0) ctkThrowException(array_get_error_message());
	return ret;
}
static CtkPrimitive mat18("matGetColumn", matGetColumn);


static CtkObject matSetColumn(int nArgs, CtkObject* args)
{
	if (nArgs < 3) invalid_input();
	void *ptr = CtkObject2Ptr(args[0]);
	for (int i = 2; i < nArgs; i += 2) {
		if (!IS_NUMBER(args[i-1])) invalid_input();
		if (IS_NUMBER(args[i])) {
			if (array_set_column(ptr,
								 GET_NUMBER(args[i]),
								 GET_NUMBER(args[i-1])) == 0) ctkThrowException(array_get_error_message());
		}
		else {
			if (array_set_column2(ptr,
								  CtkObject2Ptr(args[i]),
								  GET_NUMBER(args[i-1])) == 0) ctkThrowException(array_get_error_message());
		}
	}
	return ctkNull;
}
static CtkPrimitive mat20("matSetColumn", matSetColumn);


static CtkObject matGetByIndex(int nArgs, CtkObject* args)
{
	if (nArgs < 2 || !IS_USER_TYPE(args[1])) invalid_input();
	CtkObject ret = mat_make_array(args[0], false);
	if (array_get_by_index(CtkObject2Ptr(ret),
						   CtkObject2Ptr(args[0]),
						   CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());
	return ret;
}
static CtkPrimitive mat21("matGetByIndex", matGetByIndex);


static CtkObject matSetByIndex(int nArgs, CtkObject* args)
{
	if (nArgs < 3) invalid_input();
	if (IS_NUMBER(args[2])) {
		if (array_set_by_index(CtkObject2Ptr(args[0]),
							   GET_NUMBER(args[2]),
							   CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());
	}
	else {
		if (array_set_by_index2(CtkObject2Ptr(args[0]),
								CtkObject2Ptr(args[2]),
							    CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());
	}
	return ctkNull;
}
static CtkPrimitive mat22("matSetByIndex", matSetByIndex);


static CtkObject matDeleteRow(int nArgs, CtkObject* args)
{
	if (nArgs < 2 || !IS_NUMBER(args[1])) invalid_input();
	if (array_delete_row(CtkObject2Ptr(args[0]),
						 GET_NUMBER(args[1])) == 0) ctkThrowException(array_get_error_message());
	return ctkNull;
}
static CtkPrimitive mat23("matDeleteRow", matDeleteRow);


static CtkObject matDeleteColumn(int nArgs, CtkObject* args)
{
	if (nArgs < 2 || !IS_NUMBER(args[1])) invalid_input();
	if (array_delete_column(CtkObject2Ptr(args[0]),
							GET_NUMBER(args[1])) == 0) ctkThrowException(array_get_error_message());
	return ctkNull;
}
static CtkPrimitive mat24("matDeleteColumn", matDeleteColumn);


static CtkObject matInsertRow(int nArgs, CtkObject* args)
{
	if (nArgs < 3 || !IS_NUMBER(args[1])) invalid_input();
	if (IS_NUMBER(args[2])) {
		if (array_insert_row(CtkObject2Ptr(args[0]),
							 GET_NUMBER(args[2]),
							 GET_NUMBER(args[1])) == 0) ctkThrowException(array_get_error_message());
	}
	else {
		if (array_insert_row2(CtkObject2Ptr(args[0]),
							  CtkObject2Ptr(args[2]),
							  GET_NUMBER(args[1])) == 0) ctkThrowException(array_get_error_message());
	}
	return ctkNull;
}
static CtkPrimitive mat25("matInsertRow", matInsertRow);


static CtkObject matInsertColumn(int nArgs, CtkObject* args)
{
	if (nArgs < 3 || !IS_NUMBER(args[1])) invalid_input();
	if (IS_NUMBER(args[2])) {
		if (array_insert_column(CtkObject2Ptr(args[0]),
								GET_NUMBER(args[2]),
								GET_NUMBER(args[1])) == 0) ctkThrowException(array_get_error_message());
	}
	else {
		if (array_insert_column2(CtkObject2Ptr(args[0]),
								 CtkObject2Ptr(args[2]),
								 GET_NUMBER(args[1])) == 0) ctkThrowException(array_get_error_message());
	}
	return ctkNull;
}
static CtkPrimitive mat26("matInsertColumn", matInsertColumn);


static CtkObject matGetBlock(int nArgs, CtkObject* args)
{
	if (nArgs < 5 ||
		!IS_NUMBER(args[1]) ||
		!IS_NUMBER(args[2]) ||
		!IS_NUMBER(args[3]) ||
		!IS_NUMBER(args[4])) invalid_input();
	CtkObject ret = mat_make_array(args[0], false);
	if (array_get_by_block(CtkObject2Ptr(ret),
						   CtkObject2Ptr(args[0]),
						   GET_NUMBER(args[1]),
						   GET_NUMBER(args[2]),
						   GET_NUMBER(args[3]),
						   GET_NUMBER(args[4])) == 0) ctkThrowException(array_get_error_message());
	return ret;
}
static CtkPrimitive mat27("matGetBlock", matGetBlock);


static CtkObject matSetBlock(int nArgs, CtkObject* args)
{
	if (nArgs < 4 ||
		!IS_NUMBER(args[1]) ||
		!IS_NUMBER(args[2])) invalid_input();
	if (array_set_by_block(CtkObject2Ptr(args[0]),
						   CtkObject2Ptr(args[3]),
						   GET_NUMBER(args[1]),
						   GET_NUMBER(args[2])) == 0) ctkThrowException(array_get_error_message());
	return ctkNull;
}
static CtkPrimitive mat28("matSetBlock", matSetBlock);


static CtkObject matTranspose(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	if (array_transpose(CtkObject2Ptr(args[0])) == 0) ctkThrowException(array_get_error_message());
	return ctkNull;
}
static CtkPrimitive mat31("matTranspose", matTranspose);


static CtkObject matSwapRows(int nArgs, CtkObject* args)
{
	if (nArgs < 3 ||
		!IS_NUMBER(args[1]) ||
		!IS_NUMBER(args[2])) invalid_input();
	if (array_swap_rows(CtkObject2Ptr(args[0]),
						GET_NUMBER(args[1]),
						GET_NUMBER(args[2])) == 0) ctkThrowException(array_get_error_message());
	return ctkNull;
}
static CtkPrimitive mat32("matSwapRows", matSwapRows);


static CtkObject matSwapColumns(int nArgs, CtkObject* args)
{
	if (nArgs < 3 ||
		!IS_NUMBER(args[1]) ||
		!IS_NUMBER(args[2])) invalid_input();
	if (array_swap_columns(CtkObject2Ptr(args[0]),
						GET_NUMBER(args[1]),
						GET_NUMBER(args[2])) == 0) ctkThrowException(array_get_error_message());
	return ctkNull;
}
static CtkPrimitive mat33("matSwapColumns", matSwapColumns);


static CtkObject matFlipRows(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	if (array_flip_rows(CtkObject2Ptr(args[0])) == 0) ctkThrowException(array_get_error_message());
	return ctkNull;
}
static CtkPrimitive mat34("matFlipRows", matFlipRows);


static CtkObject matFlipColumns(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	if (array_flip_columns(CtkObject2Ptr(args[0])) == 0) ctkThrowException(array_get_error_message());
	return ctkNull;
}
static CtkPrimitive mat35("matFlipColumns", matFlipColumns);


static CtkObject matShiftRows(int nArgs, CtkObject* args)
{
	if (nArgs < 2 || !IS_NUMBER(args[1])) invalid_input();
	if (array_shift_rows(CtkObject2Ptr(args[0]),
						 GET_NUMBER(args[1])) == 0) ctkThrowException(array_get_error_message());
	return ctkNull;
}
static CtkPrimitive mat36("matShiftRows", matShiftRows);


static CtkObject matProduct(int nArgs, CtkObject* args)
{
	if (nArgs < 2) invalid_input();
	if (array_product(CtkObject2Ptr(args[0]),
					  CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());
	return ctkNull;
}
static CtkPrimitive mat37("matProduct", matProduct);


static CtkObject matSort(int nArgs, CtkObject* args)
{
	if (nArgs < 2 || !IS_NUMBER(args[1])) invalid_input();
	if (nArgs > 2) {
		if (!IS_NUMBER(args[2])) invalid_input();
		if (array_sort(CtkObject2Ptr(args[0]),
					   GET_NUMBER(args[1]),
					   GET_NUMBER(args[2])) == 0) ctkThrowException(array_get_error_message());
	}
	else {
		if (array_sort(CtkObject2Ptr(args[0]),
					   GET_NUMBER(args[1]),
					   1) == 0) ctkThrowException(array_get_error_message());
	}
	return ctkNull;
}
static CtkPrimitive mat38("matSort", matSort);


static CtkObject matUnique(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	if (nArgs > 1) {
		if (!IS_NUMBER(args[1])) invalid_input();
		if (array_unique(CtkObject2Ptr(args[0]),
						 GET_NUMBER(args[1])) == 0) ctkThrowException(array_get_error_message());
	}
	else {
		if (array_unique(CtkObject2Ptr(args[0]),
						 1) == 0) ctkThrowException(array_get_error_message());
	}
	return ctkNull;
}
static CtkPrimitive mat39("matUnique", matUnique);


static CtkObject matMin(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	void * ptr = CtkObject2Ptr(args[0]);
	if (!array_check_type(ptr)) ctkThrowException(array_get_error_message());
	double v = array_min(ptr);
	CtkObject ret;
	if (array_is_double(ptr) || array_is_float(ptr))
		MAKE_REAL(ret, v);
	else
		MAKE_INTEGER(ret, v);

	return ret;
}
static CtkPrimitive mat40("matMin", matMin);


static CtkObject matMax(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	void * ptr = CtkObject2Ptr(args[0]);
	if (!array_check_type(ptr)) ctkThrowException(array_get_error_message());
	double v = array_max(ptr);
	CtkObject ret;
	if (array_is_double(ptr) || array_is_float(ptr))
		MAKE_REAL(ret, v);
	else
		MAKE_INTEGER(ret, v);

	return ret;
}
static CtkPrimitive mat41("matMax", matMax);


static CtkObject matSum(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	void * ptr = CtkObject2Ptr(args[0]);
	if (!array_check_type(ptr)) ctkThrowException(array_get_error_message());
	double v = array_sum(ptr);
	CtkObject ret;
	if (array_is_double(ptr) || array_is_float(ptr))
		MAKE_REAL(ret, v);
	else
		MAKE_INTEGER(ret, v);

	return ret;
}
static CtkPrimitive mat42("matSum", matSum);


static CtkObject matConfine(int nArgs, CtkObject* args)
{
	if (nArgs < 3 ||
		!IS_NUMBER(args[1]) ||
		!IS_NUMBER(args[2])) invalid_input();
	if (array_confine(CtkObject2Ptr(args[0]),
					  GET_NUMBER(args[1]),
					  GET_NUMBER(args[2])) == 0) ctkThrowException(array_get_error_message());
	return ctkNull;
}
static CtkPrimitive mat43("matConfine", matConfine);


static CtkObject matImport(int nArgs, CtkObject* args)
{
	if (nArgs < 4 ||
		!IS_RAW_POINTER(args[1]) ||
		!IS_NUMBER(args[2]) ||
		!IS_NUMBER(args[3])) invalid_input();
	if (array_import(CtkObject2Ptr(args[0]),
					 args[1].u.ptr,
					 GET_NUMBER(args[2]),
					 GET_NUMBER(args[3])) == 0) ctkThrowException(array_get_error_message());
	return ctkNull;
}
static CtkPrimitive mat45("matImport", matImport);


static CtkObject matStringToRow(int nArgs, CtkObject* args)
{
	if (nArgs < 3 ||
		!IS_NUMBER(args[1]) ||
		!IS_STRING(args[2])) invalid_input();
	if (array_parse(CtkObject2Ptr(args[0]),
					args[2].u.svalue,
					GET_NUMBER(args[1])) == 0) ctkThrowException(array_get_error_message());
	return ctkNull;
}
static CtkPrimitive mat46("matStringToRow", matStringToRow);


static CtkObject matReadText(int nArgs, CtkObject* args)
{
	if (nArgs < 1 || !IS_STRING(args[0])) invalid_input();
	 CtkObject ret = mat_create_array("double", 0, 0);
	 if (array_read_text(CtkObject2Ptr(ret), args[0].u.svalue) == 0) ctkThrowException(array_get_error_message());
	return ret;
}
static CtkPrimitive mat47("matReadText", matReadText);


#include "gshhs.h"

static CtkObject matReadGSHHS(int nArgs, CtkObject* args)
{
	if (nArgs < 2 ||
		!IS_STRING(args[0]) ||
		!IS_STRING(args[1])) invalid_input();

	FILE *fp = fopen(args[0].u.svalue, "rb");
	if (!fp) ctkThrowException(" Cannot open file for reading.");

	int level, west = 0, east = 360000000, south = -90000000, north = 90000000;

	if      (!strcmp(args[1].u.svalue, "land"))   level = 1;
	else if (!strcmp(args[1].u.svalue, "lake"))   level = 2;
	else if (!strcmp(args[1].u.svalue, "island")) level = 3;
	else if (!strcmp(args[1].u.svalue, "pond"))   level = 4;
	else ctkThrowException(" flag must be land, lake, island, or pond.");

	if (nArgs > 5 &&
		IS_NUMBER(args[2]) &&
		IS_NUMBER(args[3]) &&
		IS_NUMBER(args[4]) &&
		IS_NUMBER(args[5])) {
	
		west  = GET_NUMBER(args[2]) * 1000000;
		east  = GET_NUMBER(args[3]) * 1000000;
		south = GET_NUMBER(args[4]) * 1000000;
		north = GET_NUMBER(args[5]) * 1000000;
	}

	if (west >= east ||
		south >= north ||
		west < 0 ||
		east >  360000000 ||
		south < -90000000 ||
		north >  90000000) ctkThrowException(" Invalid lat-lon range.");


	XY p;
	GSHHS h;
	vector<XY> xy;
	xy.reserve(10000);

	while (fread((void *)&h, sizeof(struct GSHHS), 1, fp) == 1) {
		h.n = swabi4((unsigned int)h.n);
		h.level = swabi4((unsigned int)h.level);
		h.west = swabi4((unsigned int)h.west);
		h.east = swabi4((unsigned int)h.east);
		h.south = swabi4((unsigned int)h.south);
		h.north = swabi4((unsigned int)h.north);

		if (h.level != level ||
			h.west > east ||
			h.east < west ||
			h.south > north ||
			h.north < south) {
			fseek(fp, sizeof(struct XY) * h.n, SEEK_CUR);
			continue;
		}

		p.x = -360000000;
		p.y = -360000000;
		xy.push_back(p);

		for (int k = 0; k < h.n; k++) {
			if (fread ((void *)&p, sizeof(struct XY), 1, fp) != 1) ctkThrowException(" Failed to read gshhs data.");
			p.x = swabi4 ((unsigned int)p.x);
			p.y = swabi4 ((unsigned int)p.y);
			if (p.x < west ||
				p.x > east ||
				p.y < south ||
				p.y > north) {
				p.x = -360000000;
				p.y = -360000000;
			}
			xy.push_back(p);
		}
	}
	
	size_t  size = xy.size();
	vector<XY> tmp;
	tmp.reserve(2 * size);

	level = 0;

	for (size_t k = 0; k < size; k++) {
		if (xy[k].y > -300000000) {
			level++;
			if (level == 2) {
				XY p1 = {xy[k-1].x, xy[k-1].y};
				tmp.push_back(p1);
				XY p2 = {xy[k].x, xy[k].y};
				tmp.push_back(p2);
				if (abs(xy[k].x - xy[k-1].x) > 90000000) {
					tmp.pop_back();
					tmp.pop_back();
					level = 0;
				}
			}
			else if (level == 3) {
				tmp.push_back(tmp[tmp.size()-1]);
				XY p3 = {xy[k].x, xy[k].y};
				tmp.push_back(p3);
				level--;
				if (abs(xy[k].x - xy[k-1].x) > 90000000) {
					tmp.pop_back();
					tmp.pop_back();
					level = 0;
				}
			}
		}
		else {
			level = 0;
		}
	}

	CtkObject ret = mat_create_array("double", tmp.size(), 2);
	void *ptr = CtkObject2Ptr(ret);
	for (k = 0; k < tmp.size(); k++) {
		XY p = tmp[k];
		array_set_at(ptr, k, 0, 1.e-6 * p.x);
		array_set_at(ptr, k, 1, 1.e-6 * p.y);
	}

	return ret;
}
static CtkPrimitive mat49("matReadGSHHS", matReadGSHHS);


static CtkObject matcos(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	CtkObject ret;
	if (IS_NUMBER(args[0])) {
		MAKE_REAL(ret, cos(GET_NUMBER(args[0])));
		return ret;
	}
	else {
		ret = mat_make_array(args[0], true);
		if (amath_cos(CtkObject2Ptr(ret)) == 0) ctkThrowException(array_get_error_message());
		return ret;
	}
}
static CtkPrimitive mat50("cos", matcos);


static CtkObject matacos(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	CtkObject ret;
	if (IS_NUMBER(args[0])) {
		MAKE_REAL(ret, acos(GET_NUMBER(args[0])));
	}
	else {
		ret = mat_make_array(args[0], true);
		if (amath_acos(CtkObject2Ptr(ret)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive mat51("acos", matacos);


static CtkObject matcosh(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	CtkObject ret;
	if (IS_NUMBER(args[0])) {
		MAKE_REAL(ret, cosh(GET_NUMBER(args[0])));
	}
	else {
		ret = mat_make_array(args[0], true);
		if (amath_cosh(CtkObject2Ptr(ret)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive mat52("cosh", matcosh);


static CtkObject matsin(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	CtkObject ret;
	if (IS_NUMBER(args[0])) {
		MAKE_REAL(ret, sin(GET_NUMBER(args[0])));
	}
	else {
		ret = mat_make_array(args[0], true);
		if (amath_sin(CtkObject2Ptr(ret)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive mat53("sin", matsin);


static CtkObject matasin(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	CtkObject ret;
	if (IS_NUMBER(args[0])) {
		MAKE_REAL(ret, asin(GET_NUMBER(args[0])));
	}
	else {
		ret = mat_make_array(args[0], true);
		if (amath_asin(CtkObject2Ptr(ret)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive mat54("asin", matasin);


static CtkObject matsinh(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	CtkObject ret;
	if (IS_NUMBER(args[0])) {
		MAKE_REAL(ret, sinh(GET_NUMBER(args[0])));
	}
	else {
		ret = mat_make_array(args[0], true);
		if (amath_sinh(CtkObject2Ptr(ret)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive mat55("sinh", matsinh);


static CtkObject mattan(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	CtkObject ret;
	if (IS_NUMBER(args[0])) {
		MAKE_REAL(ret, tan(GET_NUMBER(args[0])));
	}
	else {
		ret = mat_make_array(args[0], true);
		if (amath_tan(CtkObject2Ptr(ret)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive mat57("tan", mattan);


static CtkObject matatan(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	CtkObject ret;
	if (IS_NUMBER(args[0])) {
		MAKE_REAL(ret, atan(GET_NUMBER(args[0])));
	}
	else {
		ret = mat_make_array(args[0], true);
		if (amath_atan(CtkObject2Ptr(ret)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive mat58("atan", matatan);


static CtkObject mattanh(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	CtkObject ret;
	if (IS_NUMBER(args[0])) {
		MAKE_REAL(ret, tanh(GET_NUMBER(args[0])));
	}
	else {
		ret = mat_make_array(args[0], true);
		if (amath_tanh(CtkObject2Ptr(ret)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive mat59("tanh", mattanh);


static CtkObject matsqrt(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	CtkObject ret;
	if (IS_NUMBER(args[0])) {
		MAKE_REAL(ret, sqrt(GET_NUMBER(args[0])));
	}
	else {
		ret = mat_make_array(args[0], true);
		if (amath_sqrt(CtkObject2Ptr(ret)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive mat60("sqrt", matsqrt);


static CtkObject matexp(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	CtkObject ret;
	if (IS_NUMBER(args[0])) {
		MAKE_REAL(ret, exp(GET_NUMBER(args[0])));
	}
	else {
		ret = mat_make_array(args[0], true);
		if (amath_exp(CtkObject2Ptr(ret)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive mat61("exp", matexp);


static CtkObject matlog(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	CtkObject ret;
	if (IS_NUMBER(args[0])) {
		if (GET_NUMBER(args[0]) == 0) invalid_input();
		MAKE_REAL(ret, log(GET_NUMBER(args[0])));
	}
	else {
		ret = mat_make_array(args[0], true);
		if (amath_log(CtkObject2Ptr(ret)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive mat62("log", matlog);


static CtkObject matlog10(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	CtkObject ret;
	if (IS_NUMBER(args[0])) {
		if (GET_NUMBER(args[0]) == 0) invalid_input();
		MAKE_REAL(ret, log10(GET_NUMBER(args[0])));
	}
	else {
		ret = mat_make_array(args[0], true);
		if (amath_log10(CtkObject2Ptr(ret)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive mat63("log10", matlog10);


static CtkObject matceil(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	CtkObject ret;
	if (IS_NUMBER(args[0])) {
		MAKE_REAL(ret, ceil(GET_NUMBER(args[0])));
	}
	else {
		ret = mat_make_array(args[0], true);
		if (amath_ceil(CtkObject2Ptr(ret)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive mat64("ceil", matceil);


static CtkObject matfloor(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	CtkObject ret;
	if (IS_NUMBER(args[0])) {
		MAKE_REAL(ret, floor(GET_NUMBER(args[0])));
	}
	else {
		ret = mat_make_array(args[0], true);
		if (amath_floor(CtkObject2Ptr(ret)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive mat65("floor", matfloor);


static CtkObject matabs(int nArgs, CtkObject* args)
{
	if (nArgs < 1) invalid_input();
	CtkObject ret;
	if (IS_NUMBER(args[0])) {
		MAKE_REAL(ret, fabs(GET_NUMBER(args[0])));
	}
	else {
		ret = mat_make_array(args[0], true);
		if (amath_abs(CtkObject2Ptr(ret)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive mat66("abs", matabs);


static CtkObject matpow(int nArgs, CtkObject* args)
{
	if (nArgs < 2) invalid_input();
	CtkObject ret;
	if (IS_NUMBER(args[1])) {
		if (IS_NUMBER(args[0])) {
			MAKE_REAL(ret, pow(GET_NUMBER(args[0]), GET_NUMBER(args[1])));
		}
		else {
			ret = mat_make_array(args[0], true);
			if (amath_pow(CtkObject2Ptr(ret),
						  GET_NUMBER(args[1])) == 0) ctkThrowException(array_get_error_message());
		}
	}
	else {
		if (IS_NUMBER(args[0])) {
			ret = mat_make_array(args[1], true);
			if (amath_pow_ex(CtkObject2Ptr(ret),
							 GET_NUMBER(args[0])) == 0) ctkThrowException(array_get_error_message());
		}
		else {
			ret = mat_make_array(args[0], true);
			if (amath_pow2(CtkObject2Ptr(args[0]),
							CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());
		}
	}
	return ret;
}
static CtkPrimitive mat68("pow", matpow);


static CtkObject matMakeIdentity(int nArgs, CtkObject* args)
{
	if (nArgs < 1 || !IS_NUMBER(args[0])) invalid_input();
	CtkObject ret = mat_create_array("double", 0, 0);
	if (amath_make_identity(CtkObject2Ptr(ret),
							GET_NUMBER(args[0])) == 0) ctkThrowException(array_get_error_message());
	return ret;
}
static CtkPrimitive mat69("matMakeIdentity", matMakeIdentity);


static CtkObject matMakeRotateX(int nArgs, CtkObject* args)
{
	if (nArgs < 1 || !IS_NUMBER(args[0])) invalid_input();
	CtkObject ret = mat_create_array("double", 0, 0);
	if (amath_make_rotatex(CtkObject2Ptr(ret),
						   GET_NUMBER(args[0])) == 0) ctkThrowException(array_get_error_message());
	return ret;
}
static CtkPrimitive mat70("matMakeRotateX", matMakeRotateX);


static CtkObject matMakeRotateY(int nArgs, CtkObject* args)
{
	if (nArgs < 1 || !IS_NUMBER(args[0])) invalid_input();
	CtkObject ret = mat_create_array("double", 0, 0);
	if (amath_make_rotatey(CtkObject2Ptr(ret),
						   GET_NUMBER(args[0])) == 0) ctkThrowException(array_get_error_message());
	return ret;
}
static CtkPrimitive mat71("matMakeRotateY", matMakeRotateY);


static CtkObject matMakeRotateZ(int nArgs, CtkObject* args)
{
	if (nArgs < 1 || !IS_NUMBER(args[0])) invalid_input();
	CtkObject ret = mat_create_array("double", 0, 0);
	if (amath_make_rotatez(CtkObject2Ptr(ret),
						   GET_NUMBER(args[0])) == 0) ctkThrowException(array_get_error_message());
	return ret;
}
static CtkPrimitive mat72("matMakeRotateZ", matMakeRotateZ);


static CtkObject matMakeScale(int nArgs, CtkObject* args)
{
	if (nArgs < 3 ||
		!IS_NUMBER(args[0]) ||
		!IS_NUMBER(args[1]) ||
		!IS_NUMBER(args[2])) invalid_input();
	CtkObject ret = mat_create_array("double", 0, 0);
	if (amath_make_scale(CtkObject2Ptr(ret),
						 GET_NUMBER(args[0]),
						 GET_NUMBER(args[1]),
						 GET_NUMBER(args[2])) == 0) ctkThrowException(array_get_error_message());
	return ret;
}
static CtkPrimitive mat75("matMakeScale", matMakeScale);


static CtkObject matMakeTranslate(int nArgs, CtkObject* args)
{
	if (nArgs < 3 ||
		!IS_NUMBER(args[0]) ||
		!IS_NUMBER(args[1]) ||
		!IS_NUMBER(args[2])) invalid_input();
	CtkObject ret = mat_create_array("double", 0, 0);
	if (amath_make_translate(CtkObject2Ptr(ret),
							 GET_NUMBER(args[0]),
							 GET_NUMBER(args[1]),
							 GET_NUMBER(args[2])) == 0) ctkThrowException(array_get_error_message());
	return ret;
}
static CtkPrimitive mat76("matMakeTranslate", matMakeTranslate);


static CtkObject matClone(int nArgs, CtkObject* args)
{
	if (nArgs < 2) invalid_input();
	if (array_clone(CtkObject2Ptr(args[0]),
					CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());
	return ctkNull;
}
static CtkPrimitive mat77("matClone", matClone);


////////////////// manipulate array by operators //////////////////////////////

static bool is_equal_type(CtkObject &lvalue, CtkObject &rvalue)
{
	int ltype = 0, rtype = 0;
	if (IS_USER_TYPE(lvalue)) ltype = array_check_type(CtkObject2Ptr(lvalue));
	if (IS_USER_TYPE(rvalue)) rtype = array_check_type(CtkObject2Ptr(rvalue));
	if (ltype == 0 && rtype == 0) ctkThrowException(": User type not a matrix.");
	if (ltype == rtype) return true;
	return false;
}


static int make_matrix_operands(CtkObject &lvalue, CtkObject &rvalue, CtkObject &arg1, CtkObject &arg2)
{
	int ltype = 0, rtype = 0;
	if (IS_USER_TYPE(lvalue)) ltype = array_check_type(CtkObject2Ptr(lvalue));
	if (IS_USER_TYPE(rvalue)) rtype = array_check_type(CtkObject2Ptr(rvalue));
	if (ltype == 0 && rtype == 0) ctkThrowException(": User type not a matrix.");
	if (ltype == rtype) {
		arg1 = mat_make_array(lvalue, true);
		arg2 = mat_make_array(rvalue, true);
		return  1;
	}
	else if (ltype > rtype) {
		arg1 = mat_make_array(lvalue, true);
		arg2 = mat_make_array(lvalue, false);
		if (rtype > 0) {
			// array promotion
			if (!array_clone(CtkObject2Ptr(arg2),
							 CtkObject2Ptr(rvalue))) ctkThrowException(array_get_error_message());
		}
		else {
			// make array from number
			size_t nrow, ncol;
			array_get_size(CtkObject2Ptr(lvalue), &nrow, &ncol);
			array_resize(CtkObject2Ptr(arg2), nrow, ncol, false);
			if (!IS_NUMBER(rvalue)) invalid_input();
			array_fill(CtkObject2Ptr(arg2), GET_NUMBER(rvalue), 0);
		}
		return  1;
	}
	else {
		arg1 = mat_make_array(rvalue, false);
		arg2 = mat_make_array(rvalue, true);
		if (ltype > 0) {
			// array promotion
			if (!array_clone(CtkObject2Ptr(arg1),
							 CtkObject2Ptr(lvalue))) ctkThrowException(array_get_error_message());
		}
		else {
			// array from number
			size_t nrow, ncol;
			array_get_size(CtkObject2Ptr(rvalue), &nrow, &ncol);
			array_resize(CtkObject2Ptr(arg1), nrow, ncol, false);
			if (!IS_NUMBER(lvalue)) invalid_input();
			array_fill(CtkObject2Ptr(arg1), GET_NUMBER(lvalue), 0);
		}
		return -1;
	}
}


static CtkObject matAddExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2) invalid_input();
	CtkObject arg1, arg2;
	make_matrix_operands(args[0], args[1], arg1, arg2);
	if (array_add2(CtkObject2Ptr(arg1),
				   CtkObject2Ptr(arg2)) == 0) ctkThrowException(array_get_error_message());
	return arg1;
}
static CtkPrimitive op1("AddExpr", matAddExpr);


static CtkObject matMulExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2) invalid_input();
	CtkObject arg1, arg2;
	make_matrix_operands(args[0], args[1], arg1, arg2);
	if (array_multiply2(CtkObject2Ptr(arg1),
						CtkObject2Ptr(arg2)) == 0) ctkThrowException(array_get_error_message());
	return arg1;
}
static CtkPrimitive op2("MulExpr", matMulExpr);


static CtkObject matSubExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2) invalid_input();
	CtkObject arg1, arg2;
	int flag = make_matrix_operands(args[0], args[1], arg1, arg2);
	if (flag > 0) {
		if (array_subtract2(CtkObject2Ptr(arg1),
							CtkObject2Ptr(arg2)) == 0) ctkThrowException(array_get_error_message());
		return arg1;
	}
	else {
		if (array_negate(CtkObject2Ptr(arg2)) == 0) ctkThrowException(array_get_error_message());
		if (array_add2(CtkObject2Ptr(arg2),
						CtkObject2Ptr(arg1)) == 0) ctkThrowException(array_get_error_message());
		return arg2;
	}
	return ctkNull;
}
static CtkPrimitive op3("SubExpr", matSubExpr);


static CtkObject matDivExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2) invalid_input();
	CtkObject arg1, arg2;
	int flag = make_matrix_operands(args[0], args[1], arg1, arg2);
	if (flag > 0) {
		if (array_divide2(CtkObject2Ptr(arg1),
						CtkObject2Ptr(arg2)) == 0) ctkThrowException(array_get_error_message());
		return arg1;
	}
	else {
		if (array_element_inverse(CtkObject2Ptr(arg2)) == 0) ctkThrowException(array_get_error_message());
		if (array_multiply2(CtkObject2Ptr(arg2),
							CtkObject2Ptr(arg1)) == 0) ctkThrowException(array_get_error_message());
		return arg2;
	}
	return ctkNull;
}
static CtkPrimitive op4("DivExpr", matDivExpr);


static CtkObject matModExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2) invalid_input();
	CtkObject arg1, arg2;
	int flag = make_matrix_operands(args[0], args[1], arg1, arg2);
	if (flag > 0) {
		if (amath_mod2(CtkObject2Ptr(arg1),
						CtkObject2Ptr(arg2)) == 0) ctkThrowException(array_get_error_message());
		return arg1;
	}
	else {
		if (amath_mod2(CtkObject2Ptr(arg2),
						CtkObject2Ptr(arg1)) == 0) ctkThrowException(array_get_error_message());
		return arg2;
	}
	return ctkNull;
}
static CtkPrimitive op5("ModExpr", matModExpr);


static CtkObject matOrExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2) invalid_input();
	CtkObject ret = mat_create_array("int8", 0, 0);
	if (is_equal_type(args[0], args[1])) {
		if (array_or(CtkObject2Ptr(ret),
					 CtkObject2Ptr(args[0]),
					 CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());
	}
	else {
		CtkObject arg1, arg2;
		make_matrix_operands(args[0], args[1], arg1, arg2);
		if (array_or(CtkObject2Ptr(ret),
					 CtkObject2Ptr(arg1),
					 CtkObject2Ptr(arg2)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive op6("OrExpr", matOrExpr);


static CtkObject matAndExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2) invalid_input();
	CtkObject ret = mat_create_array("int8", 0, 0);
	if (is_equal_type(args[0], args[1])) {
		if (array_and(CtkObject2Ptr(ret),
					 CtkObject2Ptr(args[0]),
					 CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());
	}
	else {
		CtkObject arg1, arg2;
		make_matrix_operands(args[0], args[1], arg1, arg2);
		if (array_and(CtkObject2Ptr(ret),
					 CtkObject2Ptr(arg1),
					 CtkObject2Ptr(arg2)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive op7("AndExpr", matAndExpr);


static CtkObject matXorExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2) invalid_input();
	CtkObject ret = mat_create_array("int8", 0, 0);
	if (is_equal_type(args[0], args[1])) {
		if (array_xor(CtkObject2Ptr(ret),
					 CtkObject2Ptr(args[0]),
					 CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());
	}
	else {
		CtkObject arg1, arg2;
		make_matrix_operands(args[0], args[1], arg1, arg2);
		if (array_xor(CtkObject2Ptr(ret),
					 CtkObject2Ptr(arg1),
					 CtkObject2Ptr(arg2)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive op8("XorExpr", matXorExpr);


static CtkObject matShlExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2 || !IS_NUMBER(args[1])) invalid_input();
	CtkObject ret = mat_make_array(args[0], true);
	if (array_shift_columns(CtkObject2Ptr(ret),
							GET_NUMBER(args[1])) == 0) ctkThrowException(array_get_error_message());

	return ret;
}
static CtkPrimitive op9("ShlExpr", matShlExpr);


static CtkObject matShrExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2 || !IS_NUMBER(args[1])) invalid_input();
	CtkObject ret = mat_make_array(args[0], true);
	if (array_shift_columns(CtkObject2Ptr(ret),
							-GET_NUMBER(args[1])) == 0) ctkThrowException(array_get_error_message());

	return ret;
}
static CtkPrimitive op10("ShrExpr", matShrExpr);


static CtkObject matEqExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2) invalid_input();
	CtkObject ret = mat_create_array("int8", 0, 0);
	if (is_equal_type(args[0], args[1])) {
		if (array_equal2(CtkObject2Ptr(ret),
						 CtkObject2Ptr(args[0]),
						 CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());
	}
	else {
		CtkObject arg1, arg2;
		make_matrix_operands(args[0], args[1], arg1, arg2);
		if (array_equal2(CtkObject2Ptr(ret),
						 CtkObject2Ptr(arg1),
						CtkObject2Ptr(arg2)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive op11("EqExpr", matEqExpr);


static CtkObject matNeExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2) invalid_input();
	CtkObject ret = mat_create_array("int8", 0, 0);
	if (is_equal_type(args[0], args[1])) {
		if (array_nequal2(CtkObject2Ptr(ret),
						  CtkObject2Ptr(args[0]),
						  CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());
	}
	else {
		CtkObject arg1, arg2;
		make_matrix_operands(args[0], args[1], arg1, arg2);
		if (array_nequal2(CtkObject2Ptr(ret),
						  CtkObject2Ptr(arg1),
						  CtkObject2Ptr(arg2)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive op12("NeExpr", matNeExpr);


static CtkObject matGtExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2) invalid_input();
	CtkObject ret = mat_create_array("int8", 0, 0);
	if (is_equal_type(args[0], args[1])) {
		if (array_greater2(CtkObject2Ptr(ret),
							CtkObject2Ptr(args[0]),
							CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());
	}
	else {
		CtkObject arg1, arg2;
		make_matrix_operands(args[0], args[1], arg1, arg2);
		if (array_greater2(CtkObject2Ptr(ret),
							CtkObject2Ptr(arg1),
							CtkObject2Ptr(arg2)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive op13("GtExpr", matGtExpr);


static CtkObject matGeExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2) invalid_input();
	CtkObject ret = mat_create_array("int8", 0, 0);
	if (is_equal_type(args[0], args[1])) {
		if (array_gequal2(CtkObject2Ptr(ret),
							CtkObject2Ptr(args[0]),
							CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());
	}
	else {
		CtkObject arg1, arg2;
		make_matrix_operands(args[0], args[1], arg1, arg2);
		if (array_gequal2(CtkObject2Ptr(ret),
							CtkObject2Ptr(arg1),
							CtkObject2Ptr(arg2)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive op14("GeExpr", matGeExpr);


static CtkObject matLtExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2) invalid_input();
	CtkObject ret = mat_create_array("int8", 0, 0);
	if (is_equal_type(args[0], args[1])) {
		if (array_less2(CtkObject2Ptr(ret),
						CtkObject2Ptr(args[0]),
						CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());
	}
	else {
		CtkObject arg1, arg2;
		make_matrix_operands(args[0], args[1], arg1, arg2);
		if (array_less2(CtkObject2Ptr(ret),
						CtkObject2Ptr(arg1),
						CtkObject2Ptr(arg2)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive op15("LtExpr", matLtExpr);


static CtkObject matLeExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2) invalid_input();
	CtkObject ret = mat_create_array("int8", 0, 0);
	if (is_equal_type(args[0], args[1])) {
		if (array_lequal2(CtkObject2Ptr(ret),
							CtkObject2Ptr(args[0]),
							CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());
	}
	else {
		CtkObject arg1, arg2;
		make_matrix_operands(args[0], args[1], arg1, arg2);
		if (array_lequal2(CtkObject2Ptr(ret),
							CtkObject2Ptr(arg1),
							CtkObject2Ptr(arg2)) == 0) ctkThrowException(array_get_error_message());
	}
	return ret;
}
static CtkPrimitive op16("LeExpr", matLeExpr);


static CtkObject matSetAddExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2 ) invalid_input();
	if (IS_NUMBER(args[1])) {
		if (array_add(CtkObject2Ptr(args[0]),
						GET_NUMBER(args[1])) == 0) ctkThrowException(array_get_error_message());
	}
	else {
		CtkObject arg;
		if (is_equal_type(args[0], args[1])) {
			arg = args[1];
		}
		else {
			arg = mat_make_array(args[0], false);
			if (array_clone(CtkObject2Ptr(arg),
							CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());
		}
		if (array_add2(CtkObject2Ptr(args[0]),
						CtkObject2Ptr(arg)) == 0) ctkThrowException(array_get_error_message());
	}
	return args[0];
}
static CtkPrimitive op17("SetAddExpr", matSetAddExpr);


static CtkObject matSetSubExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2) invalid_input();
	if (IS_NUMBER(args[1])) {
		if (array_subtract(CtkObject2Ptr(args[0]),
							GET_NUMBER(args[1])) == 0) ctkThrowException(array_get_error_message());
	}
	else {
		CtkObject arg;
		if (is_equal_type(args[0], args[1])) {
			arg = args[1];
		}
		else {
			arg = mat_make_array(args[0], false);
			if (array_clone(CtkObject2Ptr(arg),
							CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());
		}
		if (array_subtract2(CtkObject2Ptr(args[0]),
							CtkObject2Ptr(arg)) == 0) ctkThrowException(array_get_error_message());
	}
	return args[0];
}
static CtkPrimitive op18("SetSubExpr", matSetSubExpr);


static CtkObject matSetMulExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2) invalid_input();
	if (IS_NUMBER(args[1])) {
		if (array_multiply(CtkObject2Ptr(args[0]),
							GET_NUMBER(args[1])) == 0) ctkThrowException(array_get_error_message());
	}
	else {
		CtkObject arg;
		if (is_equal_type(args[0], args[1])) {
			arg = args[1];
		}
		else {
			arg = mat_make_array(args[0], false);
			if (array_clone(CtkObject2Ptr(arg),
							CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());
		}
		if (array_multiply2(CtkObject2Ptr(args[0]),
							CtkObject2Ptr(arg)) == 0) ctkThrowException(array_get_error_message());
	}

	return args[0];
}
static CtkPrimitive op19("SetMulExpr", matSetMulExpr);


static CtkObject matSetDivExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2) invalid_input();

	if (IS_NUMBER(args[1])) {
		if (array_divide(CtkObject2Ptr(args[0]),
						GET_NUMBER(args[1])) == 0) ctkThrowException(array_get_error_message());
	}
	else {
		CtkObject arg;
		if (is_equal_type(args[0], args[1])) {
			arg = args[1];
		}
		else {
			arg = mat_make_array(args[0], false);
			if (array_clone(CtkObject2Ptr(arg),
							CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());
		}
		if (array_divide2(CtkObject2Ptr(args[0]),
							CtkObject2Ptr(arg)) == 0) ctkThrowException(array_get_error_message());
	}

	return args[0];
}
static CtkPrimitive op20("SetDivExpr", matSetDivExpr);


static CtkObject matSetModExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2) invalid_input();

	if (IS_NUMBER(args[1])) {
		if (amath_mod(CtkObject2Ptr(args[0]),
					GET_NUMBER(args[1])) == 0) ctkThrowException(array_get_error_message());
	}
	else {
		CtkObject arg;
		if (is_equal_type(args[0], args[1])) {
			arg = args[1];
		}
		else {
			arg = mat_make_array(args[0], false);
			if (array_clone(CtkObject2Ptr(arg),
							CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());
		}
		if (amath_mod2(CtkObject2Ptr(args[0]),
						CtkObject2Ptr(arg)) == 0) ctkThrowException(array_get_error_message());
	}

	return args[0];
}
static CtkPrimitive op21("SetModExpr", matSetModExpr);


static CtkObject matSetOrExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2 ||
		!array_is_int8(CtkObject2Ptr(args[0])) ||
		!array_is_int8(CtkObject2Ptr(args[1])) ) invalid_input();
	if (array_or(CtkObject2Ptr(args[0]),
				 CtkObject2Ptr(args[0]),
				 CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());

	return args[0];
}
static CtkPrimitive op22("SetOrExpr", matSetOrExpr);


static CtkObject matSetAndExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2 ||
		!array_is_int8(CtkObject2Ptr(args[0])) ||
		!array_is_int8(CtkObject2Ptr(args[1])) ) invalid_input();
	if (array_and(CtkObject2Ptr(args[0]),
				  CtkObject2Ptr(args[0]),
				  CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());

	return args[0];
}
static CtkPrimitive op23("SetAndExpr", matSetAndExpr);


static CtkObject matSetXorExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2 ||
		!array_is_int8(CtkObject2Ptr(args[0])) ||
		!array_is_int8(CtkObject2Ptr(args[1])) ) invalid_input();
	if (array_xor(CtkObject2Ptr(args[0]),
				  CtkObject2Ptr(args[0]),
				  CtkObject2Ptr(args[1])) == 0) ctkThrowException(array_get_error_message());

	return args[0];
}
static CtkPrimitive op24("SetXorExpr", matSetXorExpr);


static CtkObject matSetShlExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2 || !IS_NUMBER(args[1]) ) invalid_input();
	if (array_shift_columns(CtkObject2Ptr(args[0]),
							GET_NUMBER(args[1])) == 0) ctkThrowException(array_get_error_message());
	return args[0];
}
static CtkPrimitive op25("SetShlExpr", matSetShlExpr);


static CtkObject matSetShrExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 2 || !IS_NUMBER(args[1]) ) invalid_input();
	if (array_shift_columns(CtkObject2Ptr(args[0]),
							-GET_NUMBER(args[1])) == 0) ctkThrowException(array_get_error_message());
	return args[0];
}
static CtkPrimitive op26("SetShrExpr", matSetShrExpr);


static CtkObject matNegExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 1) invalid_input();
	CtkObject ret = mat_make_array(args[0], true);
	if (array_negate(CtkObject2Ptr(ret)) == 0) ctkThrowException(array_get_error_message());
	return ret;
}
static CtkPrimitive op27("NegExpr", matNegExpr);


static CtkObject matPreIncExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 1) invalid_input();
	if (array_add(CtkObject2Ptr(args[0]), 1) == 0) ctkThrowException(array_get_error_message());
	return args[0];
}
static CtkPrimitive op28("PreIncExpr", matPreIncExpr);


static CtkObject matPreDecExpr(int nArgs, CtkObject* args)
{
	if (nArgs != 1) invalid_input();
	if (array_subtract(CtkObject2Ptr(args[0]), 1) == 0) ctkThrowException(array_get_error_message());
	return args[0];
}
static CtkPrimitive op29("PreDecExpr", matPreDecExpr);


static CtkObject matGetAtExpr(int nArgs, CtkObject* args)
{
	if (nArgs < 2 || !IS_NUMBER(args[1])) invalid_input();
	void *ptr = CtkObject2Ptr(args[0]);
	int irow = 0, icol = 0;
#ifdef ROW_MAJOR
	icol = GET_NUMBER(args[1]);
#else
	irow = GET_NUMBER(args[1]);
#endif
	double v = array_get_at(ptr, irow, icol);
	CtkObject ret;
	if (array_is_double(ptr) || array_is_float(ptr)) {
		MAKE_REAL(ret, v);
	}
	else {
		MAKE_INTEGER(ret, v);
	}
	return ret;
}
static CtkPrimitive mat15("GetAtExpr", matGetAtExpr);


static CtkObject matSetAtExpr(int nArgs, CtkObject* args)
{
	if (nArgs < 3 || !IS_NUMBER(args[1]) || !IS_NUMBER(args[2])) invalid_input();
	int irow = 0, icol = 0;
#ifdef ROW_MAJOR
	icol = GET_NUMBER(args[1]);
#else
	irow = GET_NUMBER(args[1]);
#endif
	if (array_set_at(CtkObject2Ptr(args[0]),
					 irow,
					 icol,
					 GET_NUMBER(args[2])) == 0) ctkThrowException(array_get_error_message());
	return ctkNull;
}
static CtkPrimitive mat16("SetAtExpr", matSetAtExpr);


static CtkObject matIsRowMajor(int nArgs, CtkObject* args)
{
	CtkObject ret;
#ifdef ROW_MAJOR
	MAKE_INTEGER(ret, 1);
#else
	MAKE_INTEGER(ret, 0);
#endif
	return ret;
}
static CtkPrimitive mat78("matIsRowMajor", matIsRowMajor);

